/*
 * a1-i.f.c
 * Copyright (C) 2011-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <a1-inter.h>
#include <a1-types.h>
#include <e4.h>
#include <fd.h>
#include <lxcall.h>
#include <lxcast.h>

#define POST_BITS \
    X1f4_LXCALL_POST_ARGS | X1f4_LXCALL_POST_DATA | X1f4_LXCALL_POST_LINE

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

typedef struct pick_type {
    struct fxdata_type finite;
    struct x1f4_dxcast_type dxcast, serial;
    struct x1f4_function_type function;
    struct x1f4_linetext_type linetext;
} pick_type;

int
_x1f4_a1_fold_call(struct a1_instant_type *instant_data, void *output,
		   void **input, struct x1f4_linetext_type *linetext_data,
		   int head, int tail,
		   const struct x1f4_function_type *function_text, int lead,
		   int lock)
{
    int linked, status;
    const struct x1f4_function_type *function_data;
    unsigned class, count, flags, grade, parts, scale = 0;
    void *beta;

    function_data = &instant_data->fail_node->linetext.function;

    if (function_data->flags & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	function_data = instant_data->s.function_data;
    }

    flags = function_data->flags;

    count = function_data->count;

    if (head < 0) {
	head += count;
    }
    if (tail < 0) {
	tail += count;
    }

    class = function_text->count - lead;

    grade = class + 1;

    if (lock < 0) {
	lock += grade;
    }

    parts = function_text->flags;

    if ((flags | parts) & X1f4_E4_POST_TYPE) {
	if (linetext_data->function.flags
	    & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	    scale = 1;
	} else {
	}
    }

    linked = sizeof(struct pick_type);

    status = instant_data->v.link
	(instant_data->v.data, &beta,
	 linked + (sizeof(void *) + (sizeof(int) << scale)) * grade
	 + sizeof(void *));
    if (status) {
	status = _x1f4_a1_stat_link(instant_data->shuffle_data);
    } else {
	int *args, call, type;
	struct a1_shuffle_type *shuffle_data;
	struct pick_type *call_data;
	struct x1f4_function_type *function_call;
	unsigned bits;
	void **frame;

	call_data = beta;

	function_call = &call_data->function;

	function_call->count = grade;

	function_call->flags = scale ? X1f4_E4_POST_TYPE : 0;

	frame = (void *) (call_data + 1);
	frame++;

	function_call->args = args = (void *) (frame + grade);

	frame[-1] = input[-1];

	tail++;

	shuffle_data = instant_data->shuffle_data;

	type = linetext_data->function.type;

	if (type != X1f4_E4_CASE) {
	    l_USER(&call_data->dxcast, output);

	    call_data->dxcast.call = 0;
	    call_data->dxcast.type = type;

	    output = &call_data->dxcast.data;
	} else {
	}

	bits = POST_BITS;

	call = head;

	_x1f4_a1_hold_call
	    (frame, input, parts, flags, input[-1], args, instant_data,
	     function_text->args, function_data->args, class, lead, lock,
	     count, call, scale);

	{
	    /* the args, count, flags set should do */
	    call_data->linetext = *linetext_data;

	    call_data->linetext.function.type = X1f4_E4_CASE;

	    status = _libx1f4i0_lxcall_post_call
		(&call_data->linetext, function_call, output, 0, frame, bits,
		 &instant_data->v, &shuffle_data->p, &shuffle_data->q,
		 &shuffle_data->x, shuffle_data->y.eelookup_data);
	    if (status) {
		call = tail;
	    } else {
		call++;
	    }
	}
	while (call != tail) {
	    if (call == count) {
		call = 0;
	    }

	    _x1f4_a1_keep_call
		(frame, flags, input[-1], args, instant_data,
		 function_data->args, class, lock, count, call, scale);

	    /* the args, count, flags set should do */
	    call_data->linetext = *linetext_data;

	    call_data->linetext.function.type = X1f4_E4_CASE;

	    status = _libx1f4i0_lxcall_rail_call
		(&call_data->linetext, function_call, output, 0, frame, bits,
		 &shuffle_data->p, &shuffle_data->q, &shuffle_data->x,
		 shuffle_data->y.eelookup_data, 1, &lock, &call_data->finite,
		 &call_data->serial);
	    if (status) {
		break;
	    }

	    call++;
	}
    }

    return status;
}
