/*
 * a1-l.o.c
 * Copyright (C) 2011-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <a1-inter.h>
#include <a1-types.h>
#include <e4.h>
#include <lxcast.h>

#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

extern const void *const x1f4_a1_walk_link;

static int beta_call(void *, void *, void **);

static const int c_c_m_m_m[] = {
/* *INDENT-OFF* */
    X1f4_E4_CASE,
    X1f4_E4_MODE,
    X1f4_E4_MODE,
    X1f4_E4_MODE,
    X1f4_E4_MODE
/* *INDENT-ON* */
};
const struct x1f4_function_type _x1f4_a1_l_o = {
/* *INDENT-OFF* */
    "ocall",
    beta_call,
    X1f4_E4_CASE,
    c_c_m_m_m,
    4,
    X1f4_E4_KEEP_CALL | X1f4_E4_LINK_PASS | X1f4_E4_SIDE_LIST
    | X1f4_E4_WALK_LINK,
    5
/* *INDENT-ON* */
};

static int
beta_call(void *context, void *output, void **input)
{
    int status;

    do {
	X1f4_E4_C_MODE head, lock, tail;
	struct a1_instant_type *instant_data;
	struct x1f4_dxcast_type *dxcast_data;
	const struct x1f4_function_type *function_data, *function_text;
	struct x1f4_linetext_type *linetext_data;
	unsigned class, count;
	void **track, **used;

	if (output == x1f4_a1_walk_link) {
	    track = input[1];
	} else {
	    track = input;
	}

	used = track[-1];

	instant_data = used[X1f4_E4_USED_META];

	dxcast_data = I_USER(track[0]);

	if (X1f4_E4_CALL < dxcast_data->type) {
	} else {
	    status = _x1f4_a1_stat_call
		(instant_data->shuffle_data, "ocall", 5);
	    if (1) {
		break;
	    }
	}

	linetext_data = I_USER(&dxcast_data->data);
	if (linetext_data->function.function) {
	} else {
	    status = _x1f4_a1_stat_zero
		(instant_data->shuffle_data, "ocall", 5);
	    if (1) {
		break;
	    }
	}

	lock = I_MODE(track[1]);

	function_text = used[X1f4_E4_RENDITION];

	class = function_text->count - 4 + 1;
	if (lock < -(int) class || !(lock < (int) class)) {
	    status = _x1f4_a1_stat_land
		(instant_data->shuffle_data, lock, class);
	    if (1) {
		break;
	    }
	}

	head = I_MODE(track[2]);
	tail = I_MODE(track[3]);

	function_data = &instant_data->fail_node->linetext.function;

	if (function_data->flags
	    & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    function_data = instant_data->s.function_data;
	}

	count = function_data->count;
	if (head < -(int) count || !(head < (int) count)) {
	    status = _x1f4_a1_stat_rend
		(instant_data->shuffle_data, head, count);
	    if (1) {
		break;
	    }
	}
	if (tail < -(int) count || !(tail < (int) count)) {
	    status = _x1f4_a1_stat_rend
		(instant_data->shuffle_data, tail, count);
	    if (1) {
		break;
	    }
	}

	if (output == x1f4_a1_walk_link) {
	    status = _x1f4_a1_lift_call
		(instant_data, output, input, linetext_data, head, tail,
		 function_text, 4, lock);
	} else {
	    status = _x1f4_a1_fold_call
		(instant_data, output, input, linetext_data, head, tail,
		 function_text, 4, lock);
	}
    } while (0);

    return status;
}
