/*
 * bqfv.i.c
 * Copyright (C) 2009-2012, 2015, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <bqfx-config.h>

#include <bqfx-defs.h>
#include <bqfx-inter.h>
#include <bqfx-names.h>
#include <bqfx-types.h>

#define bqfset(bqfset)			((struct bqfset_type *) (bqfset))

static int case_line(void *, unsigned, void **, unsigned,
		      int (*) (void *, void *),void *);
static int cast_line(void *, void **, unsigned, int (*) (void *, void *),
		     void *);
static int lose_line(void *, void *, void *, unsigned,
		     int (*) (void *, void *), void *);
static int miss_line(void *, unsigned, void *, void *, void **, unsigned,
		     int (*) (void *, void *), void *);
static int part_line(void *, void **, void **, unsigned,
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);
static int port_line(void *, unsigned, void **, void **, unsigned,
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);
static int trap_line(void *, unsigned, void *, unsigned,
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);
static int trip_line(void *, void *, unsigned,
		     int (*) (void *, void *, const void *),
		     int (*) (void *, void *), void *);

static int
case_line(void *bqdata, unsigned rate, void **fine, unsigned size,
	  int (*free) (void *, void *), void *context)
{
    integral_q news;
    void *data, *node;

    node = *fine;

    rate--;

    news = *(integral_q *) node;

    news >>= 1;

    data = node;
    data = (char *) data + size;

    while (news) {
	if (news & 1) {
	    free(context, data);
	}

	news >>= 1;

	data = (char *) data + size;
    }

    news = *(integral_q *) node;

    fine = (void **) ((char *) node + size * integral_q_bits);

    if (rate) {
	do {
	    if (news & 2) {
		case_line(bqdata, rate, fine, size, free, context);
	    }

	    fine++;

	    if (2) {
		case_line(bqdata, rate, fine, size, free, context);
	    }

	    fine++;

	    news >>= 2;
	} while (news);
    } else {
	do {
	    if (news & 2) {
		cast_line(bqdata, fine, size, free, context);
	    }

	    fine++;

	    if (2) {
		cast_line(bqdata, fine, size, free, context);
	    }

	    fine++;

	    news >>= 2;
	} while (news);
    }

    bqfset(bqdata)->link_m.free(bqfset(bqdata)->link_m.data, node);

    return -1;
}


static int
cast_line(void *bqdata, void **fine, unsigned size,
	  int (*free) (void *, void *), void *context)
{
    integral_q news;
    void *data, *node;

    node = *fine;

    news = *(integral_q *) node;

    news >>= 1;

    data = node;
    data = (char *) data + size;

    while (news) {
	if (news & 1) {
	    free(context, data);
	}

	news >>= 1;

	data = (char *) data + size;
    }

    bqfset(bqdata)->link_m.free(bqfset(bqdata)->link_m.data, node);

    return -1;
}


static int
lose_line(void *bqdata, void *node, void *data, unsigned size,
	  int (*free) (void *, void *), void *context)
{
    integral_q news;
    unsigned lead;

    lead = ((char *) data - (char *) node) / size;

    news = *(integral_q *) node;
    news &= ((integral_q) 1 << lead) - 1;

    news >>= 1;

    data = node;
    data = (char *) data + size;

    while (news) {
	if (news & 1) {
	    free(context, data);
	}

	news >>= 1;
	data = (char *) data + size;
    }

    bqfset(bqdata)->link_m.free(bqfset(bqdata)->link_m.data, node);

    return -1;
}


static int
miss_line(void *bqdata, unsigned rate, void *node, void *data, void **fine,
	  unsigned size, int (*free) (void *, void *), void *context)
{
    integral_q news;
    unsigned lead, side;

    lead = ((char *) data - (char *) node) / size;

    news = *(integral_q *) node;
    if (lead < integral_q_bits) {
	news &= ((integral_q) 1 << lead) - 1;
    }

    news >>= 1;

    data = node;
    data = (char *) data + size;

    while (news) {
	if (news & 1) {
	    free(context, data);
	}

	news >>= 1;
	data = (char *) data + size;
    }

    lead = fine - (void **) ((char *) node + size * integral_q_bits);
    side = (lead + 1) & ~1;

    news = *(integral_q *) node;
    if (side < integral_q_bits) {
	news &= ((integral_q) 1 << side) - 1;
    }
    if (lead & 1) {
	news ^= (integral_q) 1 << (lead - 1);
    }

    fine = (void **) ((char *) node + size * integral_q_bits);

    if (rate) {
	while (news) {
	    if (news & 2) {
		case_line(bqdata, rate, fine, size, free, context);
	    }

	    fine++;

	    if (news & 1) {
		case_line(bqdata, rate, fine, size, free, context);
	    }

	    fine++;

	    news >>= 2;
	}
    } else {
	while (news) {
	    if (news & 2) {
		cast_line(bqdata, fine, size, free, context);
	    }

	    fine++;

	    if (news & 1) {
		cast_line(bqdata, fine, size, free, context);
	    }

	    fine++;

	    news >>= 2;
	}
    }

    bqfset(bqdata)->link_m.free(bqfset(bqdata)->link_m.data, node);

    return -1;
}


static int
part_line(void *bqdata, void **near, void **spin, unsigned size,
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;
    void *node;

    status = bqfset(bqdata)->link_m.link
	(bqfset(bqdata)->link_m.data, &node, integral_q_bits * size);
    if (status) {
	status = LINK_ERROR;
    } else {
	void *data;

	data = node;

	do {
	    integral_q news;
	    void *text;

	    text = *spin;

	    news = *(integral_q *) text;

	    *(integral_q *) data = news;

	    data = (char *) data + size;
	    text = (char *) text + size;

	    if (news & 2) {
		status = copy(context, data, text);
		if (status) {
		    status = CALL_ERROR;
		    if (1) {
			break;
		    }
		}
	    }

	    news >>= 2;

	    do {
		data = (char *) data + size;
		text = (char *) text + size;
		status = copy(context, data, text);
		if (status) {
		    status = CALL_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    data = (char *) data + size;
		    text = (char *) text + size;
		    if (news & 2) {
			status = copy(context, data, text);
			if (status) {
			    status = CALL_ERROR;
			    if (1) {
				break;
			    }
			}
		    }
		}

		news >>= 2;
	    } while (news);
	} while (0);

	if (status) {
	    lose_line(bqdata, node, data, size, free, context);
	} else {
	   *near = node;
	}
    }

    return status;
}


static int
port_line(void *bqdata, unsigned rate, void **near, void **spin, unsigned size,
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;
    void *node;

    status = bqfset(bqdata)->link_m.link
	(bqfset(bqdata)->link_m.data, &node,
	 integral_q_bits * (size + SIZEOF_integral_q));
    if (status) {
	status = LINK_ERROR;
    } else {
	void *data, **fine;

	data = node;

	fine = (void **) ((char *) data + size * integral_q_bits);

	rate--;

	do {
	    integral_q news;
	    void **line, *text;

	    text = *spin;

	    line = (void **) ((char *) text + size * integral_q_bits);

	    news = *(integral_q *) text;

	    *(integral_q *) data = news;

	    data = (char *) data + size;
	    text = (char *) text + size;

	    if (news & 2) {
		status = copy(context, data, text);
		if (status) {
		    status = CALL_ERROR;
		    if (1) {
			break;
		    }
		}
	    }

	    data = (char *) data + size;
	    text = (char *) text + size;

	    news >>= 2;

	    do {
		status = copy(context, data, text);
		if (status) {
		    status = CALL_ERROR;
		    if (1) {
			break;
		    }
		}

		data = (char *) data + size;
		text = (char *) text + size;

		if (news & 2) {
		    status = copy(context, data, text);
		    if (status) {
			status = CALL_ERROR;
			if (1) {
			    break;
			}
		    }
		}

		data = (char *) data + size;
		text = (char *) text + size;

		news >>= 2;
	    } while (news);
	    if (news) {
		break;
	    }

	    news = *(integral_q *) *spin;

	    if (rate) {
		do {
		    if (news & 2) {
			status = port_line
			    (bqdata, rate, fine, line, size, copy, free,
			     context);
			if (status) {
			    break;
			}
		    }

		    fine++;
		    line++;

		    status = port_line
			(bqdata, rate, fine, line, size, copy, free, context);
		    if (status) {
			break;
		    }

		    fine++;
		    line++;

		    news >>= 2;
		} while (news);
	    } else {
		do {
		    if (news & 2) {
			status = part_line
			    (bqdata, fine, line, size, copy, free, context);
			if (status) {
			    break;
			}
		    }

		    fine++;
		    line++;

		    status = part_line
			(bqdata, fine, line, size, copy, free, context);
		    if (status) {
			break;
		    }

		    fine++;
		    line++;

		    news >>= 2;
		} while (news);
	    }
	} while (0);

	if (status) {
	    miss_line(bqdata, rate, node, data, fine, size, free, context);
	} else {
	   *near = node;
	}
    }

    return status;
}


static int
trap_line(void *bqdata, unsigned rate, void *text, unsigned size,
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;

    rate--;

#if __MICRO_FSETS__
    do {
#endif				/* __MICRO_FSETS__ */
	unsigned i, j, mind;
	void *data, **fine, **line;

	mind = *(integral_q *) text;

	data = bqfset(bqdata)->link_a.fpnews.node;

#if __MICRO_FSETS__
	status = bqfset(bqdata)->link_m.mode
	    (bqfset(bqdata)->link_m.data, &data,
	     integral_q_bits_and_half * (size + SIZEOF_integral_q));
	if (status) {
	    status = MODE_ERROR;
	    if (1) {
		break;
	    }
	} else {
	    bqfset(bqdata)->link_a.fpnews.node = data;
	}
#endif				/* __MICRO_FSETS__ */

	fine = (void **) ((char *) data + size * integral_q_bits_and_half);
	line = (void **) ((char *) text + size * integral_q_bits_and_half);

	status = 0;

	j = mind + 1;
	i = mind;
	for (; i; i--) {
	    data = (char *) data + size;
	    text = (char *) text + size;
	    status = copy(context, data, text);
	    if (status) {
		status = CALL_ERROR;
		if (1) {
		    break;
		}
	    }
	}
	if (i) {
	} else {
	    if (rate) {
		for (; j; j--) {
		    status = port_line
			(bqdata, rate, fine, line, size, copy, free, context);
		    if (status) {
			break;
		    } else {
			fine++;
			line++;
		    }
		}
	    } else {
		for (; j; j--) {
		    status = part_line
			(bqdata, fine, line, size, copy, free, context);
		    if (status) {
			break;
		    } else {
			fine++;
			line++;
		    }
		}
	    }
	}
	if (i | j) {
	    j = mind + 1 - j;
	    if (rate) {
		for (; j; j--) {
		    fine--;
		    case_line(bqdata, rate, fine, size, free, context);
		}
	    } else {
		for (; j; j--) {
		    fine--;
		    cast_line(bqdata, fine, size, free, context);
		}
	    }
	    i = mind - i;
	    for (; i; i--) {
		data = (char *) data - size;
		free(context, data);
	    }
	} else {
	    integral_q *node;

	    node = (void *) ((char *) data - size * mind);
	    *node = mind;

	    bqfset(bqdata)->link_a.fpnews.link_v.rate = rate + 1;
	}
#if __MICRO_FSETS__
    } while (0);
#endif				/* __MICRO_FSETS__ */

    return status;
}


static int
trip_line(void *bqdata, void *text, unsigned size,
	  int (*copy) (void *, void *, const void *),
	  int (*free) (void *, void *), void *context)
{
    int status;

#if __MICRO_FSETS__
    do {
#endif				/* __MICRO_FSETS__ */
	unsigned i, mind;
#if __MICRO_FSETS__
	unsigned sans;
#endif				/* __MICRO_FSETS__ */
	void *data;

	mind = *(integral_q *) text;

	data = bqfset(bqdata)->link_a.fpnews.node;

#if __MICRO_FSETS__
	sans = mind;
	if (sans) {
	    sans++;
	    sans |= sans >> 1;
	    sans |= sans >> 2;
	    sans |= sans >> 4;
	    sans |= sans >> 8;
	    sans ^= sans >> 1;
	    sans |= sans >> 1;
	    if (sans < mind + 1) {
		sans <<= 1;
	    }
	} else {
	    sans = 3;
	}

	if (sans == integral_q_bits_and_half) {
	    sans *= size + SIZEOF_integral_q;
	} else {
	    sans *= size;
	}

	status = bqfset(bqdata)->link_m.mode
	    (bqfset(bqdata)->link_m.data, &data, sans);
	if (status) {
	    status = MODE_ERROR;
	    if (1) {
		break;
	    }
	} else {
	    bqfset(bqdata)->link_a.fpnews.node = data;
	}
#endif				/* __MICRO_FSETS__ */

	status = 0;

	i = mind;
	for (; i; i--) {
	    data = (char *) data + size;
	    text = (char *) text + size;
	    status = copy(context, data, text);
	    if (status) {
		status = CALL_ERROR;
		if (1) {
		    break;
		}
	    }
	}
	if (i) {
	    mind -= i;
	    for (; mind; mind--) {
		data = (char *) data - size;
		free(context, data);
	    }
	} else {
	    integral_q *node;

	    node = (void *) ((char *) data - size * mind);
	    *node = mind;
	}
#if __MICRO_FSETS__
    } while (0);
#endif				/* __MICRO_FSETS__ */

    return status;
}


int
_libx1f4l2_rank_bqfset(void *bqdata, void *bqtext,
		       int (*copy) (void *, void *, const void *),
		       int (*free) (void *, void *), void *context)
{
    int status;
    unsigned rate, size;
    void *node;

    size = bqfset(bqtext)->link_a.fpnews.link_v.size;

    rate = bqfset(bqtext)->link_a.fpnews.link_v.rate;

    node = bqfset(bqtext)->link_a.fpnews.node;

    if (rate) {
	status = trap_line(bqdata, rate, node, size, copy, free, context);
    } else {
	status = trip_line(bqdata, node, size, copy, free, context);
    }

    return status;
}
