/*
 * c1.7.9.c
 * Copyright (C) 2006-2014, 2016-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * TODO
 *
 * reorder variables so those needing processing come first and modify count to
 * indicate their number
 */

#include <stddef.h>
#include <string.h>

#include <c1-defs.h>
#include <c1-inter.h>
#include <c1-parse.h>
#include <c1-types.h>
#include <e4.h>

#define true(e)				1

#define M_FREE(___m, text) \
    parser(___m)->m.free(parser(___m)->m.data, (text))
#define M_LINK(___m, lock, size) \
    parser(___m)->m.link(parser(___m)->m.data, (void *) (lock), (size))

#define rule_copy(parser, ___l) \
    (((___l) - parser(parser)->thirteen) & LINE_COPY)

#define parser(parser) \
    ((struct parser_type *) (parser))

static int deck_ever(const char *, unsigned, const void *,
		     const struct x1f4_variable_type **, void **);
static int deck_miss(struct parser_type *, void *, struct c1_node_type *,
		     const char *, unsigned);

static int
deck_ever(const char *text, unsigned miss, const void *context,
	  const struct x1f4_variable_type **variable, void **state)
{
    int status;

    /*
     * the expression is `C* = L;', no need to properly test for `L'.
     */
    if (*text == 'L' && true(miss == 1)) {
	struct x1f4_variable_type *variable_data;

	variable_data = ((void **) context)[1];

	variable_data->flags = 0;
	variable_data->length = 1;
	variable_data->name = "L";

	*variable = variable_data;

	status = 0;
    } else {
	void **line;

	line = (void *) context;

	status =
	    _x1f4_c1_select_variable(text, miss, line[0], variable, state);
	if (status) {
	} else {
	    struct x1f4_variable_type *variable_data;

	    variable_data = line[1];
	    variable_data->type = (*variable)->type;
	}
    }

    return status;
}


static int
deck_miss(struct parser_type *parser_data, void *fold,
	  struct c1_node_type *node_data, const char *c, unsigned w)
{
    char call[13];
    int status;
    struct x1f4_e4_type *e4_data;
    struct x1f4_variable_type L;
    void *expression;
    const void *line[2];

    e4_data = &parser_data->e4;

    e4_data->terminator = 0x3b;

    line[0] = e4_data->variable_set.context;
    line[1] = &L;

    memcpy(call, c, w);
    memcpy(call + w, " = L;", 6);

    e4_data->variable_set.context = line;
    e4_data->variable_set.get = deck_ever;

    status = x1f4_iNit_expression
	(&expression, call, parser_data->e4_flags, e4_data);
    if (status) {
	if (true(status == X1f4_E4_ALLOC_ERROR)) {
	    status = ALLOC_ERROR;
	}
    } else {
	node_data[0].data.data = expression;

	status = x1f4_rack_expression(expression, &L, fold);
	if (status) {
	    status = ALLOC_ERROR;

	    x1f4_fini_expression(&expression);
	} else {
	    x1f4_size_expression
		(expression, &parser_data->s1st, &parser_data->s2nd);
	}
    }

    e4_data->variable_set.context = line[0];
    e4_data->variable_set.get = _x1f4_c1_select_variable;

    return status;
}


int
_x1f4_c1_rule_fold(void *parser, void *fold, const char *g, const char *c,
		   unsigned w)
{
    int status;
    struct c1_node_type *node_data;

    status = M_LINK(parser, &node_data, sizeof(struct c1_node_type));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	status = deck_miss(parser, fold, node_data, c, w);
	if (status) {
	} else {
	    node_data[0].state = rule_copy(parser, g);

	    {
		struct c1_node_type *tail_node;

		tail_node = parser(parser)->tail_node;
		if (tail_node) {
		    tail_node->node_data = node_data;
		} else {
		    parser(parser)->head_node = node_data;
		}

		parser(parser)->tail_node = node_data;
	    }
	}

	if (status) {
	    M_FREE(parser, node_data);
	}
    }

    return status;
}

