/*
 * e4.7.c
 * Copyright (C) 2006-2014, 2017, 2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * TODO
 *
 *.write a non recursive version of _fast_hack so that nodes are not visited
 * one thousand times
 */

#include <config.h>

#include <e4-config.h>

#include <float.h>
#include <limits.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.0.h>
#include <e4-byte.h>
#include <e4-defs.h>
#include <e4-inter.h>
#include <e4-parse.h>
#include <e4-types.h>
#include <e4fine-e4.h>
#include <lcardinal-x.h>
#include <lxfset.h>
#if __EXPONENTIAL_REAL_SYNTAX__
# include <real.e.h>
#else
# include <real.h>
#endif				/* __EXPONENTIAL_REAL_SYNTAX__ */

#define __USE_XLCARDINAL__		1

/*
 * class set only definitions
 */
#define MAKE_CLASS			(ZERO_CLASS + 1)
#define ROLL_CLASS			(ZERO_CLASS + 2)
#define SIDE_CLASS			(ZERO_CLASS + 3)

#define ZONE_CLASS			(ZERO_CLASS + 3)

/*
 * expectation definitions
 */
#define DATA_CLASS			(ZONE_CLASS + 1)
#define DONE_CLASS			(ZONE_CLASS + 2)
#define JOIN_CLASS			(ZONE_CLASS + 3)
#define NEXT_CLASS			(ZONE_CLASS + 4)
#define POST_CLASS			(ZONE_CLASS + 5)

#define SET_NEXT__________ARGS_0 \
    struct next_type *, struct parser_type *, int *
#define SET_NEXT__________ARGS_1 \
    struct next_type *next_data, struct parser_type *parser_data, int *expect
#define SET_NEXT__________ARGS_2 \
    next_data, parser_data, expect

#define LOGIC_BITS \
    ((1 << LAST) | (1 << LOCK) | (1 << PICK))
#define SCALAR_BITS \
    ((1 << BILL) | (1 << MODE) | (1 << REAL))

#define EXPRESSION(expression) \
    ((e4_expression_type *) expression)

#define E_FREE(mcontext, mdata) \
    (EXPRESSION(mcontext)->m.free(EXPRESSION(mcontext)->m.data, (mdata)))
#define E_LINK(mcontext, mdata, size) \
    (EXPRESSION(mcontext)->m.link					      \
	(EXPRESSION(mcontext)->m.data, (void *) (mdata), (size)))
#define E_MODE(mcontext, mdata, size) \
    (EXPRESSION(mcontext)->m.mode					      \
	(EXPRESSION(mcontext)->m.data, (void *) (mdata), (size)))

#define M_FREE(mcontext, mdata) \
    (mcontext)->m.free((mcontext)->m.data, (mdata))
#define M_LINK(mcontext, mdata, size) \
    (mcontext)->m.link((mcontext)->m.data, (void *) (mdata), (size))
#define M_MODE(mcontext, mdata, size) \
    (mcontext)->m.mode((mcontext)->m.data, (void *) (mdata), (size))

extern const char _x1f4_e4_tc_reflex[];
extern const struct e4_odbx_type _x1f4_e4_pick[];
extern const unsigned _x1f4_c_type_0[], _x1f4_c_type_1[];

static int case_cc_0(GET_ATOM__________ARGS_0);
static int case_cc_1(GET_ATOM__________ARGS_0);
static int case_cc_2(GET_ATOM__________ARGS_0);
static int case_cc_3(GET_ATOM__________ARGS_0);
static int case_cc_4(GET_ATOM__________ARGS_0);
static int case_cc_5(GET_ATOM__________ARGS_0);
static int case_cc_6(GET_ATOM__________ARGS_0);
static int case_cc_7(GET_ATOM__________ARGS_0);
static int case_cc_8(GET_ATOM__________ARGS_0);
static int case_ce_8(GET_ATOM__________ARGS_0);
static int case_cc_9(GET_ATOM__________ARGS_0);
static int case_cc__(GET_ATOM__________ARGS_0);
static int case_side(struct e4_atom_type *, unsigned,
		     const struct e4_last_type *, struct parser_type *);
static int ever_last(struct next_type *, struct parser_type *, const char *,
		     unsigned, const struct e4_last_type *);
static int init_data(struct parser_type *);
static int join_hack(struct link_type **, struct e4_hack_type **,
		     struct parser_type *);
static int join_rail(struct parser_type *);
static int keep_ever(GET_ATOM__________ARGS_0, const char *, unsigned,
		     const struct e4_ever_type *, void *);
static int lead_atom(GET_ATOM__________ARGS_0);
static int lead_back(GET_ATOM__________ARGS_0);
static int lead_char(GET_ATOM__________ARGS_0);
static int lead_flat(GET_ATOM__________ARGS_0);
static int lead_name(GET_ATOM__________ARGS_0);
static int lead_odb1(GET_ATOM__________ARGS_0);
static int lead_odb2(GET_ATOM__________ARGS_0);
static int lead_swap(struct next_type *, struct parser_type *, int);
static int lead_text(GET_ATOM__________ARGS_0);
static int link_data(struct parser_type *);
static int make_atom(struct e4_atom_type **, struct parser_type *);
static int make_hack(struct e4_hack_type **, struct parser_type *);
static int make_list(struct e4_hack_type **, const struct e4_last_type *,
		     struct parser_type *, unsigned);
static int near_side(struct parser_type *);
static int part_open(struct parser_type *);
static int pick_atom(int *, int *, struct e4_atom_type **,
		     struct e4_atom_type *, struct e4_hack_type *,
		     struct parser_type *);
static int pick_this(int *, int *, struct e4_atom_type **,
		     struct e4_hack_type *, struct parser_type *);
static int push_hack(struct link_type **, struct e4_hack_type *,
		     struct parser_type *);
static int spin_dana(struct next_type *, struct parser_type *);
static int spin_ever(struct next_type *, struct parser_type *);
static int spin_last(struct next_type *, struct parser_type *);
static int trap_bill(SET_NEXT__________ARGS_0);
static int trap_case(struct next_type *, struct parser_type *,
		     int (*)(struct next_type *, struct parser_type *));
static int trap_dana(const struct e4_last_type *, const struct e4_ever_type *,
		     void *, struct parser_type *);
static int trap_ever(SET_NEXT__________ARGS_0);
static int trap_join(struct next_type *, const struct e4_last_type *,
		     struct parser_type *);
static int trap_last(SET_NEXT__________ARGS_0);
static int trap_lock(SET_NEXT__________ARGS_0);
static int trap_lose(SET_NEXT__________ARGS_0);
static int trap_make(SET_NEXT__________ARGS_0);
static int trap_mode(SET_NEXT__________ARGS_0);
static int trap_odb1(SET_NEXT__________ARGS_0);
static int trap_odb2(SET_NEXT__________ARGS_0);
static int trap_open(SET_NEXT__________ARGS_0);
static int trap_pick(struct next_type *, const struct e4_last_type *,
		     struct parser_type *);
static int trap_real(SET_NEXT__________ARGS_0);
static int trap_roll(SET_NEXT__________ARGS_0);
static int trap_side(SET_NEXT__________ARGS_0);
static int trap_step(SET_NEXT__________ARGS_0);
static int trap_text(SET_NEXT__________ARGS_0);

static void ever_slip(struct parser_type *);
static void ever_turn(struct parser_type *);
static void fill_data(struct parser_type *);
static void fill_type(unsigned char *, const unsigned *, unsigned);
static void free_long(struct parser_type *);
static void free_odb1(struct e4_odb1_type *, void *);
static void free_trip(struct parser_type *, struct trip_type *);

static int (*const case_seleCtion[])(GET_ATOM__________ARGS_0) = {
/* *INDENT-OFF* */
    case_cc_0,
    case_cc_1,
    case_cc_2,
    case_cc_3,
    case_cc_4,
    case_cc_5,
    case_cc_6,
    case_cc_7,
    case_ce_8,
    case_cc_9,
    case_cc__
/* *INDENT-ON* */
}, (*const case_selection[])(GET_ATOM__________ARGS_0) = {
/* *INDENT-OFF* */
    case_cc_0,
    case_cc_1,
    case_cc_2,
    case_cc_3,
    case_cc_4,
    case_cc_5,
    case_cc_6,
    case_cc_7,
    case_cc_8,
    case_cc_9,
    case_cc__
/* *INDENT-ON* */
}, (*const set_class[])(SET_NEXT__________ARGS_0) = {
/* *INDENT-OFF* */
    trap_bill,
    trap_odb1,
    trap_odb2,
    trap_ever,
    trap_last,
    trap_lock,
    trap_lose,
    trap_mode,
    trap_open,
    trap_real,
    trap_step,
    trap_text,
    trap_make,
    trap_roll,
    trap_side
/* *INDENT-ON* */
};
static const unsigned c_type_0[] = {
/*
 *
 *
 *   !   # $ % &       * +   -   /
 *                     :   < = > ?
 * @
 *                       [   ] ^
 *
 *                         |   ~
 *
 *
 *
 *
 *
 *
 *
 *
 */
/* *INDENT-OFF* */
    0x00000000, 0xf400ac7a, 0x68000001, 0x50000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
/* *INDENT-ON* */
}, c_type_1[] = {
/*
 *                   . .     .
 *
 * .                     +
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */
/* *INDENT-OFF* */
    0x00002600, 0x00000801, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
/* *INDENT-ON* */
}, c_type_2[] = {
/*
 *
 *
 *
 * 0 1 2 3 4 5 6 7 8 9
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */
/* *INDENT-OFF* */
    0x00000000, 0x03ff0000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
/* *INDENT-ON* */
}, c_type_3[] = {
#if __EXPONENTIAL_REAL_SYNTAX__
/*
 *
 *
 *                             .
 * 0 1 2 3 4 5 6 7 8 9
 *
 *
 *           e
 *     r
 *
 *
 *
 *
 *
 *
 *
 *
 */
/* *INDENT-OFF* */
    0x00000000, 0x03ff4000, 0x00000000, 0x00040020,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
/* *INDENT-ON* */
#else
/*
 *
 *
 *                             .
 * 0 1 2 3 4 5 6 7 8 9
 *
 *
 *
 *     r
 *
 *
 *
 *
 *
 *
 *
 *
 */
/* *INDENT-OFF* */
    0x00000000, 0x03ff4000, 0x00000000, 0x00040000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
/* *INDENT-ON* */
#endif				/* __EXPONENTIAL_REAL_SYNTAX__ */
}, c_type_4[] = {
/*
 *                   . .     .
 *
 * .
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */
/* *INDENT-OFF* */
    0x00002600, 0x00000001, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000
/* *INDENT-ON* */
};

static int
case_cc_0(GET_ATOM__________ARGS_1)
{
    return _x1f4_e4_080704_error_3(parser_data);
}


static int
case_cc_1(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_1				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_2				(ERROR_1 | (1 << LOSE_CLASS))
#define ERROR_3				(ERROR_2 | (1 << NEXT_CLASS))
#define ERROR_4				(ERROR_3 | (1 << OPEN_CLASS))
#define ERROR_5				(ERROR_4 | (1 << POST_CLASS))
#define ERROR_6				(ERROR_5 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_6) {
	status = _x1f4_e4_110608_error_1(parser_data);
    } else {
	status = lead_name(GET_ATOM__________ARGS_2);
    }

    return status;

#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_3
#undef ERROR_2
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc_2(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_1				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_3				(ERROR_1 | (1 << LOSE_CLASS))
#define ERROR_4				(ERROR_3 | (1 << NEXT_CLASS))
#define ERROR_5				(ERROR_4 | (1 << OPEN_CLASS))
#define ERROR_6				(ERROR_5 | (1 << POST_CLASS))
#define ERROR_7				(ERROR_6 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	status = _x1f4_e4_110608_error_2(parser_data);
    } else {
	status = lead_flat(GET_ATOM__________ARGS_2);
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_3
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc_3(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_1				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_3				(ERROR_1 | (1 << LOSE_CLASS))
#define ERROR_4				(ERROR_3 | (1 << NEXT_CLASS))
#define ERROR_5				(ERROR_4 | (1 << OPEN_CLASS))
#define ERROR_6				(ERROR_5 | (1 << POST_CLASS))
#define ERROR_7				(ERROR_6 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	status = _x1f4_e4_110608_error_3(parser_data);
    } else {
	status = lead_text(GET_ATOM__________ARGS_2);
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_3
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc_4(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DATA_CLASS))
#define ERROR_1				(ERROR_0 | (1 << DONE_CLASS))
#define ERROR_4				(ERROR_1 | (1 << NEXT_CLASS))
#define ERROR_5				(ERROR_4 | (1 << OPEN_CLASS))
#define ERROR_6				(ERROR_5 | (1 << POST_CLASS))
#define ERROR_7				(ERROR_6 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	struct link_type *link_data;

	link_data = parser_data->link_data;
	if (expect == POST_CLASS) {
	    status = _x1f4_e4_110605_error_1
		(parser_data,
		 link_data->last_data ? link_data->atom_data : NULL);
	} else {
	    do {
		if (link_data) {
		    const struct e4_last_type *last_data;

		    last_data = link_data->last_data;
		    if (last_data) {
			if (last_data->bits & SIDE_LIST) {
			    if (parser_data->hack_data->count) {
			    } else {
				if (parser_data->odb1_data) {
				} else {
				    if (last_data->count) {
				    } else {
					if (link_data->frame ^ 1) {
					} else {
					    parser_data->expression++;
					    next_data->class = LOSE_CLASS;
					    link_data->frame = 0;
					    status = 0;
					    if (1) {
						break;
					    }
					}
				    }
				}
			    }
			}
		    }
		}

		status = _x1f4_e4_110608_error_4(parser_data);
	    } while (0);
	}
    } else {
	struct link_type *link_data;

	link_data = parser_data->link_data;
	if (link_data->part) {
	    status = _x1f4_e4_110608_error_4(parser_data);
	} else {
	    parser_data->expression++;
	    next_data->class = LOSE_CLASS;
	    status = 0;
	}
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc_5(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DATA_CLASS))
#define ERROR_1				(ERROR_0 | (1 << DONE_CLASS))
#define ERROR_3				(ERROR_1 | (1 << JOIN_CLASS))
#define ERROR_5				(ERROR_3 | (1 << LOSE_CLASS))
#define ERROR_6				(ERROR_5 | (1 << NEXT_CLASS))
#define ERROR_7				(ERROR_6 | (1 << OPEN_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	if (expect == JOIN_CLASS) {
	    do {
		const struct e4_last_type *last_data;
		struct link_type *link_data;

		link_data = parser_data->link_data;

		last_data = link_data->last_data;
		if (last_data) {
		    if (last_data->bits & SIDE_LIST) {
			parser_data->expression++;
			next_data->class = SIDE_CLASS;
			status = 0;
			if (1) {
			    break;
			}
		    }
		}

		if (link_data->part) {
		    status = _x1f4_e4_170827_error_0(parser_data);
		} else {
		    status = _x1f4_e4_110605_error_0
			(parser_data,
			 last_data ? link_data->atom_data : NULL);
		}
	    } while (0);
	} else {
	    status = _x1f4_e4_110608_error_5(parser_data);
	}
    } else {
	parser_data->expression++;
	next_data->class = STEP_CLASS;
	status = 0;
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_3
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc_6(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_1				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_2				(ERROR_1 | (1 << LOSE_CLASS))
#define ERROR_3				(ERROR_2 | (1 << NEXT_CLASS))
#define ERROR_4				(ERROR_3 | (1 << POST_CLASS))
#define ERROR_5				(ERROR_4 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_5) {
	do {
	    if (parser_data->e4fine) {
		int miss = 0, type = VOID;
		struct e4_atom_type *atom_data, *atom_text = NULL;

		atom_data = parser_data->atom_data;
		type = atom_data->type;

		status = pick_atom
		    (&type, &miss, &atom_text, atom_data,
		     parser_data->hack_data, parser_data);
		if (status) {
		    break;
		}

		if (BASE_LOCK < type) {
		    const void *line;

		    if (atom_text) {
			const struct e4_ever_type *ever_data;

			status = 1;
			ever_data = atom_text->data.ever.ever_data;
			if (ever_data->bits & UNIVERSAL) {
			    status = x1f4_pick_e4fine
				(parser_data->e4fine, ever_data->type, &line);
			} else {
			    line = *(void **) atom_text->data.ever.state;
			    /*
			     * no need to check for the REFERENCE flag, as the
			     * state has been set to a function description
			     * pointer
			     */
			}
		    } else {
			status = x1f4_pick_e4fine
			    (parser_data->e4fine, type, &line);
		    }
		    if (status) {
			const struct e4_last_type *last_data;

			last_data = &((struct e4_line_type *) line)->last;
			if (miss) {
			    status = trap_dana
				(last_data, atom_text->data.ever.ever_data,
				 atom_text->data.ever.state, parser_data);
			} else {
			    status = trap_join
				(next_data, last_data, parser_data);
			}
			if (status) {
			} else {
			    parser_data->expression++;
			    next_data->class = OPEN_CLASS;
			}

			break;
		    }
		} else {
		    if (type == CASE && parser_data->recall) {
			status = trap_case(next_data, parser_data, spin_dana);
			if (status) {
			} else {
			    parser_data->expression++;
			    ever_turn(parser_data);
			    if (*parser_data->expression == ')') {
				parser_data->expression++;
				next_data->class = LOSE_CLASS;
			    } else {
				next_data->class = SIDE_CLASS;
			    }
			}

			break;
		    }
		}
	    }

	    status = _x1f4_e4_110608_error_6(parser_data);
	} while (0);
    } else {
	parser_data->expression++;
	next_data->class = OPEN_CLASS;
	status = 0;
    }

    return status;

#undef ERROR_5
#undef ERROR_4
#undef ERROR_3
#undef ERROR_2
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc_7(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_2				(ERROR_0 | (1 << LOSE_CLASS))
#define ERROR_3				(ERROR_2 | (1 << OPEN_CLASS))
#define ERROR_4				(ERROR_3 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_4) {
	status = _x1f4_e4_110608_error_7(parser_data);
    } else {
	if (expect == DATA_CLASS) {
	    status = lead_odb1(GET_ATOM__________ARGS_2);
	} else {
	    status = lead_odb2(GET_ATOM__________ARGS_2);
	}
    }

    return status;

#undef ERROR_4
#undef ERROR_3
#undef ERROR_2
#undef ERROR_0
}


static int
case_cc_8(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DATA_CLASS))
#define ERROR_2				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_4				(ERROR_2 | (1 << LOSE_CLASS))
#define ERROR_5				(ERROR_4 | (1 << OPEN_CLASS))
#define ERROR_6				(ERROR_5 | (1 << POST_CLASS))
#define ERROR_7				(ERROR_6 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	status = _x1f4_e4_080704_error_5(parser_data);
    } else {
	next_data->class = DONE_CLASS;
	status = 0;
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_2
#undef ERROR_0
}


static int
case_ce_8(GET_ATOM__________ARGS_1)
{
    return _x1f4_e4_080704_error_5(parser_data);
}


static int
case_cc_9(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_1				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_3				(ERROR_1 | (1 << LOSE_CLASS))
#define ERROR_4				(ERROR_3 | (1 << NEXT_CLASS))
#define ERROR_5				(ERROR_4 | (1 << OPEN_CLASS))
#define ERROR_6				(ERROR_5 | (1 << POST_CLASS))
#define ERROR_7				(ERROR_6 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	status = _x1f4_e4_110608_error_9(parser_data);
    } else {
	status = lead_char(GET_ATOM__________ARGS_2);
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_3
#undef ERROR_1
#undef ERROR_0
}


static int
case_cc__(GET_ATOM__________ARGS_1)
{
#define ERROR_0				((1 << DONE_CLASS))
#define ERROR_1				(ERROR_0 | (1 << JOIN_CLASS))
#define ERROR_3				(ERROR_1 | (1 << LOSE_CLASS))
#define ERROR_4				(ERROR_3 | (1 << NEXT_CLASS))
#define ERROR_5				(ERROR_4 | (1 << OPEN_CLASS))
#define ERROR_6				(ERROR_5 | (1 << POST_CLASS))
#define ERROR_7				(ERROR_6 | (1 << STEP_CLASS))

    int status;

    if ((1 << expect) & ERROR_7) {
	do {
#define STRIP_0				((1 << JOIN_CLASS))
#define STRIP_1				(STRIP_0 | (1 << NEXT_CLASS))
#define STRIP_2				(STRIP_1 | (1 << POST_CLASS))

	    if ((1 << expect) & STRIP_2) {
		const char *f;
		int c;

		f = parser_data->expression;

		parser_data->expression = f + 1;

		ever_slip(parser_data);

		c = *(const unsigned char *) parser_data->expression;

		if (C_TYPE_X(_x1f4_c_type_1, c)) {
		    int miss = 0, type = VOID;
		    struct e4_atom_type *atom_data, *atom_text = NULL;

		    atom_data = parser_data->atom_data;
		    type = atom_data->type;

		    status = pick_atom
			(&type, &miss, &atom_text, atom_data,
			 parser_data->hack_data, parser_data);
		    if (status) {
			break;
		    }

		    status = lead_swap(next_data, parser_data, type);
		    if (status) {
			if (status == STILL_ERROR) {
			    parser_data->expression = f;
			} else {
			    break;
			}
		    } else {
			break;
		    }
		}
	    }

	    status = _x1f4_e4_170926_error_0(parser_data);

#undef STRIP_2
#undef STRIP_1
#undef STRIP_0
	} while (0);
    } else {
	/*
	 * what's after this (starting `.') cannot be an integral
	 */
	status = lead_flat(GET_ATOM__________ARGS_2);
    }

    return status;

#undef ERROR_7
#undef ERROR_6
#undef ERROR_5
#undef ERROR_4
#undef ERROR_3
#undef ERROR_1
#undef ERROR_0
}


static int
case_side(struct e4_atom_type *atom_data, unsigned frame,
	  const struct e4_last_type *last_data,
	  struct parser_type *parser_data)
{
    int status;
    struct e4_flat_type *flat_data;
    unsigned bits, class, count;

    count = last_data->count;
    bits = last_data->bits;

    if (frame == count && !(bits & SLIP_LIST)) {
	frame = 0;
	class = 0;
    } else {
	class = frame;
	if (bits & (LINK_PASS | POST_TYPE)) {
	    class <<= 1;
	    bits |= POST_TYPE;
	}
    }

    status = M_LINK
	(parser_data, &flat_data,
	 sizeof(struct e4_flat_type) * 3 + sizeof(struct e4_last_type)
	 + class * sizeof(int));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	struct e4_base_type *base_data;
	struct e4_last_type *last_text;

	flat_data->flat_data = parser_data->flat_data;
	flat_data->read = DATA_FLAT;

	parser_data->flat_data = flat_data;

	base_data = hack_base(atom_data->data.last.base);

	flat_data[1].flat_data = (void *) last_data;
	flat_data[2].flat_data = base_data->text;

	last_text = (void *) (flat_data + 3);

	memcpy(last_text, last_data, sizeof(struct e4_last_type));

	if (frame) {
	    int *args;
	    const int *line;

	    args = (void *) (last_text + 1);

	    last_text->count = frame;
	    last_text->args = args;

	    line = last_data->args;

	    memcpy(args, line, count * sizeof(int));

	    args += count;

	    frame -= count;
	    if (bits & POST_TYPE) {
		int *flat;

		flat = args + frame;

		if (last_data->bits & POST_TYPE) {
		    line += count;
		    for (; count; count--) {
			*flat++ = *line++;
		    }
		} else {
		    for (; count; count--) {
			*flat++ = 0;
		    }
		}

		for (; frame; frame--) {
		    *args++ = SIDE;
		    *flat++ = 0;
		}
	    } else {
		for (; frame; frame--) {
		    *args++ = SIDE;
		}
	    }
	}

	base_data->last_data = last_text;

	last_text->bits = (bits | FLAT_LINE) & ~(SIDE_LIST | SLIP_LIST);
    }

    return status;
}


static int
ever_last(struct next_type *next_data, struct parser_type *parser_data,
	  const char *f, unsigned length, const struct e4_last_type *last_data)
{
    int status;
    struct e4_flat_type *flat_data;

    status = M_LINK(parser_data, &flat_data, __news_alloc_size__);
    if (status) {
	status = ALLOC_ERROR;
    } else {
	struct e4_ever_type *ever_miss;
	struct e4_news_type *news_data;
	void **record, *state;

	flat_data->flat_data = parser_data->flat_data;
	flat_data->read = NEWS_FLAT;

	parser_data->flat_data = flat_data;

	ever_miss = (void *) (flat_data + 1);

	news_data = (void *) (ever_miss + 1);

	state = news_data + 1;

	record = state;

	news_data->call = 1;

	news_data->line.last = *last_data;

	news_data->data = parser_data->m.data;
	news_data->free = parser_data->m.free;

	news_data->line.last.bits |= NEWS_LINK;

	if (last_data->bits & NODE_LINK) {
	    *record = (void *)
		((char *) last_data - offsetof(struct e4_line_type, last));
	} else {
	    *record = (void *) &news_data->line;
	}

	ever_miss->name = last_data->name;
	ever_miss->bits = FLAT_LINK | READ_ONLY;
	ever_miss->size = last_data->size;

	status = x1f4_type_e4fine
	    (parser_data->e4fine, &ever_miss->type, &news_data->line);
	if (status) {
	    status = ALLOC_ERROR;
	} else {
	    next_data->class = EVER_CLASS;
	    next_data->data.ever.state = state;
	    next_data->data.ever.ever_data = ever_miss;

	    if (0) {
	    } else {
		if (last_data->bits & TEXT_LINK) {
		    void *text;

		    status = parser_data->context_data->last_0.fix
			(parser_data->context_data->last_0.context,
			 last_data, &text);
		    if (status) {
		    } else {
			news_data->line.text = text;
		    }
		}
	    }
	}
    }

    return status;
}


static int
init_data(struct parser_type *parser_data)
{
    int expect = DATA_CLASS, status;
    struct e4_expression_type *expression_data;
    struct next_type next;

    parser_data->argument = 0;
    parser_data->flat_data = NULL;
    parser_data->link_class = NEXT_CLASS;
    parser_data->link_data = NULL;
    parser_data->odb1_data = NULL;
    parser_data->rail_link = NULL;
    parser_data->slip_rail = NULL;
    parser_data->trip_data = NULL;
    parser_data->trail_closing = 0;

    next.class = ~0;

    while (1) {
	status = parser_data->slip(&next, parser_data, expect);
	if (status) {
	    break;
	}
	if (next.class == DONE_CLASS) {
	    if (parser_data->link_data) {
		/*
		 * unreached?
		 */
		status = _x1f4_e4_080704_error_a(parser_data);
	    }

	    break;
	}

	status = set_class[next.class](&next, parser_data, &expect);
	if (status) {
	    break;
	}
    }

    if (status) {
    } else {
	status = join_rail(parser_data);

	parser_data->rail_link = NULL;
    }

    if (status) {
	free_long(parser_data);
    }

    free_trip(parser_data, parser_data->rail_link);

    expression_data = parser_data->expression_data;

    free_odb1(parser_data->odb1_data, expression_data);

    expression_data->flat_data = parser_data->flat_data;

    if (status) {
    } else {
	if (parser_data->bits & PACK_ESTORAGE) {
	    struct e4_hack_type *hack_data;

	    hack_data = &expression_data->hack;
	    M_MODE(parser_data, &hack_data->atom_data,
		   hack_data->count * sizeof(struct e4_atom_type));
	}

	status = link_data(parser_data);
	if (status) {
	} else {
	    _x1f4_e4_sort_expression(expression_data);
	}
    }

    free_trip(parser_data, parser_data->trip_data);

    return status;
}


static int
join_hack(struct link_type **link, struct e4_hack_type **hack,
	  struct parser_type *parser_data)
{
    int status;
    struct e4_hack_type *hack_slip;
    struct link_type *link_data, *link_slip;

    link_data = *link;

    *hack = hack_slip = link_data->data;
    *link = link_slip = link_data->link_data;

    status = join_rail(parser_data);

    parser_data->rail_link = link_data->trip_data;

    parser_data->hack_data = hack_slip;
    parser_data->link_data = link_slip;

    M_FREE(parser_data, link_data);

    return status;
}


static int
join_rail(struct parser_type *parser_data)
{
    int status = 0;
    struct trip_type *trip_data;

    trip_data = parser_data->rail_link;
    while (trip_data) {
	void *trip;

	if (trip_data->left.bits & RAIL_ONLY) {
	    if (status) {
	    } else {
		status = _x1f4_e4_170621_error_0(parser_data, trip_data);
	    }
	}

	trip = trip_data->trip_data;

	trip_data->trip_data = parser_data->trip_data;
	parser_data->trip_data = trip_data;

	trip_data = trip;
    }

    return status;
}


static int
keep_ever(GET_ATOM__________ARGS_1, const char *f, unsigned length,
	  const struct e4_ever_type *ever_data, void *state)
{
    next_data->class = EVER_CLASS;
    next_data->data.ever.state = state;
    next_data->data.ever.ever_data = ever_data;

    if (0) {
    } else {
	unsigned bits;

	bits = ever_data->bits;
	if (bits & ANCILLARY) {
	    unsigned pass;

	    pass = bits & PRINCIPAL;

	    bits &= ~ANCILLARY;
	    bits |= PRINCIPAL;

	    if (pass) {
	    } else {
		bits |= SUSPENDED;
	    }

	    ((struct e4_ever_type *) ever_data)->bits = bits;
	} else {
	    if (bits & SUSPENDED) {
		((struct e4_ever_type *) ever_data)->bits = bits & ~SUSPENDED;
	    }
	}
    }

    return 0;
}


static int
lead_atom(GET_ATOM__________ARGS_1)
{
    int c;

    ever_turn(parser_data);

    c = *(const unsigned char *) parser_data->expression;

    return case_selection[parser_data->class_select[c]]
	(GET_ATOM__________ARGS_2);
}


static int
lead_back(GET_ATOM__________ARGS_1)
{
    int c, status;

    ever_turn(parser_data);

    c = *(const unsigned char *) parser_data->expression;

    do {
	if (c == 0x29) {
	    status = case_cc_8(GET_ATOM__________ARGS_2);
	    if (status) {
	    } else {
		if (parser_data->link_data) {
		} else {
		    break;
		}
	    }
	}

	status = case_seleCtion[parser_data->class_select[c]]
	    (GET_ATOM__________ARGS_2);
    } while (0);

    return status;
}


static int
lead_char(GET_ATOM__________ARGS_1)
{
    const char *e;
    int status = 1;

    e = parser_data->expression;

    e++;

    do {
	int c, f;

	c = *(const unsigned char *) e;
	if (c == '\'') {
	    break;
	} else {
	    if (!c) {
		break;
	    } else {
		if (c != '\\') {
		    f = c;
		} else {
		    e++;

		    c = *(const unsigned char *) e;
		    if (!c) {
			break;
		    } else {
			if (c < '0' || '7' < c) {
			    f = _x1f4_e4_tc_reflex[c];
			} else {
			    f = c - '0';
			    c = ((const unsigned char *) e)[1];
			    if (c < '0' || '7' < c) {
			    } else {
				f <<= 3;
				f |= c - '0';
				e++;
				c = ((const unsigned char *) e)[1];
				if (c < '0' || '7' < c) {
				} else {
				    f <<= 3;
				    f |= c - '0';
				    e++;
				}
			    }
			}
		    }
		}
	    }
	}

	e++;

	c = *(const unsigned char *) e;
	if (c != '\'') {
	    break;
	} else {
	    e++;

	    next_data->class = MODE_CLASS;
	    next_data->data.mode = f;
	}

	status = 0;
    } while (0);

    if (status) {
	status = _x1f4_e4_081019_error_0(parser_data);
    }

    parser_data->expression = e;

    return status;
}


static int
lead_flat(GET_ATOM__________ARGS_1)
{
    const char *e;
    int c, sign = 0, status = 0;

    e = parser_data->expression;

    /*
     * reached only for digits and period, the sign parsing below is
     * superfluous.
     */

    c = *(const unsigned char *) e;
    do {
	if (c == '-') {
	    sign ^= 1;
	} else {
	    if (C_TYPE_X(c_type_1, c)) {
	    } else {
		if (parser_data->bits & SKIP_SCOMMENT) {
		    if (c ^ 35) {
			break;
		    } else {
			do {
			    ++e;
			    c = *(const unsigned char *) e;
			} while (c ^ 10);
		    }
		} else {
		    break;
		}
	    }
	}
	e++;
	c = *(const unsigned char *) e;
    } while (1);

    if (!C_TYPE_X(c_type_2, c) && c != '.') {
	/*
	 * unreached?
	 */
	status = _x1f4_e4_080704_error_e(parser_data);
    } else {
#if __USE_XLCARDINAL__
	int o;
#endif				/* __USE_XLCARDINAL__ */
	unsigned long integral;

#if !__USE_XLCARDINAL__
# error no longer working

	integral = 0;

	do {
	    c -= '0';
	    integral *= 10;
	    integral += c;
	    e++;
	    c = *(const unsigned char *) e;
	} while (C_TYPE_X(c_type_2, c));
#else
	if (c != '.') {
	    o = *(const unsigned char *) (e + 1);
	    if (C_TYPE_X(c_type_3, o)) {
		const char *v;

		x1f4_parse_xlcardinal(&integral, e, &v, 0);
		c = *v;
		if (C_TYPE_X(c_type_3, c)) {
		    if (C_TYPE_X(c_type_2, c)) {
			/*
			 * hmm, allowing bad octals as decimals...
			 */
			x1f4_parse_xlcardinal(&integral, e, &v, 10);
			c = *v;
			if (C_TYPE_X(c_type_3, c)) {
			} else {
			    e = v;
			}
		    }
		    if (e != v) {
			/*
			 * allow no dangling `.'s or `e's to change the type
			 * for _real_.
			 */
			if (c == 'r') {
			    c = *(const unsigned char *) e;
			} else {
			    o = *(const unsigned char *) (v + 1);
			    if (c == 'e' && (o == '-' || o == '+')) {
				o = *(const unsigned char *) (v + 2);
			    }
			    if (C_TYPE_X(c_type_2, o)) {
				c = *(const unsigned char *) e;
			    } else {
				c = 0;
				e = v;
			    }
			}
		    }
		} else {
		    e = v;
		}
	    } else {
		const char *v;

		v = e;
		x1f4_parse_xlcardinal(&integral, e, &e, 0);
		c = *e;
		if (e != v + 1) {
		    if (c ^ 'i') {
			if (c ^ 'z') {
			    c = 0;
			    if (e != v) {
			    } else {
				e++;
				integral = 0;
			    }
			}
		    }
		}
	    }
	}
#endif				/* !__USE_XLCARDINAL__ */
#if __EXPONENTIAL_REAL_SYNTAX__
	if (c != 'e' && c != '.' && !C_TYPE_X(c_type_2, c)) {
#elif 0
	}
#else
	if (c != '.' && !C_TYPE_X(c_type_2, c)) {
#endif				/* __EXPONENTIAL_REAL_SYNTAX__ */
	    if (c == 'z') {
		if (sign) {
		    status = _x1f4_e4_080704_error_4(parser_data);
		} else {
		    e++;
		    next_data->class = BILL_CLASS;
		    next_data->data.bill = integral;
		}
	    } else {
		if (c == 'i') {
		    e++;
		}
		next_data->class = MODE_CLASS;
#if __USE_XLCARDINAL__
		if (sign) {
		    long i;

# if LONG_MAX + LONG_MIN < 0
		    if (LONG_MAX < integral) {
			long d = LONG_MAX;

			i = 0;
			do {
			    if (integral < (unsigned long) d) {
				i -= integral;
				break;
			    } else {
				i -= d;
				integral -= d;
				d = LONG_MIN - i;
				if (!d) {
				    break;
				} else {
				    if (-LONG_MAX < d) {
					d = -d;
				    } else {
					d = LONG_MAX;
				    }
				}
			    }
			} while (1);
		    } else {
			i = -integral;
		    }
# else
		    if ((unsigned long) -LONG_MIN < integral) {
			i = LONG_MIN;
		    } else {
			i = -integral;
		    }
# endif				/* LONG_MAX + LONG_MIN < 0 */

		    next_data->data.mode = i;
		} else {
		    if (LONG_MAX < integral) {
			integral = LONG_MAX;
		    }

		    next_data->data.mode = integral;
		}
#else
		next_data->data.mode = integral;
		if (sign) {
		    next_data->data.mode = -next_data->data.mode;
		}
#endif				/* __USE_XLCARDINAL__ */
	    }
	} else {
	    const char *v;
	    double midnight;

#if __EXPONENTIAL_REAL_SYNTAX__
	    x1f4_parse_ereal(&midnight, e, &v);
#else
	    x1f4_parse_real(&midnight, e, &v);
#endif				/* __EXPONENTIAL_REAL_SYNTAX__ */
	    if (v == e) {
		status = _x1f4_e4_080704_error_d(parser_data);
	    } else {
		e = v;

		if (*e == 'r') {
		    e++;
		}

		next_data->class = REAL_CLASS;
		next_data->data.real = midnight;
		if (sign) {
		    next_data->data.real = -next_data->data.real;
		}
	    }
	}
    }

    if (status) {
    } else {
	parser_data->expression = e;
    }

    return status;
}


static int
lead_name(GET_ATOM__________ARGS_1)
{
    const char *e, *f;
    int c, status;
    const struct e4_context_type *context_data;
    unsigned length;

    context_data = parser_data->context_data;

    e = parser_data->expression;

    f = e;
    do {
	e++;
	c = *(const unsigned char *) e;
    } while (C_TYPE_X(_x1f4_c_type_0, c));
    length = e - parser_data->expression;
    parser_data->expression = e;
    ever_slip(parser_data);
    e = parser_data->expression;
    c = *(const unsigned char *) e;
    if (c == '(') {
	const struct e4_ever_type *ever_data;
	const struct e4_last_type *last_data;
	void *state;

	status = context_data->ever_0.get
	    (f, length, context_data->ever_0.context, &ever_data,
	     &state);
	if (status) {
	    status = context_data->last_0.get
		(f, length, context_data->last_0.context, &last_data);
	    if (status) {
		status = _x1f4_e4_080704_error_0(parser_data, f, length);
	    } else {
		if (last_data->bits & (SIDE_LIST | SLIP_LIST)) {
		    if (parser_data->e4fine) {
			status = ever_last
			    (next_data, parser_data, f, length, last_data);
		    } else {
			status = _x1f4_e4_080704_error_1
			    (parser_data, f, length);
		    }
		} else {
		    next_data->class = LAST_CLASS;
		    next_data->data.last.last_data = last_data;

		    if (0) {
		    } else {
			if (last_data->bits & TEXT_LINK) {
			    status = context_data->last_0.fix
				(context_data->last_0.context,
				 last_data, &next_data->data.last.text);
			}
		    }
		}
	    }
	} else {
	    do {
		if (ever_data->type < BASE_LOCK) {
#if 0
		    status = _x1f4_e4_110817_error_0(parser_data, f, length);
#else
		    status = keep_ever
			(GET_ATOM__________ARGS_2, f, length, ever_data,
			 state);
#endif				/* 0 */
		    if (1) {
			break;
		    }
		}

		if (ever_data->bits & UNIVERSAL) {
		    if (parser_data->e4fine) {
			const void *line;

			status = x1f4_pick_e4fine
			    (parser_data->e4fine, ever_data->type, &line);
			if (status) {
			    status = 0;
			    if (1) {
				last_data =
				    &((struct e4_line_type *) line)->last;
			    }
			} else {
			    /*
			     * TODO
			     * fix, not the best error report
			     */
			    status = _x1f4_e4_080704_error_0
				(parser_data, f, length);
			    if (1) {
				break;
			    }
			}
		    } else {
			/*
			 * TODO
			 * fix, not the best error report
			 */
			status = _x1f4_e4_080704_error_0
			    (parser_data, f, length);
			if (1) {
			    break;
			}
		    }
		} else {
		    if (ever_data->bits & REFERENCE) {
			void *last;

			/*
			 * no need to check for the REFERENCE flag, as the
			 * state has been set to a function description
			 * pointer
			 */

			last = *(void **) state;

			last_data = &((struct e4_line_type *) last)->last;
		    } else {
			last_data =
			    &((struct e4_line_type *) *(void **) state)->last;
		    }
		}

		next_data->class = LOCK_CLASS;
		next_data->data.lock.ever_data = ever_data;
		next_data->data.lock.last_data = last_data;
		next_data->data.lock.state = state;

		if (0) {
		} else {
		    unsigned bits;

		    bits = ever_data->bits;
		    if (bits & ANCILLARY) {
			bits &= ~ANCILLARY;
			bits |= PRINCIPAL;

			((struct e4_ever_type *) ever_data)->bits = bits;
		    }
		}
	    } while (0);
	}
    } else {
	const struct e4_ever_type *ever_data;
	void *state;

	status = context_data->ever_0.get
	    (f, length, context_data->ever_0.context, &ever_data, &state);
	if (status) {
	    if (parser_data->e4fine) {
		const struct e4_last_type *last_data;

		status = context_data->last_0.get
		    (f, length, context_data->last_0.context, &last_data);
		if (status) {
		    status = _x1f4_e4_080704_error_1(parser_data, f, length);
		} else {
		    status = ever_last
			(next_data, parser_data, f, length, last_data);
		}
	    } else {
		status = _x1f4_e4_080704_error_1(parser_data, f, length);
	    }
	} else {
	    status = keep_ever
		(GET_ATOM__________ARGS_2, f, length, ever_data, state);
	}
    }

    if (status) {
    } else {
	parser_data->expression = e;
    }

    return status;
}


static int
lead_odb1(GET_ATOM__________ARGS_1)
{
    const char *e, *f;
    int c, status = 0;
    unsigned *e_type_0, length;

    e_type_0 = parser_data->c_type_0;

    e = parser_data->expression;

    f = e;
    do {
	e++;
	c = *(const unsigned char *) e;
    } while (C_TYPE_X(e_type_0, c));
    length = e - f;
    f = e;
    if (1) {
	struct e4_odb1_type *odb1_data;

	status = M_LINK
	    (parser_data, &odb1_data,
	     sizeof(struct e4_odb1_type) + sizeof(struct text_type));
	if (status) {
	    status = ALLOC_ERROR;
	} else {
	    struct text_type *text_data;

	    odb1_data->odb1_data = NULL;

	    if (next_data->class != CAD1_CLASS) {
		next_data->class = CAD1_CLASS;
	    } else {
		odb1_data->odb1_data = next_data->data.odb1_data;
	    }

	    next_data->data.odb1_data = odb1_data;

	    text_data = (struct text_type *) (odb1_data + 1);

	    text_data->data = f;
	    text_data->size = length;

	    parser_data->expression = f;
	}
    }

    return status;
}


/*
 * function fails to live up to the parser greedy promise when a post mark is
 * recorded by the lower expression level (link_data->part is set).
 */
static int
lead_odb2(GET_ATOM__________ARGS_1)
{
    const char *e, *f;
    int c, status = 0;
    struct trip_type *slip_rail = NULL;
    unsigned *e_type_0, length;

    e_type_0 = parser_data->c_type_0;

    e = parser_data->expression;

    f = e;
    do {
	e++;
	c = *(const unsigned char *) e;
    } while (C_TYPE_X(e_type_0, c));

    length = e - parser_data->expression;

    do {
	const struct e4_odbx_type *odbx_rail = NULL, *odbx_text = NULL;
	struct link_type *link_data;
	struct trip_type *trip_data, *trip_rail = NULL;
	unsigned toxic = 0;

	if (parser_data->context_data->odb2_data) {
	} else {
	    status = _x1f4_e4_080704_error_h(parser_data);
	    if (1) {
		break;
	    }
	}

	link_data = parser_data->link_data;
	if (link_data && !parser_data->trail_closing) {
	    const char *part;

	    part = link_data->part;
	    if (part) {
		unsigned land;

		land = link_data->land;
		if (length < land) {
		} else {
		    if (memcmp((void *) f, part, land)) {
		    } else {
			next_data->class = LOSE_CLASS;

			length = land;

			trip_data = parser_data->rail_link;
			if (parser_data->slip_rail == trip_data && trip_data) {
			    if (trip_data->left.bits & RAIL_ONLY) {
				status = _x1f4_e4_170621_error_0
				    (parser_data, trip_data);
				if (1) {
				    break;
				}
			    }

			    parser_data->rail_link = trip_data->trip_data;

			    trip_data->trip_data = parser_data->trip_data;
			    parser_data->trip_data = trip_data;
			}

			trip_data = link_data->trip_data;
			if (trip_data) {
			    if (&trip_data->lead == link_data->odbx_data) {
				if (((struct e4_2nde_type *)
				     link_data->odbx_data->extension2)->bits
				    & POST_MARK) {
				    slip_rail = trip_data;
				} else {
				    parser_data->trail_closing = 1;
				    length = 0;
				}
			    }
			}

			break;
		    }
		}
	    }
	}

	trip_data = parser_data->rail_link;
	while (trip_data) {
	    const struct e4_odbx_type *odbx_case = NULL;

	    _x1f4_e4_pick_odta
		(&odbx_case, &odbx_rail, toxic, f, length,
		 trip_data->lead.name, trip_data->lead.size, parser_data);
	    if (odbx_case) {
		trip_rail = trip_data;
		toxic = odbx_rail->size;

		trip_data->left =
		    *(const struct e4_2nde_type *) odbx_case->extension2;
		trip_data->lead = *odbx_case;
		trip_data->lead.extension2 = &trip_data->left;
	    }

	    if (parser_data->trail_closing) {
		/*
		 * not to look for better, there's only one match and it is the
		 * first.
		 */

		break;
	    }

	    if (trip_data->left.bits & RAIL_ONLY) {
		break;
	    }

	    trip_data = trip_data->trip_data;
	}

	if (parser_data->trail_closing) {
	    parser_data->trail_closing = 0;
	} else {
	    _x1f4_e4_pick_odaa(&odbx_text, toxic, f, length, parser_data);

	    if (odbx_text) {
		toxic = odbx_text->size;
		odbx_rail = NULL;
	    } else {
	    }
	}

	if (odbx_rail) {
	    struct trip_type *trip_lead;

	    trip_lead = parser_data->rail_link;

	    parser_data->rail_link = trip_rail->trip_data;

	    trip_rail->trip_data = parser_data->trip_data;
	    parser_data->trip_data = trip_lead;

	    trip_rail->lead.bits |= LEAD_TRIP | RAIL_TRIP;
	    trip_rail->left.bits &= ~CALL_LINK;
	    trip_rail->rail = *odbx_rail;
	    trip_rail->post = *(struct e4_2nde_type *) odbx_rail->extension2;
	    trip_rail->post.bits |= BACK_RAIL;
	    trip_rail->post.rail = &trip_rail->lead;
	    trip_rail->rail.bits |= RAIL_TRIP;
	    trip_rail->rail.extension2 = &trip_rail->post;

	    odbx_text = &trip_rail->rail;
	}

	trip_data = parser_data->rail_link;
	if (parser_data->slip_rail == trip_data && trip_data) {
	    if (trip_data->left.bits & RAIL_ONLY) {
		status = _x1f4_e4_170621_error_0(parser_data, trip_data);
		if (1) {
		    break;
		}
	    }

	    parser_data->rail_link = trip_data->trip_data;

	    trip_data->trip_data = parser_data->trip_data;
	    parser_data->trip_data = trip_data;
	}

	if (odbx_text) {
	    length = toxic;

	    if (odbx_text->bits & E2ND_LINK) {
		const struct e4_2nde_type *e2nd_data;

		e2nd_data = odbx_text->extension2;
		if (e2nd_data->bits & (RAIL_LIST | RAIL_MARK)) {
		    struct trip_type *trip_data;

		    status = M_LINK(parser_data, &trip_data,
				    sizeof(struct trip_type));
		    if (status) {
			status = ALLOC_ERROR;
		    } else {
			trip_data->left = *e2nd_data;
			trip_data->lead = *odbx_text;
			trip_data->lead.extension2 = &trip_data->left;

			trip_data->trip_data = parser_data->rail_link;
			parser_data->rail_link = trip_data;

			odbx_text = &trip_data->lead;
		    }
		}
	    }

	    next_data->class = CAD2_CLASS;
	    next_data->data.odbx.odbx_data = odbx_text;
	} else {
	    status = _x1f4_e4_080704_error_j(parser_data, f, length);
	}
    } while (0);

    if (status) {
    } else {
	parser_data->expression = f + length;

	parser_data->slip_rail = slip_rail;
    }

    return status;
}


static int
lead_swap(struct next_type *next_data, struct parser_type *parser_data,
	  int type)
{
    const char *e, *f;
    int c, status;
    const struct e4_context_type *context_data;
    unsigned length;

    context_data = parser_data->context_data;

    e = parser_data->expression;

    f = e;
    do {
	e++;
	c = *(const unsigned char *) e;
    } while (C_TYPE_X(_x1f4_c_type_0, c));
    length = e - parser_data->expression;
    parser_data->expression = e;
    ever_slip(parser_data);
    e = parser_data->expression;
    c = *(const unsigned char *) e;
#if 0
    if (c == '(') {
#elif 0
    }
#else
    if (1) {
#endif				/* 0 */
	const struct e4_last_type *last_data;
	void *fine;

	fine = parser_data->prefix;
	if (fine) {
	    void *post, *record;

	    post = (void *) (integral_q) type;

	    status = x1f4_find_lxfset(fine, (void *) &post, &record);
	    if (status) {
		struct e4_bind_type *bind_data;

		bind_data = ((void **) record)[1];
		if (bind_data) {
		    unsigned prefix;

		    prefix = bind_data->size;
		    while (prefix) {
			if (prefix + length < 64) {
			    char name[64];

			    memcpy(name, bind_data->name, prefix);
			    memcpy(name + prefix, f, length);

			    status = context_data->last_0.get
				(name, prefix + length,
				 context_data->last_0.context, &last_data);
			} else {
			    char *name;

			    name = malloc(prefix + length);
			    if (name) {
				memcpy(name, bind_data->name, prefix);
				memcpy(name + prefix, f, length);

				status = context_data->last_0.get
				    (name, prefix + length,
				     context_data->last_0.context, &last_data);

				free(name);
			    }
			}
			if (status) {
			} else {
			    break;
			}

			bind_data++;
			prefix = bind_data->size;
		    }
		}
	    } else {
		status = 1;
	    }
	} else {
	    status = 1;
	}

	if (status) {
	    status = context_data->last_0.get
		(f, length, context_data->last_0.context, &last_data);
	}
	if (status) {
	    status = _x1f4_e4_080704_error_0(parser_data, f, length);
	} else {
	    if (last_data->bits & (SIDE_LIST | SLIP_LIST)) {
		if ((last_data->count < 2
		     && (last_data->count || last_data->bits & SIDE_LIST))
		    || c == '(') {
		    if (parser_data->e4fine) {
			status = ever_last
			    (next_data, parser_data, f, length, last_data);
			if (status) {
			} else {
			    next_data->class = ROLL_CLASS;
			}
		    } else {
			status = _x1f4_e4_170926_error_0(parser_data);
		    }
		} else {
		    status = _x1f4_e4_170926_error_0(parser_data);
		}
	    } else {
		if (last_data->count) {
		    next_data->class = MAKE_CLASS;
		    next_data->data.last.last_data = last_data;

		    if (0) {
		    } else {
			if (last_data->bits & TEXT_LINK) {
			    status = context_data->last_0.fix
				(context_data->last_0.context, last_data,
				 &next_data->data.last.text);
			}
		    }
		} else {
		    status = _x1f4_e4_170926_error_1(parser_data);
		}
	    }
	}
    } else {
	status = STILL_ERROR;
    }

    return status;
}


static int
lead_text(GET_ATOM__________ARGS_1)
{
    const char *e, *f;
    int status = 0;
    unsigned count = 0, length = 0;

    e = parser_data->expression;

    e++;

    f = e;

    while (1) {
	int c;

	status = x1f4_size_tcstring(f, &f, &length);
	if (status) {
	    break;
	} else {
	    count++;

	    f++;
	    parser_data->expression = f;
	    ever_slip(parser_data);
	    f = parser_data->expression;
	    c = *(const unsigned char *) f;

	    if (c ^ '"') {
		break;
	    } else {
		f++;
	    }
	}
    }

    if (status) {
	status = _x1f4_e4_080704_error_f(parser_data);
    } else {
	char *string;

	status = M_LINK(parser_data, &string, length + 1);
	if (status) {
	    status = ALLOC_ERROR;
	} else {
	    next_data->class = TEXT_CLASS;
	    next_data->data.text = string;

	    parser_data->expression = e;

	    count--;
	    for (; count;) {
		x1f4_copy_tcstring(&string, &parser_data->expression);
		count--;
		ever_slip(parser_data);
		parser_data->expression++;
	    }
	    {
		x1f4_copy_tcstring(&string, &parser_data->expression);
	    }
	}
    }

    return status;
}


static int
link_data(struct parser_type *parser_data)
{
    int status;
    struct e4_expression_type *expression_data;

    expression_data = parser_data->expression_data;
    status = _x1f4_e4_link_hack
	(&expression_data->hack, &expression_data->type, parser_data);
    if (status) {
    } else {
	unsigned max_list_length, max_post_length;

	_x1f4_e4_fast_miss(&expression_data->hack, 0);

	expression_data->state = 1;

	_x1f4_e4_size_expression
	    (expression_data, &max_list_length, &max_post_length);

	if (parser_data->bits & EXTERN_STACKS) {
	    const struct e4_context_type *context_data;
	    unsigned svth, *sxxx;

	    context_data = parser_data->context_data;

	    expression_data->post_size = max_post_length;
	    sxxx = context_data->link_s.s1st;
	    svth = sizeof(struct e4_post_type) * max_post_length;
	    if (*sxxx < svth) {
		*sxxx = svth;
	    }
	    expression_data->list_size = max_list_length;
	    sxxx = context_data->link_s.s2nd;
	    svth = sizeof(void *) * max_list_length;
	    if (*sxxx < svth) {
		*sxxx = svth;
	    }

	    expression_data->bits |= EXTERN_STACKS_MESS;
	} else {
	    status = M_LINK
		(parser_data, &expression_data->post_data,
		 sizeof(struct e4_post_type) * max_post_length);
	    if (status) {
		status = ALLOC_ERROR;
	    } else {
		expression_data->post_size = max_post_length;
		if (!max_list_length) {
		    expression_data->list = NULL;
		    expression_data->list_size = 0;
		} else {
		    status = M_LINK(parser_data, &expression_data->list,
				    sizeof(void *) * max_list_length);
		    if (status) {
			status = ALLOC_ERROR;
		    } else {
			expression_data->list_size = max_list_length;
		    }
		}
	    }
	}
    }

    return status;
}


static int
make_atom(struct e4_atom_type **atom, struct parser_type *parser_data)
{
    int status = 0;
    struct e4_hack_type *hack_data;
    unsigned count;

    hack_data = parser_data->hack_data;

    count = hack_data->count;
    if (count & ((count - 1) | 7)) {
	struct e4_atom_type *atom_data;

	atom_data = (struct e4_atom_type *) hack_data->atom_data + count;
	*atom = atom_data;
	parser_data->atom_data = atom_data;
	hack_data->count++;
    } else {
	unsigned class;

	class = count ? count << 1 : 8;
	status = M_MODE(parser_data, &hack_data->atom_data,
		        class * sizeof(struct e4_atom_type));
	if (status) {
	    status = ALLOC_ERROR;
	} else {
	    struct e4_atom_type *atom_data;
	    unsigned i;

	    atom_data = hack_data->atom_data;

	    atom_data += count;
	    hack_data->count++;
	    *atom = atom_data;
	    parser_data->atom_data = atom_data;
	    for (i = class - count; i; i--) {
		atom_data->type = MODE;
		atom_data->odb1 = NULL;
		atom_data++;
	    }
	}
    }

    return status;
}


static int
make_hack(struct e4_hack_type **hack, struct parser_type *parser_data)
{
    int status;
    struct e4_hack_type *hack_data;

    status = M_LINK(parser_data, &hack_data, sizeof(struct e4_hack_type));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	unsigned i;

	*hack = hack_data;
	for (i = 1; i; i--) {
	    hack_data->count = 0;
	    hack_data->atom_data = NULL;
	    hack_data->bits = READ_ONLY;
	    hack_data++;
	}
    }

    return status;
}


static int
make_list(struct e4_hack_type **hack, const struct e4_last_type *last_data,
	  struct parser_type *parser_data, unsigned offset)
{
#define BACK_HACK(count, excess) \
    (offset + (count) * sizeof(struct e4_hack_type))

    int status;
    struct e4_hack_type *hack_data;
    unsigned count;

    count = last_data->count;

    status = M_LINK(parser_data, &hack_data, BACK_HACK(count, excess));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	unsigned i;

	parser_data->offset = offset;

	hack_data = (struct e4_hack_type *) ((char *) hack_data + offset);

	*hack = hack_data;
	for (i = count; i; i--) {
	    hack_data->count = 0;
	    hack_data->atom_data = NULL;
	    hack_data->bits = READ_ONLY;
	    hack_data++;
	}
    }

    return status;

#undef BACK_HACK
}


static int
near_side(struct parser_type *parser_data)
{
    int status;
    struct e4_hack_type *hack_data;

    hack_data = parser_data->hack_data;

    do {
	const struct e4_last_type *last_data;
	struct link_type *link_data;
	unsigned class, frame, offset;
	void *hack;

	link_data = parser_data->link_data;
	last_data = link_data->last_data;

	frame = link_data->frame;

	if (frame == last_data->count) {
	} else {
	    if ((frame & (frame - 1)) | (frame < 8)) {
		link_data->frame = frame + 1;

		hack_data->count = 0;
		hack_data->atom_data = NULL;
		hack_data->bits = READ_ONLY;

		status = 0;

		break;
	    }
	}

	offset = link_data->offset;

	hack = (char *) (hack_data - frame) - offset;

#define BACK_HACK(count, excess) \
    (offset + (count) * sizeof(struct e4_hack_type))

	if (frame & (frame - 1)) {
	    class = frame | frame >> 001;
	    class = class | class >> 002;
	    class = class | class >> 004;
	    class = class | class >> 010;
	    class = class | class >> 020;
	    class = frame << 1 & ~class;
	} else {
	    class = frame << 1;
	}
	if (class < 8) {
	    class = 8;
	}

	status = M_MODE(parser_data, &hack, BACK_HACK(class, excess));
	if (status) {
	    status = ALLOC_ERROR;
	} else {
	    link_data->frame = frame + 1;

	    hack_data = (struct e4_hack_type *) ((char *) hack + offset);

	    link_data->atom_data->data.last.base = hack_data;

	    hack_data += frame;

	    hack_data->count = 0;
	    hack_data->atom_data = NULL;
	    hack_data->bits = READ_ONLY;
	}

#undef BACK_HACK

	parser_data->hack_data = hack_data;
    } while (0);

    return status;
}


static int
part_open(struct parser_type *parser_data)
{
    int status;
    struct e4_atom_type *atom_data;
    struct e4_hack_type *arguments, *hack_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
    } else {
	status = make_hack(&arguments, parser_data);
	if (status) {
	} else {
	    struct link_type *link_data;

	    atom_data->data.hack.data = arguments;
	    atom_data->type = THIS;

	    if (parser_data->odb1_data) {
		atom_data->odb1 = parser_data->odb1_data;
		parser_data->odb1_data = NULL;
	    }

	    link_data = parser_data->link_data;

	    hack_data = parser_data->hack_data;

	    status = push_hack(&link_data, hack_data, parser_data);
	    if (status) {
		M_FREE(parser_data, arguments);
	    } else {
#if 0
		link_data->atom_data = NULL;
#endif				/* 0 */
		link_data->count = 0;
		link_data->last_data = NULL;

		hack_data = arguments;

		parser_data->hack_data = hack_data;
		parser_data->link_data = link_data;

		if (0) {
		} else {
		    parser_data->link_class = JOIN_CLASS;
		}

		atom_data = hack_data->atom_data;
		parser_data->atom_data = atom_data + hack_data->count - 1;
	    }
	}
    }

    return status;
}


static int
pick_atom(int *type, int *miss, struct e4_atom_type **atom,
	  struct e4_atom_type *atom_data, struct e4_hack_type *hack_data,
	  struct parser_type *parser_data)
{
    int status = 0;

#if 0
    if (atom_data->odb1) {
	/*
	 * Not really, no.  Don't want this.  Parentheses should have higher
	 * priority.
	 */
	struct e4_hack_type hack;

	hack.count = 1;
	hack.atom_data = atom_data;
	status = pick_this(type, miss, atom, &hack, parser_data);
    } else {
#elif 0
    }
#else
    do {
#endif				/* 0 */
	int pick;

	pick = atom_data->type;
	if (atom_data != hack_data->atom_data) {
	    const struct e4_odbx_type *odbx_data;

	    odbx_data = atom_data[-1].odb2._1st;
	    if (odbx_data) {
		if (0200 < odbx_data->pass) {
		} else {
		    if (odbx_data != _x1f4_e4_pick) {
			*type = odbx_data->type;

			break;
		    }
		}
	    }
	}
	if (1 << pick & LOGIC_BITS) {
	    *type = last_base(atom_data)->last_data->type;
	} else {
	    if (pick == THIS) {
		status = pick_this
		    (type, miss, atom, atom_data->data.hack.data, parser_data);
	    } else {
		if (pick == EVER) {
		    *atom = atom_data;
		    *miss = 1;
		    *type = atom_data->data.ever.ever_data->type;
		} else {
		}
	    }
	}
    } while (0);

    return status;
}


static int
pick_this(int *type, int *miss, struct e4_atom_type **atom,
	  struct e4_hack_type *hack_data, struct parser_type *parser_data)
{
    int status;
    struct e4_atom_type *atom_data;
    unsigned count;

    do {
	atom_data = hack_data->atom_data;

	count = hack_data->count;
	if (count ^ 1) {
	} else {
	    if (atom_data->odb1) {
	    } else {
		status = pick_atom
		    (type, miss, atom, atom_data, hack_data, parser_data);
		if (1) {
		    break;
		}
	    }
	}

	status = _x1f4_e4_type_hack(hack_data, type, parser_data);
    } while (0);

    return status;
}


static int
push_hack(struct link_type **link, struct e4_hack_type *hack_data,
	  struct parser_type *parser_data)
{
    int status;
    struct link_type *link_data;

    status = M_LINK(parser_data, &link_data, sizeof(struct link_type));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	link_data->link_data = *link;
	link_data->data = hack_data;
	link_data->part = NULL;
	link_data->trip_data = parser_data->rail_link;

	parser_data->rail_link = NULL;

	*link = link_data;
    }

    return status;
}


static int
spin_dana(struct next_type *next_data, struct parser_type *parser_data)
{
    return trap_dana
	(&(*parser_data->recall)->last, parser_data->ever_call,
	 parser_data->recall, parser_data);
}


static int
spin_ever(struct next_type *next_data, struct parser_type *parser_data)
{
    int expect, status;

    parser_data->relink = 1;
    status = trap_last(next_data, parser_data, &expect);
    parser_data->relink = 0;

    return status;
}


static int
spin_last(struct next_type *next_data, struct parser_type *parser_data)
{
    int expect, status;

    parser_data->relink = 1;
    status = trap_ever(next_data, parser_data, &expect);
    parser_data->relink = 0;
    if (status) {
    } else {
	const struct e4_ever_type *ever_data;
	const struct e4_last_type *last_data;
	const void *line;

	ever_data = next_data->data.ever.ever_data;
	line = *(void **) next_data->data.ever.state;
	last_data = &((struct e4_line_type *) line)->last;

	status = trap_dana
	    (last_data, ever_data, next_data->data.ever.state, parser_data);
    }

    return status;
}


static int
trap_bill(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
    } else {
	C_BILL *slip;

	atom_data->type = BILL;
	slip = (C_BILL *) &atom_data->data.lead.data;
	*slip = next_data->data.bill;

	if (parser_data->odb1_data) {
	    atom_data->odb1 = parser_data->odb1_data;
	    parser_data->odb1_data = NULL;
	}

	*expect = parser_data->link_class;
    }

    return status;
}


static int
trap_case(struct next_type *next_data, struct parser_type *parser_data,
	  int (*spin)(struct next_type *, struct parser_type *))
{
    int status;
    struct e4_atom_type *atom_data, *atom_text, turn;
    struct e4_hack_type *hack_data;
    unsigned pick;

    atom_data = parser_data->atom_data;
    hack_data = parser_data->hack_data;
    atom_text = hack_data->atom_data;

    while (atom_data != atom_text) {
	const struct e4_odbx_type *odbx_data;

	odbx_data = atom_data[-1].odb2._1st;
	if (odbx_data) {
	    if (odbx_data->pass == 0200) {
		atom_data--;
	    } else {
		break;
	    }
	} else {
	    break;
	}
    }

    pick = parser_data->atom_data - atom_data;
    if (pick) {
	hack_data->count -= pick;
	parser_data->atom_data = atom_data;
    }

    turn = *atom_data;
    turn.odb1 = NULL;

    atom_data->type = MODE;

    status = spin(next_data, parser_data);
    if (status) {
    } else {
	int expect;

	status = trap_open(next_data, parser_data, &expect);
	if (status) {
	} else {
	    status = make_atom(&atom_text, parser_data);
	    if (status) {
	    } else {
		*atom_text = turn;
		turn.type = MODE;
		for (; pick; pick--) {
		    status = make_atom(&atom_text, parser_data);
		    if (status) {
			break;
		    }
		    atom_data++;
		    *atom_text = *atom_data;
		    atom_data->type = MODE;
		    atom_data->odb1 = NULL;
		}
	    }
	}
    }

    if (turn.type != MODE) {
	_x1f4_e4_free_atom(&turn, parser_data->expression_data);
    }

    for (; pick; pick--) {
	atom_data++;
	_x1f4_e4_free_atom(atom_data, parser_data->expression_data);
    }

    return status;
}


static int
trap_dana(const struct e4_last_type *last_data,
	  const struct e4_ever_type *ever_data, void *state,
	  struct parser_type *parser_data)
{
    int status;
    struct e4_atom_type *atom_data;
    void *odb1;

    atom_data = parser_data->atom_data;

    odb1 = atom_data->odb1;
    atom_data->odb1 = NULL;

    _x1f4_e4_free_atom(atom_data, parser_data->expression_data);

    atom_data->odb1 = odb1;

    if (0) {
    } else {
	struct e4_hack_type *arguments;

	status = make_list
	    (&arguments, last_data, parser_data,
	     sizeof(struct e4_base_type) + sizeof(struct e4_fine_type));
	if (status) {
	    atom_data->type = MODE;
	} else {
	    struct e4_fine_type *fine_data;

	    fine_data = hack_fine(arguments);

	    fine_data->ever_data = ever_data;
	    fine_data->line = state;

	    hack_base(arguments)->last_data = last_data;

	    atom_data->odb2._1st = NULL;
	    atom_data->type = LOCK;
	    atom_data->data.last.base = arguments;

	    if (parser_data->odb1_data) {
		atom_data->odb1 = parser_data->odb1_data;
		parser_data->odb1_data = NULL;
	    }
	}
    }

    return status;
}


static int
trap_ever(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    if (parser_data->relink) {
	status = 0;
	atom_data = parser_data->atom_data;
    } else {
	status = make_atom(&atom_data, parser_data);
    }
    if (status) {
    } else {
	atom_data->type = EVER;
	atom_data->data.ever.state = next_data->data.ever.state;
	atom_data->data.ever.ever_data = next_data->data.ever.ever_data;

	if (parser_data->odb1_data) {
	    atom_data->odb1 = parser_data->odb1_data;
	    parser_data->odb1_data = NULL;
	}

	*expect = parser_data->link_class;
    }

    return status;
}


static int
trap_join(struct next_type *next_data, const struct e4_last_type *last_data,
	  struct parser_type *parser_data)
{
    int status;

    do {
	if (last_data->type == VOID) {
	    struct e4_atom_type *atom_data;
	    struct e4_hack_type *hack_data;
	    unsigned count;

	    hack_data = parser_data->hack_data;

	    atom_data = hack_data->atom_data;
#if 0
	    atom_data++;

	    /* use to be like this.  remove, maybe */

	    count = hack_data->count - 1;
	    for (; count; count--) {
		if (atom_data->type ^ PICK) {
		    status = _x1f4_e4_130704_error_0(parser_data);
		    if (1) {
			break;
		    }
		} else {
		    atom_data++;
		}
	    }
#else
	    count = hack_data->count - 1;
	    for (; count; count--) {
		if (((struct e4_odbx_type *) atom_data->odb2._1st)->pass
		    ^ 0200) {
		    status = _x1f4_e4_130704_error_0(parser_data);
		    if (1) {
			break;
		    }
		} else {
		    atom_data++;
		}
	    }
#endif				/* 0 */
	    if (count) {
		break;
	    }
	}

	status = trap_pick(next_data, last_data, parser_data);
    } while (0);

    return status;
}


static int
trap_last(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    if (parser_data->relink) {
	status = 0;
	atom_data = parser_data->atom_data;
    } else {
	status = make_atom(&atom_data, parser_data);
    }
    if (status) {
    } else {
	struct e4_hack_type *arguments;
	const struct e4_last_type *last_data;

	last_data = next_data->data.last.last_data;

	status = make_list
	    (&arguments, last_data, parser_data, sizeof(struct e4_base_type));
	if (status) {
	} else {
	    atom_data->odb2._1st = NULL;
	    atom_data->type = LAST;
	    atom_data->data.last.base = arguments;
	    hack_base(arguments)->last_data = last_data;

	    if (last_data->bits & TEXT_LINK) {
		hack_base(arguments)->text = next_data->data.last.text;
	    }

	    if (parser_data->odb1_data) {
		atom_data->odb1 = parser_data->odb1_data;
		parser_data->odb1_data = NULL;
	    }

	    *expect = OPEN_CLASS;
	}
    }

    return status;
}


static int
trap_lock(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
    } else {
	struct e4_hack_type *arguments;
	const struct e4_last_type *last_data;

	last_data = next_data->data.lock.last_data;

	status = make_list
	    (&arguments, last_data, parser_data,
	     sizeof(struct e4_base_type) + sizeof(struct e4_fine_type));
	if (status) {
	} else {
	    struct e4_fine_type *fine_data;

	    fine_data = hack_fine(arguments);

	    fine_data->ever_data = next_data->data.lock.ever_data;
	    fine_data->line = next_data->data.lock.state;

	    hack_base(arguments)->last_data = last_data;

	    atom_data->odb2._1st = NULL;
	    atom_data->type = LOCK;
	    atom_data->data.last.base = arguments;

	    if (parser_data->odb1_data) {
		atom_data->odb1 = parser_data->odb1_data;
		parser_data->odb1_data = NULL;
	    }

	    *expect = OPEN_CLASS;
	}
    }

    return status;
}


static int
trap_lose(SET_NEXT__________ARGS_1)
{
    int status = 0;
    struct e4_hack_type *hack_data, *past_hack;
    const struct e4_last_type *last_data;
    struct link_type *link_data;
    unsigned frame;

    hack_data = parser_data->hack_data;
    link_data = parser_data->link_data;

    past_hack = hack_data;

    last_data = link_data->last_data;

    frame = link_data->frame;

    status = join_hack(&link_data, &hack_data, parser_data);
    if (status) {
    } else {
	struct e4_atom_type *atom_data;
	unsigned bits;

	if (last_data) {
	    parser_data->argument--;
	    if (parser_data->bits & PACK_ESTORAGE) {
		if (last_data->count) {
		    M_MODE(parser_data, &past_hack->atom_data,
			   past_hack->count * sizeof(struct e4_atom_type));
		}
	    }

	    bits = last_data->bits;
	} else {
	    if (past_hack->count ^ 1) {
	    } else {
		past_hack->atom_data->odb2._1st = NULL;
	    }

	    if (1) {
		if (parser_data->bits & PACK_ESTORAGE) {
		    M_MODE(parser_data, &past_hack->atom_data,
			   past_hack->count * sizeof(struct e4_atom_type));
		}
	    }

	    bits = 0;
	}

	if (link_data) {
	    if (link_data->last_data) {
		/*
		 * 0 is allowed for variadic functions with no mandatory
		 * arguments
		 */
		if (link_data->count < 2) {
		    *expect = JOIN_CLASS;
		    parser_data->link_class = JOIN_CLASS;
		} else {
		    *expect = POST_CLASS;
		    parser_data->link_class = POST_CLASS;
		}
	    } else {
		if (1) {
		    *expect = JOIN_CLASS;
		    parser_data->link_class = JOIN_CLASS;
		}
	    }
	} else {
	    parser_data->link_class = NEXT_CLASS;
	    *expect = NEXT_CLASS;
	}

	atom_data = hack_data->atom_data;
	atom_data += hack_data->count - 1;

	parser_data->atom_data = atom_data;

	if (bits & (SIDE_LIST | SLIP_LIST)) {
	    status = case_side(atom_data, frame, last_data, parser_data);
	}
    }

    return status;
}


static int
trap_make(SET_NEXT__________ARGS_1)
{
    int status;

    status = trap_case(next_data, parser_data, spin_ever);
    if (status) {
    } else {
	const char *e;
	int c;
	const struct e4_last_type *last_data;

	last_data = next_data->data.last.last_data;

	e = parser_data->expression;
	c = *(const unsigned char *) e;
	if (c == '(') {
	    parser_data->expression = e + 1;

	    ever_turn(parser_data);

	    if (*parser_data->expression == ')') {
		if (last_data->count == 1) {
		    *expect = LOSE_CLASS;
		} else {
		    status = _x1f4_e4_110608_error_4(parser_data);
		}
	    } else {
		if (last_data->count == 1) {
		    status = _x1f4_e4_170926_error_2
			(parser_data, parser_data->link_data->atom_data);
		} else {
		    status = trap_step(SET_NEXT__________ARGS_2);
		}
	    }
	} else {
	    if (last_data->count == 1) {
		if (c == ')' || c == ',' || c == parser_data->terminator
		    || c == '.' || C_TYPE_X(parser_data->c_type_0, c)) {
		    status = trap_lose(SET_NEXT__________ARGS_2);
		} else {
		    status = _x1f4_e4_170926_error_0(parser_data);
		}
	    } else {
		status = _x1f4_e4_170926_error_0(parser_data);
	    }
	}
    }

    return status;
}


static int
trap_mode(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
    } else {
	C_MODE *slip;

	atom_data->type = MODE;
	slip = (C_MODE *) &atom_data->data.lead.data;
	*slip = next_data->data.mode;

	if (parser_data->odb1_data) {
	    atom_data->odb1 = parser_data->odb1_data;
	    parser_data->odb1_data = NULL;
	}

	*expect = parser_data->link_class;
    }

    return status;
}


static int
trap_odb1(SET_NEXT__________ARGS_1)
{
    parser_data->odb1_data = next_data->data.odb1_data;

    *expect = DATA_CLASS;

    return 0;
}


static int
trap_odb2(SET_NEXT__________ARGS_1)
{
    int status = 0;
    struct e4_atom_type *atom_data;
    const struct e4_odbx_type *odbx_data;

    atom_data = parser_data->atom_data;

    odbx_data = next_data->data.odbx.odbx_data;

    atom_data->odb2._1st = odbx_data;

    *expect = DATA_CLASS;

    if (odbx_data->bits & E2ND_LINK) {
	const struct e4_2nde_type *e2nd_data;

	e2nd_data = odbx_data->extension2;
	if (e2nd_data->bits & (POST_MARK | RAIL_ONLY)) {
	    status = part_open(parser_data);
	    if (status) {
	    } else {
		struct link_type *link_data;

		link_data = parser_data->link_data;

		link_data->odbx_data = odbx_data;

		if (e2nd_data->bits & POST_MARK) {
		    const struct e4_mark_type *mark_data;

		    mark_data = e2nd_data->post;

		    link_data->land = mark_data->size;
		    link_data->part = mark_data->part;
		} else {
		    const struct e4_odbx_type *odbx_rail;

		    odbx_rail = e2nd_data->rail;

		    link_data->land = odbx_rail->size;
		    link_data->part = odbx_rail->name;
		}
	    }
	}
    }

    return status;
}


static int
trap_open(SET_NEXT__________ARGS_1)
{
    int status;

    do {
	struct e4_hack_type *hack_data;
	struct link_type *link_data;
	unsigned count;

	hack_data = parser_data->hack_data;
	link_data = parser_data->link_data;

	count = hack_data->count;
	if (count) {
	    struct e4_atom_type *atom_data;

	    atom_data = parser_data->atom_data;

	    /*
	     * function call is if... no binary attached.
	     * otherwise, it is an opening parenthesis coming after a completed
	     * function call.
	     */
	    if (1 << atom_data->type & LOGIC_BITS && !atom_data->odb2._1st) {
		const struct e4_last_type *last_data;

		status = push_hack(&link_data, hack_data, parser_data);
		if (status) {
		    break;
		}

		hack_data = atom_data->data.last.base;

		last_data = hack_base(hack_data)->last_data;

		link_data->atom_data = atom_data;
		link_data->last_data = last_data;
		link_data->count = last_data->count;
		link_data->frame = last_data->count;
		link_data->offset = parser_data->offset;

		parser_data->argument++;

		parser_data->hack_data = hack_data;
		parser_data->link_data = link_data;

		atom_data = hack_data->atom_data;
		parser_data->atom_data = atom_data + hack_data->count - 1;

		if (!last_data->count) {
		    if (last_data->bits & SIDE_LIST) {
			*expect = DATA_CLASS;
			parser_data->link_class = JOIN_CLASS;
			status = near_side(parser_data);
		    } else {
			*expect = LOSE_CLASS;
		    }
		} else {
		    *expect = DATA_CLASS;
		    if (last_data->count == 1) {
			parser_data->link_class = JOIN_CLASS;
		    } else {
			parser_data->link_class = POST_CLASS;
		    }
		}

		break;
	    }
	}

	*expect = DATA_CLASS;

	status = part_open(parser_data);
    } while (0);

    return status;
}


static int
trap_pick(struct next_type *next_data, const struct e4_last_type *last_data,
	  struct parser_type *parser_data)
{
    int status;
    struct e4_atom_type *atom_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
    } else {
	struct e4_hack_type *arguments;

	status = make_list
	    (&arguments, last_data, parser_data,
	     sizeof(struct e4_base_type) + sizeof(struct e4_near_type));
	if (status) {
	} else {
	    hack_base(arguments)->last_data = last_data;

	    atom_data->odb2._1st = NULL;
	    atom_data->type = PICK;

	    atom_data->data.last.base = arguments;

	    if (parser_data->odb1_data) {
		atom_data->odb1 = parser_data->odb1_data;
		parser_data->odb1_data = NULL;
	    }

	    (atom_data - 1)->odb2._1st = _x1f4_e4_pick;
	}
    }

    return status;
}


static int
trap_real(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
    } else {
	C_REAL *slip;

	atom_data->type = REAL;
	slip = (C_REAL *) &atom_data->data.lead.data;
	*slip = next_data->data.real;

	if (parser_data->odb1_data) {
	    atom_data->odb1 = parser_data->odb1_data;
	    parser_data->odb1_data = NULL;
	}

	*expect = parser_data->link_class;
    }

    return status;
}


static int
trap_roll(SET_NEXT__________ARGS_1)
{
    int status;

    status = trap_case(next_data, parser_data, spin_last);
    if (status) {
    } else {
	const struct e4_last_type *last_data;
	const void *line;

	line = *(void **) next_data->data.ever.state;
	last_data = &((struct e4_line_type *) line)->last;

	if (*parser_data->expression == '(') {
	    parser_data->expression++;
	    ever_turn(parser_data);

	    if (*parser_data->expression == ')') {
		if (last_data->count < 2) {
		    *expect = LOSE_CLASS;
		} else {
		    status = _x1f4_e4_110608_error_4(parser_data);
		}
	    } else {
		if (last_data->count < 2) {
		    if (last_data->bits & SIDE_LIST) {
			status = trap_side(SET_NEXT__________ARGS_2);
		    } else {
			status = _x1f4_e4_170926_error_2
			    (parser_data, parser_data->link_data->atom_data);
		    }
		} else {
		    status = trap_step(SET_NEXT__________ARGS_2);
		}
	    }
	} else {
	    status = trap_lose(SET_NEXT__________ARGS_2);
	}
    }

    return status;
}


static int
trap_side(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_hack_type *hack_data;

    hack_data = parser_data->hack_data;

    status = join_rail(parser_data);
    if (status) {
    } else {
	parser_data->rail_link = NULL;

	*expect = DATA_CLASS;

	parser_data->link_class = JOIN_CLASS;

	if (parser_data->bits & PACK_ESTORAGE) {
	    M_MODE(parser_data, &hack_data->atom_data,
		   hack_data->count * sizeof(struct e4_atom_type));
	}

	hack_data++;

	parser_data->hack_data = hack_data;

	status = near_side(parser_data);
    }

    return status;
}


static int
trap_step(SET_NEXT__________ARGS_1)
{
    int status;

    do {
	unsigned count;
	struct e4_hack_type *hack_data;
	struct link_type *link_data;

	hack_data = parser_data->hack_data;

	if (parser_data->rail_link) {
	    status = join_rail(parser_data);

	    parser_data->rail_link = NULL;

	    if (status) {
		break;
	    }
	} else {
	    status = 0;
	}

	link_data = parser_data->link_data;
	count = link_data->count;
	count--;
	link_data->count = count;
	*expect = DATA_CLASS;
	if (count == 1) {
	    parser_data->link_class = JOIN_CLASS;
	} else {
	    parser_data->link_class = POST_CLASS;
	}

	if (parser_data->bits & PACK_ESTORAGE) {
	    M_MODE(parser_data, &hack_data->atom_data,
		   hack_data->count * sizeof(struct e4_atom_type));
	}

	hack_data++;
	parser_data->hack_data = hack_data;
    } while (0);

    return status;
}


static int
trap_text(SET_NEXT__________ARGS_1)
{
    int status;
    struct e4_atom_type *atom_data;

    status = make_atom(&atom_data, parser_data);
    if (status) {
	M_FREE(parser_data, next_data->data.text);
    } else {
	C_TEXT *slip;

	atom_data->type = TEXT;
	slip = (C_TEXT *) &atom_data->data.lead.data;
	*slip = next_data->data.text;

	if (parser_data->odb1_data) {
	    atom_data->odb1 = parser_data->odb1_data;
	    parser_data->odb1_data = NULL;
	}

	*expect = parser_data->link_class;
    }

    return status;
}


static void
ever_slip(struct parser_type *parser_data)
{
    const char *e;

    e = parser_data->expression;

    while (1) {
	int c;

	c = *(const unsigned char *) e;
	if (C_TYPE_X(c_type_4, c)) {
	    e++;
	} else {
	    if (c ^ 35) {
		break;
	    } else {
		do {
		    ++e;
		    c = *(const unsigned char *) e;
		} while (c ^ 10);
	    }
	}
    }

    parser_data->expression = e;
}


static void
ever_turn(struct parser_type *parser_data)
{
    const char *e;
    int v;

    v = parser_data->bits & SKIP_SCOMMENT ? 35 : -35;

    e = parser_data->expression;

    while (1) {
	int c;

	c = *(const unsigned char *) e;
	if (C_TYPE_X(c_type_4, c)) {
	    e++;
	} else {
	    if (c == v) {
		do {
		    ++e;
		    c = *(const unsigned char *) e;
		} while (c ^ 10);
	    } else {
		break;
	    }
	}
    }

    parser_data->expression = e;
}


static void
fill_data(struct parser_type *parser_data)
{
    unsigned *e_type_0;
    unsigned char *selection;

    selection = parser_data->class_select;

    memset(selection, 0, 256);

    fill_type(selection, _x1f4_c_type_1, 1);
    fill_type(selection, c_type_2, 2);

    selection['"'] = 3;
    selection[')'] = 4;
    selection[','] = 5;
    selection['('] = 6;

    selection[047] = 9;

    selection['.'] = 10;

    e_type_0 = parser_data->c_type_0;

    e_type_0[0] = c_type_0[0];
    e_type_0[1] = c_type_0[1];
    e_type_0[2] = c_type_0[2];
    e_type_0[3] = c_type_0[3];
    e_type_0[4] = c_type_0[4];
    e_type_0[5] = c_type_0[5];
    e_type_0[6] = c_type_0[6];
    e_type_0[7] = c_type_0[7];

    if (parser_data->bits & SKIP_SCOMMENT) {
	e_type_0[1] &= ~8;
    }

    fill_type(selection, e_type_0, 7);

    selection[((unsigned) parser_data->terminator) & 255] = 8;

    parser_data->slip = lead_atom;
    if (parser_data->bits & SUIT_SLIPTEST) {
	if (parser_data->completion == BACK_COMPLETION) {
	    parser_data->slip = lead_back;
	}
    }
}


static void
fill_type(unsigned char *selection, const unsigned *c_type_x, unsigned class)
{
    unsigned c = 0;

    for (; c < 256; c++) {
	if (C_TYPE_X(c_type_x, c)) {
	    selection[c] = class;
	}
    }
}


static void
free_long(struct parser_type *parser_data)
{
    struct e4_expression_type *expression_data;
    struct link_type *link_data;

    expression_data = parser_data->expression_data;

    link_data = parser_data->link_data;
    while (link_data) {
	const struct e4_last_type *last_data;
	struct link_type *free_link;
	struct trip_type *trip_data;

	last_data = link_data->last_data;
	if (last_data) {
	    unsigned count, frame;

	    count = last_data->count;
	    frame = link_data->frame - count;
	    if (frame) {
		struct e4_hack_type *hack_data;

		hack_data = link_data->atom_data->data.last.base;
		hack_data += count;
		for (; frame; frame--) {
		    _x1f4_e4_free_hack(hack_data, expression_data);
		    hack_data++;
		}
	    }
	}

	trip_data = link_data->trip_data;
	if (trip_data) {
	    free_trip(parser_data, trip_data);
	}

	free_link = link_data;
	link_data = link_data->link_data;
	M_FREE(parser_data, free_link);
    }
}


static void
free_odb1(struct e4_odb1_type *odb1_data, void *expression)
{
    while (odb1_data) {
	struct e4_odb1_type *free_odb1;

	free_odb1 = odb1_data;
	odb1_data = odb1_data->odb1_data;
	E_FREE(expression, free_odb1);
    }
}


static void
free_trip(struct parser_type *parser_data, struct trip_type *trip_data)
{
    while (trip_data) {
	struct trip_type *free_trip;

	free_trip = trip_data;
	trip_data = trip_data->trip_data;
	M_FREE(parser_data, free_trip);
    }
}


int
x1f4_inIt_expression(struct e4_expression_type **expression,
		     const char **expressIon, unsigned bits,
		     const struct e4_context_type *context_data)
{
    int status;
    struct e4_expression_type *expression_data;
    struct parser_type parser;

    if (bits & RESOURCE_LINK) {
	parser.m = context_data->link_m;
    } else {
	parser.m.free = _x1f4_e4_free_data;
	parser.m.link = _x1f4_e4_link_data;
	parser.m.mode = _x1f4_e4_mode_data;
    }

    status = parser.m.link
	(parser.m.data, (void *) &expression_data,
	 sizeof(struct e4_expression_type));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	if (bits & CASECALL_LINK) {
	    parser.ever_call = context_data->near_f.ever_data;
	    parser.recall = context_data->near_f.line;
	} else {
	    parser.recall = NULL;
	}

	if (bits & DEPREFIX_LINK) {
	    parser.prefix = context_data->near_e.fine;
	} else {
	    parser.prefix = NULL;
	}

	if (bits & DIRECT_ACCESS) {
	    expression_data->bits = DIRECT_DATA_ACCESS;
	    if (bits & DIRECT_DEVICE) {
		parser.dddevice = context_data->near_d.lead;
	    } else {
		parser.dddevice = NULL;
	    }
	} else {
	    expression_data->bits = 0;
	    parser.dddevice = NULL;
	}

	if (bits & ERROR_STORAGE) {
	    parser.back_data = context_data->link_e.back;
	} else {
	    parser.back_data = NULL;
	}

	if (bits & TRANSFER_LINK) {
	    parser.e4fine = context_data->link_i.fine;
	} else {
	    parser.e4fine = NULL;
	}

	expression_data->flat_data = NULL;
	expression_data->hack.count = 0;
	expression_data->hack.atom_data = NULL;
	expression_data->list = NULL;
	expression_data->m = parser.m;
	expression_data->state = 0;
	expression_data->post_data = NULL;
	parser.context_data = context_data;
	parser.completion = context_data->completion;
	parser.division = *expressIon;
	parser.expression = parser.division;
	parser.expression_data = expression_data;
	parser.bits = bits;
	parser.hack_data = &expression_data->hack;
	parser.relink = 0;
	parser.terminator = context_data->terminator;
	fill_data(&parser);
	status = init_data(&parser);
	if (status) {
	    x1f4_fini_expression(&expression_data);
	} else {
	    *expression = expression_data;
	}

	*expressIon = parser.expression;
    }

    return status;
}
