/*
 * e4.e.c
 * Copyright (C) 2007-2014, 2016, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <ctype.0.h>
#include <e4-defs.h>
#include <e4-errors.h>
#include <e4-inter.h>
#include <e4-parse.h>
#include <e4-types.h>

#define parser(text) \
    ((struct parser_type *) (text))

extern const unsigned _x1f4_c_type_0[];

static int pick_110608_error_1(struct e4_back_type *, void *);
static int pick_120712_error_0(struct e4_back_type *, void *);

static void copy_trip(struct e4_mark_type *, struct trip_type *);
static void seek_mind(const char **, unsigned *, const char *, unsigned,
		      void *);
static void type_mind(struct e4_mind_type *, const struct e4_atom_type *,
		      void *);

static int
pick_110608_error_1(struct e4_back_type *back_data, void *parser)
{
    const char *e, *expression;
    int c;

    expression = parser(parser)->expression;

    e = expression;

    do {
	e++;
	c = *(const unsigned char *) e;
    } while (C_TYPE_X(_x1f4_c_type_0, c));

    back_data->code = NEARNAME;

    back_data->data.nearname.name = expression;
    back_data->data.nearname.size = e - expression;

    return 0;
}


static int
pick_120712_error_0(struct e4_back_type *back_data, void *parser)
{
    int delete = 0;
    struct link_type *link_data;

    link_data = parser(parser)->link_data;
    if (link_data) {
	const struct e4_last_type *last_data;

	last_data = link_data->last_data;
	if (last_data) {
	    if (last_data->count || last_data->bits & SIDE_LIST) {
	    } else {
		back_data->code = NEARVOID;
		delete = PARSE_ERROR;
		type_mind
		    (&back_data->data.nearvoid.mind, link_data->atom_data,
		     parser);
	    }
	}
    }

    return delete;
}


static void
copy_trip(struct e4_mark_type *mark_data, struct trip_type *trip_data)
{
    mark_data[0].part = trip_data->lead.name;
    mark_data[0].size = trip_data->lead.size;
    mark_data[2].part = trip_data->rail.name;
    mark_data[2].size = trip_data->rail.size;

    if (trip_data->left.bits & POST_MARK) {
	mark_data[1] = *trip_data->left.post;
    } else {
	mark_data[1].size = 0;
    }
}


static void
seek_mind(const char **post, unsigned *cast, const char *name, unsigned size, void *parser)
{
    const char *division, *expression;
    unsigned c;

    division = parser(parser)->division;
    expression = parser(parser)->expression;

    expression -= size;

    c = *(unsigned char *) name;

    name++;
    size--;

    while (1) {
	if (*(unsigned char *) expression ^ c) {
	} else {
	    const char *mind, *seek;
	    unsigned i;

	    mind = name;
	    seek = expression + 1;
	    i = size;
	    for (; i; i--) {
		if (*mind ^ *seek) {
		    break;
		} else {
		    mind++;
		    seek++;
		}
	    }

	    if (i) {
	    } else {
		*cast = size + 1;
		*post = expression;
		if (1) {
		    break;
		}
	    }
	}

	if (expression == division) {
	    *cast = 0;
	    *post = NULL;
	    if (1) {
		break;
	    }
	} else {
	    expression--;
	}
    }
}


static void
type_mind(struct e4_mind_type *mind_data, const struct e4_atom_type *atom_data,
	  void *parser)
{
    if (atom_data) {
	int type;

	type = atom_data->type;

	mind_data->type = type;

	if (type ^ LAST) {
	    if (type ^ PICK) {
		const struct e4_ever_type *ever_data;
		unsigned size;

		ever_data = lock_fine(atom_data)->ever_data;

		size = ever_data->size;

		seek_mind
		    (&mind_data->name, &mind_data->size, ever_data->name, size,
		     parser);
	    }
	} else {
	    const struct e4_last_type *last_data;
	    unsigned size;

	    last_data = last_base(atom_data)->last_data;

	    size = last_data->size;

	    seek_mind
		(&mind_data->name, &mind_data->size, last_data->name, size,
		 parser);
	}
    } else {
	mind_data->type = VOID;
    }
}


int
_x1f4_e4_080704_error_0(void *parser, const char *name, unsigned size)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = L_LOOKUP;
	back_data->data.l_lookup.name = name;
	back_data->data.l_lookup.size = size;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_1(void *parser, const char *name, unsigned size)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = E_LOOKUP;
	back_data->data.e_lookup.name = name;
	back_data->data.e_lookup.size = size;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_3(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = BYTEMISS;
	back_data->data.bytemiss.byte = parser(parser)->expression;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_4(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = F_SYNTAX;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_5(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = EXPECTED;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_6(void *parser, const char *name, unsigned size,
			int f1st, int f2nd)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = BINARY_R;
	back_data->data.binary_r.f1st = f1st;
	back_data->data.binary_r.f2nd = f2nd;
	back_data->data.binary_r.name = name;
	back_data->data.binary_r.size = size;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_7(void *parser, struct e4_atom_type *atom_data)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	const struct e4_ever_type *ever_data;
	unsigned size;

	back_data->code = C_LVALUE;

	ever_data = atom_data->data.ever.ever_data;

	size = ever_data->size;

	seek_mind
	    (&back_data->data.c_lvalue.name, &back_data->data.c_lvalue.size,
	     ever_data->name, size, parser);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_a(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = 0;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_b(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = I_LVALUE;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_c(void *parser, struct e4_hack_type *hack_data)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	while (hack_data->count == 1) {
	    struct e4_atom_type *atom_data;

	    atom_data = hack_data->atom_data;
	    if (atom_data->type != THIS) {
		break;
	    } else {
		hack_data = atom_data->data.hack.data;
	    }
	}
	if (hack_data->count != 1) {
	    _x1f4_e4_080704_error_b(parser);
	} else {
	    struct e4_atom_type *atom_data;

	    atom_data = hack_data->atom_data;
	    if (atom_data->type != EVER) {
		_x1f4_e4_080704_error_b(parser);
	    } else {
		_x1f4_e4_080704_error_7(parser, atom_data);
	    }
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_d(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = R_SCALAR;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_e(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = B_SCALAR;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_f(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = C_STRING;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_g(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = U_PREFIX;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_h(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = U_BINARY;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_i(void *parser, const char *name, unsigned size)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = P_LOOKUP;
	back_data->data.p_lookup.name = name - size;
	back_data->data.p_lookup.size = size;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_j(void *parser, const char *name, unsigned size)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = B_LOOKUP;
	back_data->data.b_lookup.name = name;
	back_data->data.b_lookup.size = size;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_080704_error_k(void *parser, const void *atom, unsigned rank,
			int f1st, int f2nd)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = MISSTYPE;
	back_data->data.misstype.f1st = f1st;
	back_data->data.misstype.f2nd = f2nd;
	back_data->data.misstype.rank = rank;
	type_mind(&back_data->data.misstype.mind, atom, parser);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_081014_error_0(struct e4_back_type *back_data, const void *atom,
			unsigned rank, void *parser)
{
    if (back_data) {
	back_data->code = I_SVALUE;
	back_data->data.i_svalue.rank = rank;
	type_mind(&back_data->data.i_svalue.mind, atom, parser);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_081014_error_1(void *parser, struct e4_hack_type *hack_data,
			const void *atom, unsigned rank, int f1st, int f2nd)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	while (hack_data->count == 1) {
	    struct e4_atom_type *atom_data;

	    atom_data = hack_data->atom_data;
	    if (atom_data->type != THIS) {
		break;
	    } else {
		hack_data = atom_data->data.hack.data;
	    }
	}
	if (hack_data->count != 1) {
	    _x1f4_e4_081014_error_0(back_data, atom, rank, parser);
	} else {
	    struct e4_atom_type *atom_data;

	    atom_data = hack_data->atom_data;
	    if (atom_data->type != EVER) {
		_x1f4_e4_081014_error_0(back_data, atom, rank, parser);
	    } else {
		back_data->code = T_SVALUE;
		back_data->data.t_svalue.f1st = f1st;
		back_data->data.t_svalue.f2nd = f2nd;
		back_data->data.t_svalue.rank = rank;
		type_mind(&back_data->data.t_svalue.mind, atom, parser);
	    }
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_081014_error_2(void *parser, struct e4_hack_type *hack_data,
			const void *atom, unsigned rank)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	while (hack_data->count == 1) {
	    struct e4_atom_type *atom_data;

	    atom_data = hack_data->atom_data;
	    if (atom_data->type != THIS) {
		break;
	    } else {
		hack_data = atom_data->data.hack.data;
	    }
	}
	if (hack_data->count != 1) {
	    _x1f4_e4_081014_error_0(back_data, atom, rank, parser);
	} else {
	    struct e4_atom_type *atom_data;

	    atom_data = hack_data->atom_data;
	    if (atom_data->type != EVER) {
		_x1f4_e4_081014_error_0(back_data, atom, rank, parser);
	    } else {
		back_data->code = C_SVALUE;
		back_data->data.c_svalue.rank = rank;
		type_mind(&back_data->data.c_svalue.mind, atom, parser);
	    }
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_081019_error_0(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = C_SYNTAX;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110605_error_0(void *parser, const void *atom)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = SURESTEP;
	type_mind(&back_data->data.surestep.mind, atom, parser);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110605_error_1(void *parser, const void *atom)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = SURELOSE;
	type_mind(&back_data->data.surelose.mind, atom, parser);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_1(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	if (pick_120712_error_0(back_data, parser)) {
	} else {
	    pick_110608_error_1(back_data, parser);
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_2(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	if (pick_120712_error_0(back_data, parser)) {
	} else {
	    back_data->code = NEARREAL;
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_3(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = NEARTEXT;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_4(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = NEARLOSE;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_5(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = NEARSTEP;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_6(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = NEAROPEN;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_7(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = NEARJOIN;
	back_data->data.nearjoin.join =
	    *(const unsigned char *) parser(parser)->expression;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_110608_error_9(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = NEARCITE;
    }

    return PARSE_ERROR;
}


#if 0
int
_x1f4_e4_110817_error_0(void *parser, const char *name, unsigned size)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = MISSCALL;
	back_data->data.misscall.name = name;
	back_data->data.misscall.size = size;
    }

    return PARSE_ERROR;
}
#endif				/* 0 */


int
_x1f4_e4_130607_error_0(void *parser, int f1st, int f2nd)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = MISSTURN;
	back_data->data.missturn.f1st = f1st;
	back_data->data.missturn.f2nd = f2nd;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_130704_error_0(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = LONGVOID;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170621_error_0(void *parser, void *trip)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	struct trip_type *trip_data;

	trip_data = trip;

	back_data->code = OPENSIGN;
	back_data->data.opensign.name = trip_data->lead.name;
	back_data->data.opensign.size = trip_data->lead.size;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170621_error_1(void *parser, const struct e4_odbx_type *odbx_data,
			void *trip)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = U_REMOTE;
	back_data->data.u_remote.name = odbx_data->name;
	back_data->data.u_remote.size = odbx_data->size;
	copy_trip(back_data->data.u_remote.post, trip);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170621_error_2(void *parser, const struct e4_odbx_type *odbx_data,
			const struct e4_odbx_type *odbx_text)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = B_REMOTE;
	back_data->data.b_remote.name = odbx_data->name;
	back_data->data.b_remote.size = odbx_data->size;
	copy_trip
	    (back_data->data.b_remote.post,
	     (void *) ((char *) odbx_text - offsetof(trip_type, lead)));
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170621_error_3(void *parser, void *trip, int f3rd)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	const int *args;
	struct trip_type *trip_data;

	trip_data = trip;

	args = trip_data->lead.args;

	back_data->code = TREBLE_R;
	back_data->data.treble_r.f1st = args[0];
	back_data->data.treble_r.f2nd = args[1];
	back_data->data.treble_r.f3rd = f3rd;
	copy_trip(back_data->data.treble_r.post, trip_data);
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170621_error_4(void *parser, void *trip, int f1st, int f2nd)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	struct trip_type *trip_data;

	trip_data = trip;

	back_data->code = SECOND_R;
	back_data->data.second_r.f1st = f1st;
	back_data->data.second_r.f2nd = f2nd;
	back_data->data.second_r.j1st.name = trip_data->lead.name;
	back_data->data.second_r.j1st.size = trip_data->lead.size;
	back_data->data.second_r.j3rd.name = trip_data->rail.name;
	back_data->data.second_r.j3rd.size = trip_data->rail.size;

	if (trip_data->left.bits & POST_MARK) {
	    const struct e4_mark_type *mark_data;

	    mark_data = trip_data->left.post;

	    back_data->data.second_r.j2nd.name = mark_data->part;
	    back_data->data.second_r.j2nd.size = mark_data->size;
	} else {
	    back_data->data.second_r.j2nd.size = 0;
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170811_error_0(void *parser, const struct e4_odbx_type *odbx_data,
			const struct e4_odbx_type *odbx_text)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = T_REMOTE;
	back_data->data.b_remote.name = odbx_data->name;
	back_data->data.b_remote.size = odbx_data->size;
	copy_trip
	    (back_data->data.t_remote.post,
	     (void *) ((char *) odbx_text - offsetof(trip_type, rail)));
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170827_error_0(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	struct link_type *link_data;

	back_data->code = MISSPART;
	link_data = parser(parser)->link_data;
	back_data->data.misspart.name = link_data->part;
	back_data->data.misspart.size = link_data->land;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170926_error_0(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	if (pick_120712_error_0(back_data, parser)) {
	} else {
	    back_data->code = POSTREAL;
	}
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170926_error_1(void *parser)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = ZEROPASS;
    }

    return PARSE_ERROR;
}


int
_x1f4_e4_170926_error_2(void *parser, const void *atom)
{
    struct e4_back_type *back_data;

    back_data = parser(parser)->back_data;
    if (back_data) {
	back_data->code = SAFESTEP;
	type_mind(&back_data->data.safestep.mind, atom, parser);
    }

    return PARSE_ERROR;
}
