/*
 * lxcall-a.8.c
 * Copyright (C) 2011-2014, 2016, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * it may be that _post_call assumes things about the transferring function
 * (function_data argument).  originally called by variadics such as `call',
 * `trap', `xcall', it may still guess the said function has the properties of
 * those.
 */

/*
 * the function must copy with synthetic _walk_link outputs set by clients like
 * call_n to intercept the function call.
 */

/*
 * some (most) callers expect a -dispatch- record after a sequenced call.  if
 * the record is to be optimized away (and the function call to be done
 * immediately), callers expectations are to be addressed.
 */

/*
 * before the first argument is overwritten.  this may be a problem for
 * functions reusing arguments lists and skipping the start (a non zero
 * -offset- is an indication of that).  of those, the trap*s and wrap*s
 * (actually not calling this) do restore the overwritten.  call_n asks
 * (too eagerly perhaps) for a new arguments list.  the remaining -call- does
 * the same for sequenced execution.
 * this function does restore the overwritten for framed execution, does -not-
 * for sequenced execution.
 */

#include <stddef.h>

#include <e4.h>
#include <it.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <lxcast.h>
#include <track.h>
#include <trans.h>

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

extern const void *const x1f4_a1_walk_link;

static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(void *, void *, void **,
		     const struct x1f4_function_type *, unsigned, unsigned,
		     struct x1f4_frame_type *, struct x1f4_track_type *,
		     struct x1f4_linetext_type *);
static int slip_call(void *, void **, void **,
		     const struct x1f4_function_type *, unsigned, unsigned,
		     struct x1f4_frame_type *, struct x1f4_track_type *,
		     struct x1f4_linetext_type *);

static int
miss_call(void *text, int success, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
pick_call(void *output, void *meta, void **track,
	  const struct x1f4_function_type *function_data, unsigned offset,
	  unsigned bits, struct x1f4_frame_type *frame_data,
	  struct x1f4_track_type *track_data,
	  struct x1f4_linetext_type *linetext_data)
{
    int status;
    void *zeta;

    zeta = track[-1];

    track[-1] = meta;

    do {
	int type;

	if (linetext_data->function.flags
	    & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    status = _libx1f4i0_lxcall_pick_call
		(track, function_data, offset, bits, frame_data, track_data,
		 &linetext_data);
	    if (status) {
		break;
	    }
	}

	type = linetext_data->function.type;

	if (type == X1f4_E4_CASE) {
	} else {
	    void *miss;

	    status = frame_data->link
		(frame_data->data, &miss, sizeof(struct x1f4_dxcast_type));
	    if (status) {
		status = _libx1f4i0_lxcall_stat_link(track_data);
		break;
	    } else {
		struct x1f4_dxcast_type *dxcast_data;

		dxcast_data = miss;

		l_USER(miss, output);

		dxcast_data->call = 0;
		dxcast_data->type = type;

		output = &dxcast_data->data;
	    }
	}

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);
    } while (0);

    track[-1] = zeta;

    return status;
}


static int
slip_call(void *output, void **input, void **track,
	  const struct x1f4_function_type *function_data, unsigned offset,
	  unsigned bits, struct x1f4_frame_type *frame_data,
	  struct x1f4_track_type *track_data,
	  struct x1f4_linetext_type *linetext_data)
{
    int status;

    do {
	int type;
	unsigned cast, mind, plot, turn;
	void *miss;

	type = linetext_data->function.type;

	if (type == X1f4_E4_CASE) {
	    cast = 0;
	} else {
	    cast = sizeof(struct x1f4_dxcast_type);
	}

	turn = linetext_data->function.flags;

	if (turn & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    if (bits & POST_LINE) {
		mind = 0;
	    } else {
		mind = sizeof(struct x1f4_linetext_type);
	    }
	    if (bits & POST_ARGS) {
	    } else {
		if (turn & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		    mind += (function_data->count - offset) * sizeof(int) << 1;
		} else {
		    if (linetext_data->function.count) {
			mind += sizeof(int) * (function_data->count - offset);
		    }
		}
	    }
	} else {
	    mind = 0;
	}

	if (bits & POST_PLOT) {
	    plot = 0;
	} else {
	    plot = sizeof(struct x1f4_dispatch_type);
	}

	if (cast | mind | plot) {
	    status = frame_data->link
		(frame_data->data, &miss,
		 sizeof(struct x1f4_dispatch_type) + cast + mind);
	    if (status) {
		status = _libx1f4i0_lxcall_stat_link(track_data);
		break;
	    }
	}
	{
	    struct x1f4_dispatch_type *dispatch_data;
	    void *subtext;

	    if (bits & POST_PLOT) {
		dispatch_data = input[2];
	    } else {
		dispatch_data = miss;

		miss = dispatch_data + 1;
	    }

	    output = input[0];

	    if (cast) {
		struct x1f4_dxcast_type *dxcast_data;

		dxcast_data = (void *) miss;

		l_USER(dxcast_data, output);

		dxcast_data->call = 0;
		dxcast_data->type = type;

		output = &dxcast_data->data;

		miss = dxcast_data + 1;
	    }

	    subtext = (void *) linetext_data->context;

	    if (turn & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
		status = _libx1f4i0_lxcall_slip_call
		    (track, function_data, offset, bits, &linetext_data, miss,
		     &subtext);
		if (status) {
		    break;
		}
	    }

	    dispatch_data->back = NULL;
	    dispatch_data->call = miss_call;

	    dispatch_data->transfer.context = subtext;
	    dispatch_data->transfer.function_data = &linetext_data->function;
	    dispatch_data->transfer.input = track;
	    dispatch_data->transfer.output = output;

	    input[2] = dispatch_data;
	}
    } while (0);

    return status;
}


int
_libx1f4i0_lxcall_post_call(struct x1f4_linetext_type *linetext_data,
			    const struct x1f4_function_type *function_data,
			    void *output, unsigned offset, void **input,
			    unsigned bits, struct x1f4_frame_type *frame_data,
			    struct x1f4_iturn_type *iturn_data,
			    struct x1f4_composer_type *composer_data,
			    struct x1f4_track_type *track_data,
			    const struct x1f4_eelookup_type *eelookup_data)
{
    int linked, status;
    void *meta, **store, **track;

    linked = output == x1f4_a1_walk_link;

    if (linked) {
	track = input[1];
    } else {
	track = input;
    }

    store = track - 1;

    meta = *store;

    store += offset;

    track += offset;

    status = _libx1f4i0_lxcall_even_list
	(&linetext_data->function, function_data, offset, &track, bits,
	 frame_data, iturn_data, composer_data, track_data, eelookup_data);
    if (status) {
    } else {
	if (linked) {
	    track[-1] = meta;

	    status = slip_call
		(output, input, track, function_data, offset, bits, frame_data,
		 track_data, linetext_data);
	} else {
	    status = pick_call
		(output, meta, track, function_data, offset, bits, frame_data,
		 track_data, linetext_data);
	}
    }

    return status;
}
