/*
 * lxcall-e.k.c
 * Copyright (C) 2011-2014, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <e4.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <track.h>
#include <trans.h>

int
_libx1f4i0_lxcall_lock_call(const struct x1f4_function_type *function_data,
			    unsigned offset, unsigned bits,
			    struct x1f4_trans_type *trans_data,
			    struct x1f4_track_type *track_data,
			    struct x1f4_linetext_type **linetext)
{
    int status;

    do {
	int *args;
	struct x1f4_linetext_type *linetext_data, *linetext_slip;
	unsigned black, count, flags, white;
	void *miss;

	linetext_data = *linetext;

	count = function_data->count - offset;

	flags = linetext_data->function.flags;

	/*
	 * the -args- arrays needs to be always copied, as the original may be
	 * destroyed (it will be, if created by another program).
	 * POST_ARGS takes priority.
	 */

	if (bits & POST_LINE) {
	    black = 0;
	} else {
	    black = sizeof(struct x1f4_linetext_type);
	}

	if (bits & POST_ARGS) {
	    white = 0;
	} else {
	    white = count * sizeof(int);
	}

	if (flags & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	    if (function_data->flags & X1f4_E4_POST_TYPE) {
		flags |= X1f4_E4_POST_TYPE;
		white <<= 1;
	    }
	} else {
	}

	if (black | white) {
	    status = trans_data->link(trans_data->data, &miss, black + white);
	    if (status) {
		status = _libx1f4i0_lxcall_stat_link(track_data);
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 0;
	}

	if (black) {
	    linetext_slip = miss;

	    *linetext = linetext_slip;

	    *linetext_slip = *linetext_data;
	} else {
	    linetext_slip = linetext_data;
	}

	if (white) {
	    int *line;
	    const int *plot, *text;
	    unsigned class, i, scale;

	    args = (void *) (black ? linetext_slip + 1 : miss);
	    line = args;
	    plot = linetext_slip->function.args;
	    class = linetext_slip->function.count;
	    scale = count - class;
	    text = function_data->args + offset + class;
	    i = class;
	    for (; i; i--) {
		int slip;

		slip = *plot++;
		*line++ = slip == X1f4_E4_SLIP ? *(text - i) : slip;
	    }
	    memcpy(line, text, scale * sizeof(int));
	    if (flags & X1f4_E4_POST_TYPE) {
		line += scale;
		if (linetext_slip->function.flags & X1f4_E4_POST_TYPE) {
		    memcpy(line, plot, class * sizeof(int));
		} else {
		    memset(line, 0, class * sizeof(int));
		}
		line += class;
		if (function_data->flags & X1f4_E4_POST_TYPE) {
		    memcpy(line, text + count + offset, scale * sizeof(int));
		} else {
		    memset(line, 0, scale * sizeof(int));
		}
	    }
	} else {
	    if (count) {
		int *line;
		const int *plot, *text;
		unsigned class, i, scale;

		args = (int *) function_data->args;
		line = args;
		plot = linetext_slip->function.args;
		class = linetext_slip->function.count;
		scale = count - class;
		text = function_data->args + class;
		i = class;
		for (; i; i--) {
		    int slip;

		    slip = *plot++;
		    *line++ = slip == X1f4_E4_SLIP ? *(text - i) : slip;
		}
		if (flags & X1f4_E4_POST_TYPE) {
		    line += scale;
		    if (linetext_slip->function.flags & X1f4_E4_POST_TYPE) {
			memcpy(line, plot, class * sizeof(int));
		    } else {
			memset(line, 0, class * sizeof(int));
		    }
		}
	    } else {
		args = NULL;
	    }
	}

	flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	linetext_slip->function.args = args;
	linetext_slip->function.count = count;
	linetext_slip->function.flags = flags;
    } while (0);

    return status;
}
