/*
 * lxcase-a.a.c
 * Copyright (C) 2008-2014, 2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <signal.h>

#include <e4.h>
#include <variable.h>

typedef struct constant_type {
    struct x1f4_variable_type variable;
    const void *data;
} constant_type;

static X1f4_E4_C_MODE names[] = {
/* *INDENT-OFF* */
    SIGABRT,
    SIGALRM,
    SIGCHLD,
    SIGCONT,
    SIGFPE,
    SIGHUP,
    SIGILL,
    SIGINT,
    SIGKILL,
    SIGPIPE,
    SIGQUIT,
    SIGSEGV,
    SIGSTOP,
    SIGTERM,
    SIGTSTP,
    SIGTTIN,
    SIGTTOU,
    SIGUSR1,
    SIGUSR2,
/* *INDENT-ON* */
};
static const struct constant_type constants[] = {
/* *INDENT-OFF* */
    {	{	"SIGABRT",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 0							},
    {	{	"SIGALRM",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 1							},
    {	{	"SIGCHLD",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 2							},
    {	{	"SIGCONT",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 3							},
    {	{	"SIGFPE",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	6			},
	names + 4							},
    {	{	"SIGHUP",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	6			},
	names + 5							},
    {	{	"SIGILL",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	6			},
	names + 6							},
    {	{	"SIGINT",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	6			},
	names + 7							},
    {	{	"SIGKILL",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 8							},
    {	{	"SIGPIPE",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 9							},
    {	{	"SIGQUIT",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 10							},
    {	{	"SIGSEGV",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 11							},
    {	{	"SIGSTOP",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 12							},
    {	{	"SIGTERM",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 13							},
    {	{	"SIGTSTP",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 14							},
    {	{	"SIGTTIN",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 15							},
    {	{	"SIGTTOU",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 16							},
    {	{	"SIGUSR1",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 17							},
    {	{	"SIGUSR2",		X1f4_E4_MODE,
		X1f4_E4_READ_ONLY,	7			},
	names + 18							}
/* *INDENT-ON* */
};

int
x1f4_vset_lxcase(void *lxcase, void **context)
{
    int status;
    const struct constant_type *constant_data;
    unsigned i;

    constant_data = constants;

    i = sizeof(constants) / sizeof(struct constant_type);
    for (; i; i--) {
	status = x1f4_push_variable
	    (context, constant_data->variable.name,
	     constant_data->variable.length, constant_data->variable.type,
	     constant_data->variable.flags, (void *) constant_data->data);
	if (status) {
	    break;
	} else {
	    constant_data++;
	}
    }

    return status;
}
