/*
 * lxdash-d.p.c
 * Copyright (C) 2008-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxdash-config.h>

#include <stddef.h>

#include <d_fset.h>
#include <e4.h>
#include <lxdash-drive.h>
#include <lxdash-inter.h>
#include <lxdash-types.h>
#include <lxrail.h>

extern const struct x1f4_bqnorm_type _libx1f4i0_lxdash_u_norm,
    _libx1f4i0_lxdash_v_norm;

int
_libx1f4i0_lxdash_p_derived(void *context, void *output, void **input)
{
    int status;
    struct dxdash_type *dxdash_data;
    struct effect_type effect;
    struct lxdash_type *lxdash_data;
    void *dash;
#if __CLIP_EXTRACT__
#else
    void *data;
#endif				/* __CLIP_EXTRACT__ */

    dxdash_data = I_LINE(input[0]);

    lxdash_data = context;

    dash = dxdash_data->dash;

    effect.mode = I_MODE(input[1]);

#if __CLIP_EXTRACT__
    _libx1f4i0_norm_lxfset(dash, &_libx1f4i0_lxdash_v_norm);
    status = _libx1f4i0_cast_lxfset(dash, &effect);
    _libx1f4i0_norm_lxfset(dash, &_libx1f4i0_lxdash_u_norm);
    if (status) {
	if (status == X1f4_BQFSET_MISS_CLASS) {
	    status = _libx1f4i0_lxdash_stat_miss(lxdash_data, effect.mode);
	} else {
	    status = _libx1f4i0_lxdash_stat_line(lxdash_data, 0, status);
	}
    } else {
	unsigned link;

	dxdash_data->size--;

	link = effect.link;

	status = _libx1f4i0_lxrail_post_data
	    (lxdash_data, output, link, effect.type, &effect.fxdata);
	if (status) {
	} else {
	    if (~link) {
		struct lxtext_type *lxtext_data;

		lxtext_data = lxdash_data->link_f.text;
		lxtext_data += link;
		if (lxtext_data->lxtype.free) {
		    status = lxtext_data->lxtype.free
			(lxtext_data->lxtype.context, I_USER(&effect.fxdata),
			 NULL, lock_excase(dxdash_data));
		} else {
		}
	    }
	}
    }
#else
    status = x1f4_find_lxfset(dash, &effect, &data);
    if (status) {
	struct effect_type *effect_data;
	unsigned link;

	effect_data = data;

	link = effect_data->link;

	status = _libx1f4i0_lxrail_post_data
	    (lxdash_data, output, link, effect_data->type,
	     &effect_data->fxdata);
	if (status) {
	} else {
	    if (~link) {
		struct lxtext_type *lxtext_data;

		effect_data->link = ~0;

		lxtext_data = lxdash_data->link_f.text;
		lxtext_data += link;
		if (lxtext_data->lxtype.free) {
		    status = lxtext_data->lxtype.free
			(lxtext_data->lxtype.context,
			 I_USER(&effect_data->fxdata), NULL,
			 lock_excase(dxdash_data));
		} else {
		}
	    }

	    if (status) {
	    } else {
		status = _libx1f4i0_cast_lxfset(dash, &effect);
		if (status) {
		    status =
			_libx1f4i0_lxdash_stat_line(lxdash_data, 0, status);
		} else {
		    dxdash_data->size--;
		}
	    }
	}
    } else {
	status = _libx1f4i0_lxdash_stat_miss(context, effect.mode);
    }
#endif				/* __CLIP_EXTRACT__ */

    return status;
}
