/*
 * lxdash-i.i.c
 * Copyright (C) 2011-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <d_fset.h>
#include <fd.h>
#include <it.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxdash-drive.h>
#include <lxdash-inter.h>
#include <lxdash-types.h>

#define dxdash(list) \
    ((struct dxdash_type *) (list))
#define lxdash(list) \
    ((struct lxdash_type *) (list))

typedef struct rule_type {
    int roll;
    struct direct_type *direct_data;
    struct excase_type *excase_data;
    struct lxtext_type *lxtext_data;
} rule_type;

extern const void *const x1f4_a1_walk_link;

static int ever_list(struct lxdash_type *, struct direct_type *,
		     struct dxdash_type *, unsigned *);
static int rule_dash(void *, void *);

static int
ever_list(struct lxdash_type *lxdash_data, struct direct_type *direct_data,
	  struct dxdash_type *dxdash_data, unsigned *miss)
{
    int status;
    struct rule_type rule;

    rule.direct_data = direct_data;
    rule.excase_data = lock_excase(dxdash_data);
    rule.lxtext_data = lxdash_data->link_f.text;

    rule.roll = 0;

    if (0) {
    } else {
	status = _libx1f4i0_lime_lxfset(dxdash_data->dash, &rule, rule_dash);
    }

    if (rule.roll) {
	*miss = rule.direct_data - direct_data;
    }

    return status;
}


static int
rule_dash(void *line, void *data)
{
    int link, status;
    struct direct_type *direct_data;
    struct effect_type *effect_data;
    struct rule_type *rule_data;

    rule_data = line;

    direct_data = rule_data->direct_data;

    effect_data = data;

    direct_data->link = effect_data->link;
    direct_data->type = effect_data->type;

    direct_data->fxdata = effect_data->fxdata;

    link = direct_data->link;
    if (~link) {
	struct lxtext_type *lxtext_data;

	lxtext_data = rule_data->lxtext_data + link;

	if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
	    status = lxtext_data->lxtype.node
		(lxtext_data->lxtype.context, (void *) &direct_data->fxdata,
		 I_USER(&direct_data->fxdata), rule_data->excase_data);
	    if (status) {
	    } else {
		rule_data->roll = 1;
		rule_data->direct_data++;
	    }
	} else {
	    int (*copy)(void *, void *, void **, const void *,
			const struct x1f4_nodelink_type *);

	    copy = lxtext_data->lxtype.copy;
	    if (copy) {
		status = copy
		    (lxtext_data->lxtype.context, NULL,
		     (void *) &direct_data->fxdata,
		     I_USER(&direct_data->fxdata), rule_data->excase_data);
		if (status) {
		} else {
		    rule_data->roll = 1;
		    rule_data->direct_data++;
		}
	    } else {
		status = 0;

		rule_data->direct_data++;
	    }
	}
    } else {
	status = 0;

	rule_data->direct_data++;
    }

    return status;
}


int
_libx1f4i0_lxdash_unto_list(void *lxdash, void *output, void **input,
			    void **track,
			    struct x1f4_linetext_type *linetext_data,
			    unsigned count,
			    const struct x1f4_function_type *function_text,
			    int lead, int lock, void *dxdash)
{
    int status;
    void *beta, **mind;

    status = lxdash(lxdash)->link_v.pick
	(lxdash(lxdash)->link_v.data, &mind, _libx1f4i0_lxrail_slip_poll,
	 sizeof(struct affine_type) + count * sizeof(struct direct_type),
	 &beta);
    if (status) {
	status = _libx1f4i0_lxdash_stat_link(lxdash);
    } else {
	struct affine_type *affine_data;
	struct direct_type *direct_data;

	*mind = beta;

	affine_data = beta;

	direct_data = (void *) (affine_data + 1);

	affine_data->miss = 0;
	affine_data->size = count;

	affine_data->nodelink_data = lock_excase(dxdash);
	affine_data->lxtext_data = lxdash(lxdash)->link_f.text;

	status = ever_list(lxdash, direct_data, dxdash, &affine_data->miss);
	if (status) {
	} else {
	    if (output == x1f4_a1_walk_link) {
		status = _libx1f4i0_lxrail_lift_call
		    (lxdash, output, input, linetext_data, count,
		     function_text, lead, lock, direct_data);
	    } else {
		status = _libx1f4i0_lxrail_fold_call
		    (lxdash, output, input, linetext_data, count,
		     function_text, lead, lock, direct_data);
	    }
	}
    }

    return status;
}
