/*
 * lxdata-a.0.c
 * Copyright (C) 2008-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <e4-m.0.h>
#include <e4.h>
#include <lxcall.h>
#include <lxdata-defs.h>
#include <lxdata-inter.h>
#include <lxdata-types.h>

#define MAKE_SINGLE(a, b)		a

#define fb_size() \
    (sizeof(side_o) / sizeof(struct x1f4_operator_type))
#define fp_size() \
    (sizeof(side_u) / sizeof(struct x1f4_operator_type))

#define lxdata(lxdata) \
    ((struct lxdata_type *) (lxdata))

#define screen(screen) \
    ((struct screen_type *) (screen))

#define f1_size(year) \
    (year)

#define fa_size(year) \
    ((year) * sizeof(int))

#define ff_size(year) \
    ((year) * sizeof(struct x1f4_linetext_type))

static int init_data(struct lxdata_type *, unsigned,
		     const struct screen_type *);
static int line_data(struct lxdata_type *, unsigned,
		     const struct screen_type *);
static int line_rail(struct ticket_type *, struct x1f4_operator_type *,
		     struct x1f4_2ndscope_type *, struct x1f4_linetext_type *);
static int line_text(struct lxdata_type *, unsigned,
		     const struct screen_type *, void *, void **, void **);
static int link_data(struct lxdata_type *, unsigned,
		     const struct screen_type *);

static void bo_mlmx0x(u_case_args_____0);
static void borne_lml(u_case_args_____0);
static void borne_lmm(u_case_args_____0);
static void borne_lmt(u_case_args_____0);
static void borneo__l(u_case_args_____0);
static void borneo_lb(u_case_args_____0);
static void borneo_ll(u_case_args_____0);
static void borneo_lm(u_case_args_____0);
static void borneo_lt(u_case_args_____0);
static void line_bfix(u_case_args_____0);
static void line_cset(u_case_args_____0);
static void line_dmap(u_case_args_____0);
static void line_mmap(u_case_args_____0);
static void line_mset(u_case_args_____0);
static void line_nset(u_case_args_____0);
static void line_rset(u_case_args_____0);
static void line_sfar(u_case_args_____0);
static void line_sfix(u_case_args_____0);
static void line_smap(u_case_args_____0);
static void line_tway(u_case_args_____0);
static void line_yfix(u_case_args_____0);
static void theme_lmm(u_case_args_____0);
static void themed_ll(u_case_args_____0);
static void themed_lm(u_case_args_____0);
static void themed_lt(u_case_args_____0);
static void themed__l(u_case_args_____0);

static const char fb_castf[] = "b_cast", fb_chance1f[] = "b_chance1",
    fb_characterf[] = "b_character", fb_lengthf[] = "b_length", fb_moldf[] =
    "b_mold", fb_replacef[] = "b_replace", fb_setf[] = "b_set",
    *const side_f[] = {
/* *INDENT-OFF* */
    fb_castf,
    fb_setf,
    fb_moldf,
    fb_chance1f,
    fb_characterf
/* *INDENT-ON* */
}, *const side_l[] = {
/* *INDENT-OFF* */
    fb_lengthf
/* *INDENT-ON* */
};
static const struct line_type auto_line[] = {
/* *INDENT-OFF* */
#define b_cut_count			3
    {	{	"b_cut",
		_libx1f4i0_lxdata_selogique,	X1f4_E4_TEXT,
		NULL,				b_cut_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						5		},
	borne_lmm,							},
#define b_cut_reach			(0 + be_string_count)
#define b_draft_count			1
    {	{	"b_draft",
		_libx1f4i0_lxdata_d_pending,	0,
		NULL,				b_draft_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	line_dmap,							},
#define b_draft_reach			(b_cut_reach + b_draft_count)
#define b_mold_count			2
    {	{	fb_moldf,
		_libx1f4i0_lxdata_m_pending,	X1f4_E4_TEXT,
		NULL,				b_mold_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						6		},
	line_mmap,							},
#define b_mold_reach			(b_draft_reach + (b_mold_count << 1))
#define b_project_count			2
    {	{	"b_project",
		_libx1f4i0_lxdata_splogique,	X1f4_E4_TEXT,
		NULL,				b_project_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borneo_lm,							},
#define b_project_reach			(b_mold_reach + b_project_count)
#define b_string_count			1
    {	{	"b_string",
		_libx1f4i0_lxdata_s_logique,	X1f4_E4_TEXT,
		NULL,				b_string_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	line_sfix,							},
#define b_string_reach			(b_project_reach + b_string_count)
#define be_string_count			3
    {	{	"be_string",
		_libx1f4i0_lxdata_selogique,	X1f4_E4_TEXT,
		NULL,				be_string_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borne_lmm,							},
#define be_string_reach			(b_string_reach + be_string_count)
#define bp_string_count			2
    {	{	"bp_string",
		_libx1f4i0_lxdata_splogique,	X1f4_E4_TEXT,
		NULL,				bp_string_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borneo_lm,							},
#define bp_string_reach			(be_string_reach + bp_string_count)
#define bq_string_count			3
    {	{	"bq_string",
		_libx1f4i0_lxdata_sqlogique,	X1f4_E4_TEXT,
		NULL,				bq_string_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borne_lmm,							},
#define bq_string_reach			(bp_string_reach + bq_string_count)
#define data_count			0
    {	{	"data",
		_libx1f4i0_lxdata_y_logique,	0,
		NULL,				data_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						4		},
	line_yfix,							}
#define data_reach			(bq_string_reach + data_count)
/* *INDENT-ON* */
#define auto_line_count			data_reach
}, call_line[] = {
/* *INDENT-OFF* */
#define b_new_count			1
    {	{	"b_new",
		_libx1f4i0_lxdata_n_forward,	X1f4_E4_VOID,
		NULL,				b_new_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						5		},
	line_nset,							}
#define b_new_reach			(0 + (b_new_count << 1))
/* *INDENT-ON* */
#define call_line_count			b_new_reach
}, ever_line[] = {
/* *INDENT-OFF* */
#define b_set_count			2
    {	{	fb_setf,
		_libx1f4i0_lxdata_s_pending,	0,
		NULL,				b_set_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						5		},
	line_smap,							}
#define b_set_reach			(0 + (b_set_count << 1))
/* *INDENT-ON* */
#define ever_line_count			b_set_reach
}, post_line[] = {
/* *INDENT-OFF* */
#define b_chance1_count			2
    {	{	fb_chance1f,
		_libx1f4i0_lxdata_t_logique,	X1f4_E4_MODE,
		NULL,				b_chance1_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borneo_lb,							}
#define b_chance1_reach			(0 + b_chance1_count)
/* *INDENT-ON* */
#define post_line_count			b_chance1_reach
}, slip_line[] = {
/* *INDENT-OFF* */
#define b_add_count			2
    {	{	"b_add",
		_libx1f4i0_lxdata_a_forward,	0,
		NULL,				b_add_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						5		},
	themed_ll,							},
#define b_add_reach			(0 + b_add_count)
#define b_append_count			2
    {	{	"b_append",
		_libx1f4i0_lxdata_a_logique,	X1f4_E4_VOID,
		NULL,				b_append_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	themed_lm,							},
#define b_append_reach			(b_add_reach + b_append_count)
#define b_cast_count			2
    {	{	fb_castf,
		_libx1f4i0_lxdata_c_forward,	0,
		NULL,				b_cast_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	line_cset,							},
#define b_cast_reach			(b_append_reach + b_cast_count)
#define b_character_count		2
    {	{	fb_characterf,
		_libx1f4i0_lxdata_t_logique,	X1f4_E4_MODE,
		NULL,				b_character_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						11		},
	borneo_lm,							},
#define b_character_reach		(b_cast_reach + b_character_count)
#define b_clear_count			1
    {	{	"b_clear",
		_libx1f4i0_lxdata_c_logique,	X1f4_E4_VOID,
		NULL,				b_clear_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	themed__l,							},
#define b_clear_reach			(b_character_reach + b_clear_count)
#define b_compare_count			2
    {	{	"b_compare",
		_libx1f4i0_lxdata_c_derived,	X1f4_E4_MODE,
		NULL,				b_compare_count,
		0,				9		},
	borneo_ll,							},
#define b_compare_reach			(b_clear_reach + b_compare_count)
#define b_copy_count			2
    {	{	"b_copy",
		_libx1f4i0_lxdata_c_pending,	0,
		NULL,				b_copy_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	themed_ll,							},
#define b_copy_reach			(b_compare_reach + b_copy_count)
#define b_delete_count			2
    {	{	"b_delete",
		_libx1f4i0_lxdata_d_logique,	0,
		NULL,				b_delete_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	themed_lm,							},
#define b_delete_reach			(b_copy_reach + b_delete_count)
#define b_drop_count			2
    {	{	"b_drop",
		_libx1f4i0_lxdata_d_centric,	0,
		NULL,				b_drop_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	themed_lt,							},
#define b_drop_reach			(b_delete_reach + b_drop_count)
#define b_dump_count			2
    {	{	"b_dump",
		_libx1f4i0_lxdata_d_forward,	0,
		NULL,				b_dump_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	themed_lm,							},
#define b_dump_reach			(b_drop_reach + b_dump_count)
#define b_ecopy_count			4
    {	{	"b_ecopy",
		_libx1f4i0_lxdata_b_logique,	0,
		NULL,				b_ecopy_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	line_bfix,							},
#define b_ecopy_reach			(b_dump_reach + b_ecopy_count)
#define b_edit_count			3
    {	{	"b_edit",
		_libx1f4i0_lxdata_e_pending,	0,
		NULL,				b_edit_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	theme_lmm,							},
#define b_edit_reach			(b_ecopy_reach + b_edit_count)
#define b_erase_count			3
    {	{	"b_erase",
		_libx1f4i0_lxdata_e_logique,	0,
		NULL,				b_erase_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	theme_lmm,							},
#define b_erase_reach			(b_edit_reach + b_erase_count)
#define b_even_count			3
    {	{	"b_even",
		_libx1f4i0_lxdata_e_derived,	X1f4_E4_MODE,
		NULL,				b_even_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	borne_lmt,							},
#define b_even_reach			(b_erase_reach + b_even_count)
#define b_extend_count			2
    {	{	"b_extend",
		_libx1f4i0_lxdata_a_forward,	0,
		NULL,				b_extend_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	themed_ll,							},
#define b_extend_reach			(b_even_reach + b_extend_count)
#define b_find_count			2
    {	{	"b_find",
		_libx1f4i0_lxdata_f_derived,	X1f4_E4_MODE,
		NULL,				b_find_count,
		0,				6		},
	borneo_ll,							},
#define b_find_reach			(b_extend_reach + b_find_count)
#define b_fit_count			2
    {	{	"b_fit",
		_libx1f4i0_lxdata_f_logique,	X1f4_E4_VOID,
		NULL,				b_fit_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						5		},
	borneo_lt,							},
#define b_fit_reach			(b_find_reach + b_fit_count)
#define b_frame_count			2
    {	{	"b_frame",
		_libx1f4i0_lxdata_f_forward,	X1f4_E4_MODE,
		NULL,				b_frame_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borneo_lm,							},
#define b_frame_reach			(b_fit_reach + b_frame_count)
#define b_index_count			2
    {	{	"b_index",
		_libx1f4i0_lxdata_i_forward,	X1f4_E4_MODE,
		NULL,				b_index_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borneo_lm,							},
#define b_index_reach			(b_frame_reach + b_index_count)
#define b_initial_count			1
    {	{	"b_initial",
		_libx1f4i0_lxdata_i_derived,	X1f4_E4_MODE,
		NULL,				b_initial_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borneo__l,							},
#define b_initial_reach			(b_index_reach + b_initial_count)
#define b_insert_count			3
    {	{	"b_insert",
		_libx1f4i0_lxdata_i_logique,	X1f4_E4_VOID,
		NULL,				b_insert_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	theme_lmm,							},
#define b_insert_reach			(b_initial_reach + b_insert_count)
#define b_length_count			1
    {	{	fb_lengthf,
		_libx1f4i0_lxdata_l_logique,	X1f4_E4_MODE,
		NULL,				b_length_count,
		0,				8		},
	borneo__l,							},
#define b_length_reach			(b_insert_reach + b_length_count)
#define b_look_count			3
    {	{	"b_look",
		_libx1f4i0_lxdata_l_pending,	X1f4_E4_MODE,
		NULL,				b_look_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	borne_lmt,							},
#define b_look_reach			(b_length_reach + b_look_count)
#define b_msbfn_count			3
    {	{	"b_msbfn",
		_libx1f4i0_lxdata_m_logique,	X1f4_E4_BILL,
		NULL,				b_msbfn_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borne_lmm,							},
#define b_msbfn_reach			(b_look_reach + b_msbfn_count)
#define b_msbfvn_count			2
    {	{	"b_msbfvn",
		_libx1f4i0_lxdata_m_forward,	X1f4_E4_BILL,
		NULL,				b_msbfvn_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						8		},
	line_mset,							},
#define b_msbfvn_reach			(b_msbfn_reach + (b_msbfvn_count << 1))
#define b_paste_count			3
    {	{	"b_paste",
		_libx1f4i0_lxdata_p_logique,	X1f4_E4_VOID,
		NULL,				b_paste_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borne_lmt,							},
#define b_paste_reach			(b_msbfvn_reach + b_paste_count)
#define b_patch_count			3
    {	{	"b_patch",
		_libx1f4i0_lxdata_p_forward,	X1f4_E4_VOID,
		NULL,				b_patch_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borne_lmt,							},
#define b_patch_reach			(b_paste_reach + b_patch_count)
#define b_place_count			2
    {	{	"b_place",
		_libx1f4i0_lxdata_i_forward,	X1f4_E4_MODE,
		NULL,				b_place_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borneo_lm,							},
#define b_place_reach			(b_patch_reach + b_place_count)
#define b_prefix_count			2
    {	{	"b_prefix",
		_libx1f4i0_lxdata_p_derived,	X1f4_E4_MODE,
		NULL,				b_prefix_count,
		0,				8		},
	borneo_ll,							},
#define b_prefix_reach			(b_place_reach + b_prefix_count)
#define b_probe_count			3
    {	{	"b_probe",
		_libx1f4i0_lxdata_p_pending,	X1f4_E4_MODE,
		NULL,				b_probe_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borne_lmm,							},
#define b_probe_reach			(b_prefix_reach + b_probe_count)
#define b_rate_count			3
    {	{	"b_rate",
		_libx1f4i0_lxdata_r_sequent,	X1f4_E4_MODE,
		NULL,				b_rate_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	borne_lmt,							},
#define b_rate_reach			(b_probe_reach + b_rate_count)
#define b_replace_count			3
    {	{	fb_replacef,
		_libx1f4i0_lxdata_r_logique,	X1f4_E4_MODE,
		NULL,				b_replace_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	borne_lmm,							},
#define b_replace_reach			(b_rate_reach + b_replace_count)
#define b_reverse_count			1
    {	{	"b_reverse",
		_libx1f4i0_lxdata_r_derived,	0,
		NULL,				b_reverse_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	themed__l,							},
#define b_reverse_reach			(b_replace_reach + b_reverse_count)
#define b_rule_count			5
    {	{	"b_rule",
		_libx1f4i0_lxdata_r_forward,	0,
		NULL,				b_rule_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	line_rset,							},
#define b_rule_reach			(b_reverse_reach + b_rule_count)
#define b_scan_count			3
    {	{	"b_scan",
		_libx1f4i0_lxdata_s_generic,	X1f4_E4_MODE,
		NULL,				b_scan_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	borne_lml,							},
#define b_scan_reach			(b_rule_reach + b_scan_count)
#define b_seek_count			2
    {	{	"b_seek",
		_libx1f4i0_lxdata_s_ongoing,	X1f4_E4_MODE,
		NULL,				b_seek_count,
		0,				6		},
	borneo_lt,							},
#define b_seek_reach			(b_scan_reach + b_seek_count)
#define b_side_count			3
    {	{	"b_side",
		_libx1f4i0_lxdata_s_harsher,	X1f4_E4_MODE,
		NULL,				b_side_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	borne_lmt,							},
#define b_side_reach			(b_seek_reach + b_side_count)
#define b_size_count			2
    {	{	"b_size",
		_libx1f4i0_lxdata_s_derived,	X1f4_E4_VOID,
		NULL,				b_size_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	themed_lm,							},
#define b_size_reach			(b_side_reach + b_size_count)
#define b_squeeze_count			2
    {	{	"b_squeeze",
		_libx1f4i0_lxdata_q_centric,	0,
		NULL,				b_squeeze_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						9		},
	themed_lt,							},
#define b_squeeze_reach			(b_size_reach + b_squeeze_count)
#define b_stock_count			3
    {	{	"b_stock",
		_libx1f4i0_lxdata_s_sequent,	0,
		NULL,				b_stock_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	line_sfar,							},
#define b_stock_reach			(b_squeeze_reach + b_stock_count)
#define b_suffix_count			2
    {	{	"b_suffix",
		_libx1f4i0_lxdata_f_logique,	X1f4_E4_VOID,
		NULL,				b_suffix_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	borneo_lt,							},
#define b_suffix_reach			(b_stock_reach + b_suffix_count)
#define b_swap_count			2
    {	{	"b_swap",
		_libx1f4i0_lxdata_s_forward,	X1f4_E4_VOID,
		NULL,				b_swap_count,
		X1f4_E4_KEEP_CALL,		6		},
	borneo_ll,							},
#define b_swap_reach			(b_suffix_reach + b_swap_count)
#define b_trace_count			3
    {	{	"b_trace",
		_libx1f4i0_lxdata_t_forward,	X1f4_E4_MODE,
		NULL,				b_trace_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borne_lmm,							},
#define b_trace_reach			(b_swap_reach + b_trace_count)
#define b_translate_count		3
    {	{	"b_translate",
		_libx1f4i0_lxdata_t_derived,	0,
		NULL,				b_translate_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						11		},
	line_tway,							},
#define b_translate_reach \
    (b_trace_reach + b_translate_count)
#define b_turn_count			3
    {	{	"b_turn",
		_libx1f4i0_lxdata_t_pending,	0,
		NULL,				b_turn_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						6		},
	theme_lmm,							},
#define b_turn_reach			(b_translate_reach + b_turn_count)
#define bb_cast_count			2
    {	{	"bb_cast",
		_libx1f4i0_lxdata_f_logique,	X1f4_E4_VOID,
		NULL,				bb_cast_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	themed_lt,							},
#define bb_cast_reach			(b_turn_reach + bb_cast_count)
#define bb_copy_count			2
    {	{	"bb_copy",
		_libx1f4i0_lxdata_a_forward,	0,
		NULL,				bb_copy_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	themed_ll,							},
#define bb_copy_reach			(bb_cast_reach + bb_copy_count)
#define bb_drop_count			2
    {	{	"bb_drop",
		_libx1f4i0_lxdata_dbcentric,	0,
		NULL,				bb_drop_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	themed_lt,							},
#define bb_drop_reach			(bb_copy_reach + bb_drop_count)
#define bb_drop0_count			2
    {	{	"bb_drop0",
		_libx1f4i0_lxdata_zbcentric,	0,
		NULL,				bb_drop0_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	themed_lt,							},
#define bb_drop0_reach			(bb_drop_reach + bb_drop0_count)
#define bf_drop_count			2
    {	{	"bf_drop",
		_libx1f4i0_lxdata_dfcentric,	0,
		NULL,				bf_drop_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	themed_lt,							},
#define bf_drop_reach			(bb_drop0_reach + bf_drop_count)
#define bf_drop0_count			2
    {	{	"bf_drop0",
		_libx1f4i0_lxdata_zfcentric,	0,
		NULL,				bf_drop0_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						8		},
	themed_lt,							},
#define bf_drop0_reach			(bf_drop_reach + bf_drop0_count)
#define bp_compare_count		3
    {	{	"bp_compare",
		_libx1f4i0_lxdata_cpderived,	X1f4_E4_MODE,
		NULL,				bp_compare_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						10		},
	borne_lml,							},
#define bp_compare_reach		(bf_drop0_reach + bp_compare_count)
#define bt_first_count			3
    {	{	"bt_first",
		_libx1f4i0_lxdata_f_sequent,	X1f4_E4_MODE,
		NULL,				bt_first_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						8		},
	bo_mlmx0x,							},
#define bt_first_reach \
    (bp_compare_reach + (bt_first_count << 1))
#define bt_last_count			3
    {	{	"bt_last",
		_libx1f4i0_lxdata_l_sequent,	X1f4_E4_MODE,
		NULL,				bt_last_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						7		},
	bo_mlmx0x,							},
#define bt_last_reach			(bt_first_reach + (bt_last_count << 1))
#define bt_next_count			3
    {	{	"bt_next",
		_libx1f4i0_lxdata_n_sequent,	X1f4_E4_MODE,
		NULL,				bt_next_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						7		},
	bo_mlmx0x,							},
#define bt_next_reach			(bt_last_reach + (bt_next_count << 1))
#define bt_previous_count		3
    {	{	"bt_previous",
		_libx1f4i0_lxdata_p_sequent,	X1f4_E4_MODE,
		NULL,				bt_previous_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						11		},
	bo_mlmx0x,							},
#define bt_previous_reach \
    (bt_next_reach + (bt_previous_count << 1))
#define bu_cast_count			3
    {	{	"bu_cast",
		_libx1f4i0_lxdata_p_logique,	X1f4_E4_VOID,
		NULL,				bu_cast_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	borne_lmt,							},
#define bu_cast_reach			(bt_previous_reach + bu_cast_count)
#define bu_copy_count			3
    {	{	"bu_copy",
		_libx1f4i0_lxdata_s_sequent,	0,
		NULL,				bu_copy_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK,
						7		},
	line_sfar,							}
#define bu_copy_reach			(bu_cast_reach + bu_copy_count)
/* *INDENT-ON* */
#define slip_line_count			bu_copy_reach
};
static const struct x1f4_postmark_type side_k[] = {
/* *INDENT-OFF* */
    {	MAKE_SINGLE("]", " "),	1				}
/* *INDENT-ON* */
};
static const struct x1f4_operator_type rail_f[] = {
/* *INDENT-OFF* */
    {	MAKE_SINGLE("=", " "),  NULL,		0400,
	0,			NULL,
	X1f4_E4_BACK_LINK | X1f4_E4_E2ND_LINK,
				1,
	NULL,			NULL					}
/* *INDENT-ON* */
}, side_o[] = {
/* *INDENT-OFF* */
    {	MAKE_SINGLE("=", " "),  NULL,		0400,
	0,			NULL,
	X1f4_E4_BACK_LINK | X1f4_E4_E2ND_LINK | X1f4_E4_LEFT_XSET,
				1,
	NULL,			NULL					},
    {	MAKE_SINGLE("=", " "),  NULL,		0400,
	0,			NULL,
	X1f4_E4_BACK_LINK | X1f4_E4_E2ND_LINK | X1f4_E4_LEFT_XSET,
				1,
	NULL,			NULL					},
    {	MAKE_SINGLE("=", " "),  NULL,		0400,
	0,			NULL,
	X1f4_E4_BACK_LINK | X1f4_E4_E2ND_LINK | X1f4_E4_LEFT_XSET,
				1,
	NULL,			NULL					},
    {	MAKE_SINGLE("[", " "),  NULL,		0200,
	0,			NULL,
	X1f4_E4_E2ND_LINK,	1,
	NULL,			NULL					},
    {	MAKE_SINGLE("[", " "),  NULL,		0200,
	0,			NULL,
	X1f4_E4_E2ND_LINK,	1,
	NULL,			NULL					}
/* *INDENT-ON* */
}, side_u[] = {
/* *INDENT-OFF* */
    {	MAKE_SINGLE("~", " "),  NULL,		0210,
	0,			NULL,
	X1f4_E4_E2ND_LINK,
				1,
	NULL,			NULL					}
/* *INDENT-ON* */
};

static int
init_data(struct lxdata_type *lxdata_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    int status;

    if (bits & AUTOLINK_LINK) {
	lxdata_data->link_v = screen_data->lock_v;
    } else {
	lxdata_data->link_v.data = NULL;
	lxdata_data->link_v.free = NULL;
	lxdata_data->link_v.link = NULL;
	lxdata_data->link_v.mode = NULL;
	lxdata_data->link_v.pick = NULL;
	lxdata_data->link_v.slip = NULL;
    }

    if (bits & CODELINK_LINK) {
	lxdata_data->link_w = screen_data->lock_w;
    } else {
	lxdata_data->link_w.data = NULL;
	lxdata_data->link_w.free = _x1f4_e4_free_data;
	lxdata_data->link_w.link = _x1f4_e4_link_data;
	lxdata_data->link_w.mode = _x1f4_e4_mode_data;
    }

    if (bits & RESOURCE_LINK) {
	lxdata_data->link_m = screen_data->lock_m;
    } else {
	lxdata_data->link_m.data = NULL;
	lxdata_data->link_m.free = _x1f4_e4_free_data;
	lxdata_data->link_m.link = _x1f4_e4_link_data;
	lxdata_data->link_m.mode = _x1f4_e4_mode_data;
    }

    if (bits & TEXTFLAT_LINK) {
	lxdata_data->link_e = screen_data->lock_e;
    } else {
	lxdata_data->link_e.data = NULL;
	lxdata_data->link_e.line = NULL;
	lxdata_data->link_e.post = NULL;
	lxdata_data->link_e.push = NULL;
    }

    status = link_data(lxdata_data, bits, screen_data);

    return status;
}


static int
line_data(struct lxdata_type *lxdata_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    lxdata_data->link_t.datatype[0].context = lxdata_data;

    lxdata_data->link_t.datatype[0].flat = _libx1f4i0_lxdata_flat_data;
    lxdata_data->link_t.datatype[0].lead = _libx1f4i0_lxdata_lead_data;
    lxdata_data->link_t.datatype[0].line = _libx1f4i0_lxdata_line_data;
    lxdata_data->link_t.datatype[0].link = _libx1f4i0_lxdata_link_data;
    lxdata_data->link_t.datatype[0].name = "data";
    lxdata_data->link_t.datatype[0].slip = _libx1f4i0_lxdata_slip_data;
    lxdata_data->link_t.datatype[0].size = 4;
    lxdata_data->link_t.datatype[0].type = screen_data->link_l.type;

    lxdata_data->link_t.datatype[1].context = NULL;

    lxdata_data->link_t.datatype[1].flat = NULL;
    lxdata_data->link_t.datatype[1].lead = NULL;
    lxdata_data->link_t.datatype[1].line = NULL;
    lxdata_data->link_t.datatype[1].link = NULL;
    lxdata_data->link_t.datatype[1].name = NULL;
    lxdata_data->link_t.datatype[1].slip = NULL;
    lxdata_data->link_t.datatype[1].size = 0;
    lxdata_data->link_t.datatype[1].type = 0;

    return 0;
}



static int
line_rail(struct ticket_type *ticket_data,
	  struct x1f4_operator_type *operator_rail,
	  struct x1f4_2ndscope_type *scope2nd_rail,
	  struct x1f4_linetext_type *linetext_data)
{
    int i = 2;

    while (linetext_data->function.name != fb_replacef) {
	linetext_data++;
    }

    for (; i; i--) {
	struct x1f4_2ndscope_type *scope2nd_data;

	ticket_data--;
	scope2nd_data = &ticket_data->tidespan;
	scope2nd_data->bits |= X1f4_E4_POST_MARK | X1f4_E4_RAIL_MARK;
	scope2nd_data->post = side_k;
	scope2nd_data->rail = operator_rail;

	*operator_rail = *rail_f;
	operator_rail->args = linetext_data->function.args + 1;
	operator_rail->type = X1f4_E4_MODE;
	operator_rail->extension2 = scope2nd_rail;
	scope2nd_rail->bits = 0;
	scope2nd_rail->line = linetext_data;

	operator_rail++;
	scope2nd_rail++;
    }

    return 0;
}


static int
line_text(struct lxdata_type *lxdata_data, unsigned bits,
	  const struct screen_type *screen_data, void *data, void **args,
	  void **dana)
{
    struct x1f4_linetext_type *linetext_data;

    linetext_data = data;

    if (1) {
	_libx1f4i0_lxcall_side_line
	    (lxdata_data, bits, screen_data, &linetext_data, args,
	     sizeof(slip_line) / sizeof(struct line_type), slip_line);
    }
    if (1) {
	_libx1f4i0_lxcall_side_line
	    (lxdata_data, bits, screen_data, &linetext_data, args,
	     sizeof(call_line) / sizeof(struct line_type), call_line);
    }
    if (bits & AUTOLINK_LINK) {
	_libx1f4i0_lxcall_side_line
	    (lxdata_data, bits, screen_data, &linetext_data, args,
	     sizeof(auto_line) / sizeof(struct line_type), auto_line);
	_libx1f4i0_lxcall_side_line
	    (lxdata_data, bits, screen_data, &linetext_data, args,
	     sizeof(ever_line) / sizeof(struct line_type), ever_line);
    } else {
	lxdata_data->late_f.b_draft = NULL;
	lxdata_data->late_f.b_string = NULL;
    }
    if (1) {
	_libx1f4i0_lxcall_zero_line(linetext_data);
    }
    if (1) {
	linetext_data++;
    }
    if (1) {
	_libx1f4i0_lxcall_side_line
	    (lxdata_data, bits, screen_data, &linetext_data, args,
	     sizeof(post_line) / sizeof(struct line_type), post_line);
    }
    if (1) {
	_libx1f4i0_lxcall_zero_line(linetext_data);
    }

    return 0;
}


static int
link_data(struct lxdata_type *lxdata_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    int status;
    unsigned args_count, line_count;
    void *data;

    args_count = 0;
    line_count = 2;

    {
	args_count += slip_line_count;
	line_count += sizeof(slip_line) / sizeof(struct line_type);
    }
    if (1) {
	args_count += call_line_count;
	line_count += sizeof(call_line) / sizeof(struct line_type);
    }
    if (bits & AUTOLINK_LINK) {
	args_count += auto_line_count;
	line_count += sizeof(auto_line) / sizeof(struct line_type);
	args_count += ever_line_count;
	line_count += sizeof(ever_line) / sizeof(struct line_type);
    }
    {
	args_count += post_line_count;
	line_count += sizeof(post_line) / sizeof(struct line_type);
    }

    status = lxdata_data->link_w.link
	(lxdata_data->link_w.data, &data,
	 args_count * sizeof(int)
	 + line_count * sizeof(struct x1f4_linetext_type)
	 + (sizeof(struct ticket_type) + sizeof(struct x1f4_operator_type *))
	 * (fb_size() + fp_size())
	 + (sizeof(struct x1f4_2ndscope_type)
	    + sizeof(struct x1f4_operator_type)) * 2
	 + sizeof(struct x1f4_operator_type *) * 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	void *args, *dana, *part, *rail, *rate, *side;

	part = (struct x1f4_linetext_type *) data + line_count;
	rail = (struct x1f4_2ndscope_type *) part + 2;
	side = (struct x1f4_operator_type *) rail + 2;
	side = (struct ticket_type *) side + fb_size();
	rate = (struct x1f4_operator_type **) side + fb_size() + 1;
	rate = (struct ticket_type *) rate + fp_size();
	args = (struct x1f4_operator_type **) rate + fp_size() + 1;

	lxdata_data->link_f.data = data;
	lxdata_data->link_f.rate = rate;
	lxdata_data->link_f.side = side;

	line_text(lxdata_data, bits, screen_data, data, &args, &dana);
	_libx1f4i0_lxcall_line_side(rate, data, side_u, side_l, fp_size());
	_libx1f4i0_lxcall_line_vice
	    (side, data,
	     (struct x1f4_linetext_type *) data + line_count - 1
	     - sizeof(post_line) / sizeof(struct line_type), side_o, side_f,
	     fb_size());

	line_data(lxdata_data, bits, screen_data);

	line_rail(side, rail, part, data);
    }

    return status;
}


static void
bo_mlmx0x(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = X1f4_E4_MODE;
    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_POST_XSET;
    *line++ = 0;
    *line++ = X1f4_E4_POST_XSET;

    *args = line;
}


static void
borne_lml(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_MODE;
    *line++ = type;

    *args = line;
}


static void
borne_lmm(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_MODE;

    *args = line;
}


static void
borne_lmt(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_TEXT;

    *args = line;
}


static void
borneo__l(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;

    *args = line;
}


static void
borneo_lb(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_BILL;

    *args = line;
}


static void
borneo_ll(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = type;

    *args = line;
}


static void
borneo_lm(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_MODE;

    *args = line;
}


static void
borneo_lt(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_TEXT;

    *args = line;
}


static void
line_bfix(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = type;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_MODE;

    *args = line;

    linetext_data->function.type = type;
}


static void
line_cset(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_TEXT;

    *args = line;

    linetext_data->function.type = screen(screen)->link_l.type;
}


static void
line_dmap(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = X1f4_E4_TEXT;

    *args = line;

    linetext_data->function.type = screen(screen)->link_l.type;

    lxdata(linetext_data->context)->late_f.b_draft = linetext_data;
}


static void
line_mmap(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = X1f4_E4_TEXT;
    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_POST_XSET;
    *line++ = 0;

    *args = line;
}


static void
line_mset(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_MODE;
    *line++ = 0;
    *line++ = X1f4_E4_POST_XSET;

    *args = line;
}


static void
line_nset(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_POST_XSET;

    *args = line;
}


static void
line_rset(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_MODE;
    *line++ = type;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_MODE;

    *args = line;

    linetext_data->function.type = type;
}


static void
line_sfar(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_MODE;
    *line++ = type;

    *args = line;

    linetext_data->function.type = type;
}


static void
line_sfix(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;

    *args = line;

    lxdata(linetext_data->context)->late_f.b_string = linetext_data;
}


static void
line_smap(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_l.type;
    *line++ = screen(screen)->link_l.type;
    *line++ = X1f4_E4_POST_FLAT | X1f4_E4_POST_XSET;
    *line++ = 0;

    *args = line;

    linetext_data->function.type = screen(screen)->link_l.type;
}


static void
line_tway(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_TEXT;
    *line++ = X1f4_E4_TEXT;

    *args = line;

    linetext_data->function.type = type;
}


static void
line_yfix(u_case_args_____1)
{
    linetext_data->function.type = screen(screen)->link_l.type;
}


static void
theme_lmm(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_MODE;

    *args = line;

    linetext_data->function.type = type;
}


static void
themed_ll(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = type;

    *args = line;

    linetext_data->function.type = type;
}


static void
themed_lm(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_MODE;

    *args = line;

    linetext_data->function.type = type;
}


static void
themed_lt(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;
    *line++ = X1f4_E4_TEXT;

    *args = line;

    linetext_data->function.type = type;
}


static void
themed__l(u_case_args_____1)
{
    int *line, type;

    type = screen(screen)->link_l.type;

    line = *args;

    linetext_data->function.args = line;

    *line++ = type;

    *args = line;

    linetext_data->function.type = type;
}


int
x1f4_init_lxdata(void **lxdata, unsigned bits,
		 const struct screen_type *screen_data)
{
    int (*link)(void *, void **, unsigned), status;
    void *data, *text;

    if (bits & CODELINK_LINK) {
	link = screen_data->lock_w.link;
	text = screen_data->lock_w.data;
    } else {
	link = _x1f4_e4_link_data;
	text = (void *) 0;
    }

    status = link(text, &data, sizeof(struct lxdata_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	status = init_data(data, bits, screen_data);
	if (status) {
	    if (bits & CODELINK_LINK) {
		screen_data->lock_w.free(text, data);
	    } else {
		_x1f4_e4_free_data(text, data);
	    }
	} else {
	    *lxdata = data;
	}
    }

    return status;
}
