/*
 * lxfile-f.d.c
 * Copyright (C) 2008-2011, 2013, 2016-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <e4.h>
#include <lxfile-defs.h>
#include <lxfile-drive.h>
#include <lxfile-inter.h>
#include <lxfile-types.h>

int
_libx1f4i0_lxfile_d_forward(void *context, void *output, void **input)
{
    int status;

    do {
	X1f4_E4_C_MODE mode;
	int fd;
	struct dxfile_type *dxfile_data;
	struct lxfile_type *lxfile_data;
	unsigned bits;

	dxfile_data = I_FILE(input[0]);

	lxfile_data = context;

	fd = I_MODE(input[1]);

	mode = I_MODE(input[2]);
	if (mode & (OPEN_READONLY | OPEN_WRITEONLY)) {
	    if ((mode & (OPEN_READONLY | OPEN_WRITEONLY))
		== (OPEN_READONLY | OPEN_WRITEONLY)) {
		status = _libx1f4i0_lxfile_stat_else(lxfile_data, NULL, fd);
		if (1) {
		    break;
		}
	    }
	} else {
	    status = _libx1f4i0_lxfile_stat_none(lxfile_data, NULL, fd);
	    if (1) {
		break;
	    }
	}

	if (dxfile_data->file) {
	    status = x1f4_lose_dxfile(lxfile_data, dxfile_data);
	    if (status) {
		break;
	    }
	}

	bits = lxfile_data->link_i.bits;

	if (mode & OPEN_NONBLOCKING) {
	    status = _libx1f4i0_lxfile_lock_link(lxfile_data, fd);
	    if (status) {
		break;
	    }
	} else {
	    status = _libx1f4i0_lxfile_fuse_link(lxfile_data, fd);
	    if (status) {
		break;
	    }
	}

	lxfile_data->link_i.bits = bits & ~SETCLOSE_LINK;

	if (mode & OPEN_READONLY) {
	    status = x1f4_pick_dxfile
		(lxfile_data, dxfile_data, fd, PICK_SPACE);
	} else {
	    unsigned bits = POSH_SPACE;

	    if (mode & OPEN_NONBLOCKING) {
		bits |= POSH_BLOCK;
	    }

	    status = x1f4_posh_dxfile(lxfile_data, dxfile_data, fd, bits);
	}

	if (status) {
	} else {
	    l_FILE(dxfile_data, output);
	}

	lxfile_data->link_i.bits = bits;
    } while (0);

    return status;
}

