/*
 * lxline-f.p.c
 * Copyright (C) 2008-2014, 2017, 2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxline-config.h>

#include <stddef.h>

#include <copy.h>
#include <e4.h>
#include <fd.h>
#include <l_note.h>
#include <lxcast.h>
#include <lxline-close.h>
#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>
#include <nt.h>

int
_libx1f4i0_lxline_p_forward(void *context, void *output, void **input)
{
    int status, type;
    struct dxline_type *dxline_data;
    struct lxline_type *lxline_data;
    struct lxtext_type *lxtext_data, *lxtext_link;
    unsigned miss;
    void *trap;

    lxline_data = context;

    trap = input[2];

    type = ((struct x1f4_function_type *)
	    ((void **) input[-1])[X1f4_E4_RENDITION])->args[2];
    if (type == X1f4_E4_CASE) {
	struct x1f4_dxcast_type *dxcast_data;

	dxcast_data = I_USER(trap);
	trap = &dxcast_data->data;
	type = dxcast_data->type;
    }

    lxtext_link = lxline_data->link_f.text;

    lxtext_data = lxtext_link;

    close_frame(miss, lxtext_data);

    dxline_data = I_LINE(input[0]);

    do {
	int (*copy) (void *, void *, void **, const void *,
		     const struct x1f4_nodelink_type *);
	void *data, *side;

	if (miss) {
	    copy = lxtext_data->lxtype.copy;
	    if (copy) {
		status = copy
		    (lxtext_data->lxtype.context, NULL, &side, I_USER(trap),
		     lock_excase(dxline_data));
		if (status) {
		    break;
		}
	    }
	} else {
	    copy = NULL;
	}

	if (0) {
	} else {
	    status = _libx1f4i0_post_sfnote
		(dxline_data->line, I_TEXT(input[1]),
		 effect_size, &data);
	    if (status) {
		status = _libx1f4i0_lxline_stat_text
		    (lxline_data, status, I_TEXT(input[1]));

		if (copy) {
		    lxtext_data->lxtype.free
			(lxtext_data->lxtype.context, side,
			 NULL, lock_excase(dxline_data));
		}
	    } else {
		unsigned link;

		type_effect(data, type);

		link = miss ? lxtext_data - lxtext_link : ~0;

		link_effect(data, link);

		dxline_data->size++;

		if (copy) {
		    l_USER(side, effect_data(data));
		} else {
		    copy_miss(effect_data(data), type, trap);
		}
	    }
	}
    } while (0);

    return status;
}
