/*
 * lxline-s.c.c
 * Copyright (C) 2008-2015, 2017, 2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <e4.h>
#include <fd.h>
#include <l_note.h>
#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>

int
_libx1f4i0_lxline_c_sequent(void *context, void *output, void **input)
{
    int status;

    do {
	struct dxline_type *dxline_data;
	struct lxline_type *lxline_data;
	struct lxtext_type *lxtext_data, *lxtext_link;
	void *data, *side;

	lxtext_data = context;

	dxline_data = I_LINE(input[0]);

	lxline_data = lxtext_data->text;

	lxtext_link = lxline_data->link_f.text;

	status = _libx1f4i0_post_sfnote
	    (dxline_data->line, I_TEXT(input[1]),
	     effect_size, &data);
	if (status) {
	    if (status ^ X1f4_SFNOTE_EVER_MATCH) {
		status = _libx1f4i0_lxline_stat_line(lxline_data, 0, status);
		if (1) {
		    break;
		}
	    } else {
		unsigned link;
		void *slip;

		slip = data;

		link = effect_link(slip);

		if (~link) {
		    int (*free)(void *, void *, struct excase_type **,
				const struct excase_type *);
		    struct lxtext_type *lxtext_slip;

		    lxtext_slip = lxtext_link + link;

		    free = lxtext_slip->lxtype.free;
		    if (free) {
			status = free
			    (lxtext_slip->lxtype.context,
			     E_USER(slip), NULL,
			     lock_excase(dxline_data));
			if (status) {
			    break;
			}
		    }
		}
	    }
	} else {
	    dxline_data->size++;
	}

	{
	    unsigned link, type;

	    type = lxtext_data->lxtype.code;

	    type_effect(data, type);

	    link = lxtext_data - lxtext_link;

	    status = lxtext_data->lxtype.node
		(lxtext_data->lxtype.context, &side,
		 I_USER(input[2]), lock_excase(dxline_data));
	    if (status) {
		link = ~0;
	    }

	    link_effect(data, link);

	    if (1) {
		l_USER(side, effect_data(data));
		l_USER(side, output);
	    }
	}
    } while (0);

    return status;
}
