/*
 * lxlist-c.e.c
 * Copyright (C) 2013-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <copy.h>
#include <e4.h>
#include <exerrors.h>
#include <lxcast.h>
#include <lxlist-inter.h>

#define l_LIST(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

static int rule_list(struct lxlist_type *, void *, void **);
static int slip_list(void *news);

static int
rule_list(struct lxlist_type *lxlist_data, void *list, void **input)
{
    int status = 0;
    struct x1f4_function_type *function_data;
    unsigned count;
    void **strip;

    strip = input[-1];

    function_data = strip[X1f4_E4_RENDITION];

    count = function_data->count;
    if (count) {
	struct cxbend_type *cxbend_data;
	struct x1f4_function_type function;
	void *slip[3];

	cxbend_data = lxlist_data->rule_f.cxbend + 1;

	slip[0] = strip;
	slip[1] = &list;

	function.args = function_data->args - 1;

	strip[X1f4_E4_RENDITION] = &function;

	for (; count; count--) {
	    slip[2] = *input++;
	    status = _libx1f4i0_lxlist_b_keeping(cxbend_data, NULL, slip + 1);
	    if (status) {
		break;
	    }

	    function.args++;
	}
    }

    return status;
}


static int
slip_list(void *news)
{
    int status;
    void *list, **side;

    side = news;
    list = side[1];
    if (list) {
	status = _libx1f4i0_lxlist_slip_list(*side, &list);
    } else {
	status = 0;
    }

    return status;
}


int
_libx1f4i0_lxlist_e_centric(void *context, void *output, void **input)
{
    int status;
    struct lxlist_type *lxlist_data;
    void **mind, *news;

    lxlist_data = context;

    status = lxlist_data->link_v.pick
	(lxlist_data->link_v.data, &mind, slip_list, sizeof(void *) << 1,
	 &news);
    if (status) {
	status = X1f4_EX_CRITICAL;
    } else {
	void *list, **seek;

	*mind = news;

	seek = news;

	*seek++ = lxlist_data;

	status = _libx1f4i0_lxlist_line_list(lxlist_data, &list);
	if (status) {
	    *seek = NULL;
	} else {
	    *seek = list;

	    l_LIST(list, output);

	    status = rule_list(lxlist_data, list, input);
	}
    }

    return status;
}
