/*
 * lxlist-d.g.c
 * Copyright (C) 2008-2011, 2013-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <copy.h>
#include <e4.h>
#include <fd.h>
#include <l_list.h>
#include <lxcall.h>
#include <lxlist-drive.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

int
_libx1f4i0_lxlist_g_derived(void *context, void *output, void **input)
{
    X1f4_E4_C_MODE mode;
    int status;
    struct dxlist_type *dxlist_data;
    struct lxlist_type *lxlist_data;
    unsigned size;

    lxlist_data = context;

    dxlist_data = I_LIST(input[1]);

    mode = I_MODE(input[2]);

    size = dxlist_data->size;
    if (mode < -(X1f4_E4_C_MODE) (int) size
	|| !(mode < (X1f4_E4_C_MODE) (int) size)) {
	status = _libx1f4i0_lxlist_stat_deck(lxlist_data, size, mode);
    } else {
	void *data;

	if (mode < 0) {
	    mode += size;
	}

	status = _libx1f4i0_peek_l4list(dxlist_data->list, mode, &data);
	if (status) {
	    status = _libx1f4i0_lxlist_stat_list(lxlist_data, 0, status);
	} else {
	    unsigned call, type;

	    type = effect_type(data);
	    call = *((struct x1f4_function_type *)
		     ((void **) input[-1])[X1f4_E4_RENDITION])->args;
	    if (type != call) {
		status = _libx1f4i0_lxlist_stat_type
		    (lxlist_data, mode, type, call);
	    } else {
		unsigned link;

		link = effect_link(data);

		do {
		    if (~link) {
			struct lxtext_type *lxtext_data;

			lxtext_data = lxlist_data->link_f.text;
			lxtext_data += link;
			if (lxtext_data->lxtype.lead
			    && lxtext_data->lxtype.slip) {
			    status = _libx1f4i0_lxcall_rule_node
				(&lxlist_data->link_v,
				 &lxlist_data->link_e,
				 &lxtext_data->lxtype,
				 E_USER(data), input[0]);
			    if (1) {
				break;
			    }
			}

			if (lxtext_data->lxtype.copy
			    && lxtext_data->lxtype.push) {
			    status = _libx1f4i0_lxcall_side_node
				(&lxtext_data->lxtype,
				 E_USER(data), input[0]);
			    if (1) {
				break;
			    }
			}
		    }

		    copy_miss(*input, type, effect_data(data));
		} while (0);
	    }
	}
    }

    return status;
}
