/*
 * lxlist-m.x-c.c
 * Copyright (C) 2008-2011, 2013, 2014, 2016-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxlist-config.h>

#include <e4.h>
#include <l_list.h>
#include <lxlist-drive.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

#if __HAVE_PULL_L4LIST__
# define clip(none) \
    ((struct clip_type *) (none))
#else
# error fix for lack of pull
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
typedef struct clip_type {
    unsigned type;
    void *output;
} clip_type;
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
static int clip_miss(void *, void *);
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
static int
clip_miss(void *clip, void *data)
{
    int status = 0;
    unsigned type;

    type = effect_type(data);
    switch (type) {
    case X1f4_E4_BILL:
	l_MODE(E_BILL(data), clip(clip)->output);
	break;
    case X1f4_E4_MODE:
	l_MODE(E_MODE(data), clip(clip)->output);
	break;
    case X1f4_E4_REAL:
	l_MODE(E_REAL(data), clip(clip)->output);
	break;
    default:
	clip(clip)->type = type;

	status = 1;
    }

    return status;
}
#endif				/* __HAVE_PULL_L4LIST__ */


int
_libx1f4i0_lxlist_x_press_m(void *context, void *output, void **input)
{
    int status;
    struct dxlist_type *dxlist_data;
    struct lxtext_type *lxtext_data;

    lxtext_data = context;

    dxlist_data = I_LIST(input[0]);
    if (0) {
    } else {
	X1f4_E4_C_MODE mode;
	unsigned size;

	mode = I_MODE(input[1]);

	size = dxlist_data->size;
	if (mode < -(X1f4_E4_C_MODE) (int) size
	    || !(mode < (X1f4_E4_C_MODE) (int) size)) {
	    status = _libx1f4i0_lxlist_stat_deck
		(lxtext_data->text, size, mode);
	} else {
#if __HAVE_PULL_L4LIST__
	    struct clip_type clip;
#else
	    void *data;
#endif				/* __HAVE_PULL_L4LIST__ */

	    if (mode < 0) {
		mode += size;
	    }

#if __HAVE_PULL_L4LIST__
	    clip.output = output;
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
	    status = _libx1f4i0_pull_l4list
		(dxlist_data->list, mode, &clip, clip_miss);
	    if (status) {
		status = _libx1f4i0_lxlist_stat_call
		    (lxtext_data->text, mode, clip.type, X1f4_E4_MODE, status);
	    } else {
		dxlist_data->size = size - 1;
	    }
#else
	    status = _libx1f4i0_peek_l4list(dxlist_data->list, mode, &data);
	    if (status) {
		status =
		    _libx1f4i0_lxlist_stat_list(lxtext_data->text, 0, status);
	    } else {
		unsigned type;

		type = effect_type(data);
		if (type != X1f4_E4_MODE) {
		    status = _libx1f4i0_lxlist_stat_type
			(lxtext_data->text, mode, type, X1f4_E4_MODE);
		} else {
		    X1f4_E4_C_MODE fast;

		    fast = E_MODE(data);

		    status = _libx1f4i0_miss_l4list(dxlist_data->list, mode);
		    if (status) {
			status = _libx1f4i0_lxlist_stat_list
			    (lxtext_data->text, 0, status);
		    } else {
			dxlist_data->size = size - 1;

			l_MODE(fast, output);
		    }
		}
	    }
#endif				/* __HAVE_PULL_L4LIST__ */
	}
    }

    return status;
}

