/*
 * lxlist-v.x-c.c
 * Copyright (C) 2008-2011, 2013-2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxlist-config.h>

#include <e4.h>
#include <exerrors.h>
#include <l_list.h>
#include <lxlist-drive.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

#if __HAVE_PULL_L4LIST__
# define clip(none) \
    ((struct clip_type *) (none))
#else
# error not fixed for -c
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
typedef struct clip_type {
    X1f4_E4_C_MODE mode;
    int status;
    struct dxlist_type *dxlist_data;
    struct lxtext_type *lxtext_data;
    void *output;
} clip_type;
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
static int clip_miss(void *, void *);
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
static int
clip_miss(void *clip, void *data)
{
    int status, type;
    struct lxtext_type *lxtext_data;

    lxtext_data = clip(clip)->lxtext_data;

    type = effect_type(data);
    status = _libx1f4i0_lxrail_mark_node
	(__types_rail(lxtext_data->text), clip(clip)->dxlist_data, lxtext_data,
	 &type, &effect_link(data), effect_data(data), clip(clip)->output);
    if (status) {
	if (type != lxtext_data->lxtype.code) {
	    clip(clip)->status = _libx1f4i0_lxlist_stat_type
		(lxtext_data->text, clip(clip)->mode, type,
		 lxtext_data->lxtype.code);
	} else {
	    clip(clip)->status = status;
	}
    }

    return status;
}
#endif				/* __HAVE_PULL_L4LIST__ */


int
_libx1f4i0_lxlist_x_press_v(void *context, void *output, void **input)
{
    int status;
    struct dxlist_type *dxlist_data;
    struct lxlist_type *lxlist_data;
    struct lxtext_type *lxtext_data;

    lxtext_data = context;

    lxlist_data = lxtext_data->text;

    dxlist_data = I_LIST(input[0]);
    if (0) {
    } else {
	X1f4_E4_C_MODE mode;
	unsigned size;

	mode = I_MODE(input[1]);

	size = dxlist_data->size;
	if (mode < -(X1f4_E4_C_MODE) (int) size
	    || !(mode < (X1f4_E4_C_MODE) (int) size)) {
	    status = _libx1f4i0_lxlist_stat_deck(lxlist_data, size, mode);
	} else {
#if __HAVE_PULL_L4LIST__
	    struct clip_type clip;
#else
	    void *data, *list;
#endif				/* __HAVE_PULL_L4LIST__ */

	    if (mode < 0) {
		mode += size;
	    }

#if __HAVE_PULL_L4LIST__
	    clip.dxlist_data = dxlist_data;
	    clip.lxtext_data = lxtext_data;
	    clip.mode = mode;
	    clip.output = output;
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
#else
	    list = dxlist_data->list;
#endif				/* __HAVE_PULL_L4LIST__ */

#if __HAVE_PULL_L4LIST__
	    status = _libx1f4i0_pull_l4list
		(dxlist_data->list, mode, &clip, clip_miss);
	    if (status) {
		if (status == X1f4_L4LIST_CALL_ERROR) {
		    status = clip.status;
		} else {
		    status = _libx1f4i0_lxlist_stat_list
			(lxlist_data, mode, status);
		}
	    } else {
		dxlist_data->size = size - 1;
	    }
#else
	    status = _libx1f4i0_peek_l4list(list, mode, &data);
	    if (status) {
		status = _libx1f4i0_lxlist_stat_list(lxlist_data, 0, status);
	    } else {
		unsigned type;

		type = effect_type(data);
		if (type != lxtext_data->lxtype.code) {
		    status = _libx1f4i0_lxlist_stat_type
			(lxlist_data, mode, type, lxtext_data->lxtype.code);
		} else {
		    void *fast;

		    fast = E_USER(data);

		    l_USER(fast, output);

		    status = lxtext_data->lxtype.push
			(lxtext_data->lxtype.context, output, fast,
			 lock_excase(dxlist_data));
		    if (status) {
		    } else {
			link_effect(data, ~0);

			status = _libx1f4i0_miss_l4list(list, mode);
			if (status) {
			    status = _libx1f4i0_lxlist_stat_list
				(lxlist_data, 0, status);
			} else {
			    dxlist_data->size = size - 1;
			}
		    }
		}
	    }
#endif				/* __HAVE_PULL_L4LIST__ */
	}
    }

    return status;
}
