/*
 * lxpoll-l.p.c
 * Copyright (C) 2014, 2016, 2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <signal.h>
#include <stddef.h>
#include <unistd.h>

#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>

#include <d_e101.h>
#include <e4.h>
#include <lxcall.h>
#include <lxpoll-defs.h>
#include <lxpoll-drive.h>
#include <lxpoll-inter.h>
#include <lxpoll-types.h>

#define dxpoll(dxpoll) \
    ((struct dxpoll_type *) (dxpoll))

#if defined HAVE_PSELECT
#else
# define long_trap(___l, ___d) \
    Mylene101_line_device(dxpoll(___d)->device, NULL, NULL)
#endif				/* HAVE_PSELECT */

extern const void *const x1f4_a1_walk_link;

#if defined HAVE_PSELECT
static int long_trap(struct lxpoll_type *, struct dxpoll_type *);
#endif				/* HAVE_PSELECT */
static int near_call(void *, int, void *, struct x1f4_dispatch_type **);
static int post_call(struct lxpoll_type *, struct dxpoll_type *, void **);

#if defined HAVE_PSELECT
/*
 * control is not returned for signal reception, see _bail_poll notes for more.
 */
static int
long_trap(struct lxpoll_type *lxpoll_data, struct dxpoll_type *dxpoll_data)
{
    int (*call)(void *), status;

    call = lxpoll_data->link_u.call;
    if (call) {
	sigset_t setc, setv;

	sigemptyset(&setc);
	sigaddset(&setc, SIGCHLD);

	if (sigprocmask(SIG_BLOCK, &setc, &setv)) {
	    status = Mylene101_LEAD_ERROR - 1;
	} else {
	    do {
		if (*lxpoll_data->link_u.hold) {
		    status = call(lxpoll_data->link_u.text);
		    if (status) {
			break;
		    }
		}

		status = Mylene101_line_device
		    (dxpoll_data->device, NULL, &setv);

		if (sigprocmask(SIG_SETMASK, &setv, NULL)) {
		    if (status) {
		    } else {
			status = Mylene101_LEAD_ERROR - 1;
		    }
		}
	    } while (0);
	}
    } else {
	status = Mylene101_line_device(dxpoll_data->device, NULL, NULL);
    }

    return status;
}
#endif				/* HAVE_PSELECT */


static int
near_call(void *text, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    struct dxpoll_type *dxpoll_data;
    unsigned bits;
    void *lxpoll, **track;

    track = text;

    dxpoll_data = (void *) *dispatch;
    dxpoll_data--;

    bits = (integral_q) track[BITS_LEAD];
    track[BITS_LEAD] = (void *) (integral_q) (bits & ~LIVE_LINE);

    lxpoll = track[LINK_LEAD];

    if (bits & HALF_LINE) {
	int excess;

	excess = _libx1f4i0_lxpoll_flat_line(NULL, text);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}

	dxpoll_data->notion = NULL;
    }

    if (status) {
    } else {
	if (dxpoll_data->bits & POLL_STAYED) {
	    *dispatch = NULL;

	    dxpoll_data->bits &= ~POLL_DEPEND;
	} else {
	    status = post_call(lxpoll, dxpoll_data, (void *) dispatch);
	}
    }

    return status;
}


static int
post_call(struct lxpoll_type *lxpoll_data, struct dxpoll_type *dxpoll_data,
	  void **dispatch)
{
    int status;
    struct x1f4_dispatch_type *dispatch_data;

    dxpoll_data->bits &= ~POLL_FAILED;

    dispatch_data = (void *) (dxpoll_data + 1);
    dispatch_data->back = (void *) (integral_q) -1;

    status = long_trap(lxpoll_data, dxpoll_data);
    if (status) {
	status = _libx1f4i0_lxpoll_stat_poll
	    (lxpoll_data, 0, dxpoll_data, status);
    } else {
	if (dxpoll_data->bits & POLL_FAILED) {
	    void **track;

	    track = dispatch_data->transfer.input;

	    track[LINK_LEAD] = lxpoll_data;

	    do {
		if ((integral_q) track[BITS_LEAD] & HALF_LINE) {
		    dxpoll_data->notion = track;
		}

		dxpoll_data->bits &= ~POLL_STAYED;

		dispatch_data->back = track;
		dispatch_data->call = near_call;
	    } while (0);

	    *dispatch = dispatch_data;
	} else {
	    *dispatch = NULL;

	    dxpoll_data->bits &= ~POLL_DEPEND;
	}
    }

    return status;
}


int
_libx1f4i0_lxpoll_p_logique(void *context, void *output, void **input)
{
    int status;

    if (output == x1f4_a1_walk_link) {
	struct dxpoll_type *dxpoll_data;

	dxpoll_data = I_POLL(((void **) input[1])[0]);

	if (dxpoll_data->bits & POLL_DEPEND) {
	    status = _libx1f4i0_lxpoll_stat_busy(context);
	} else {
	    dxpoll_data->bits |= POLL_DEPEND;

	    dxpoll_data->ennuie = ((void **) input[1])[-1];

	    status = post_call(context, dxpoll_data, input + 2);
	}
    } else {
	struct dxpoll_type *dxpoll_data;
	struct x1f4_dispatch_type *dispatch_data;

	dxpoll_data = I_POLL(input[0]);

	if (dxpoll_data->bits & POLL_DEPEND) {
	    status = _libx1f4i0_lxpoll_stat_busy(context);
	} else {
	    dxpoll_data->bits |= POLL_DEPEND;

	    dxpoll_data->ennuie = input[-1];

	    dispatch_data = (void *) (dxpoll_data + 1);
	    dispatch_data->back = NULL;

	    status = long_trap(context, dxpoll_data);
	    if (status) {
		status = _libx1f4i0_lxpoll_stat_poll
		    (context, 0, dxpoll_data, status);
	    }

	    dxpoll_data->bits &= ~POLL_DEPEND;
	}
    }

    return status;
}
