/*
 * lxport-i.s.c
 * Copyright (C) 2006-2014, 2016, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/wait.h>

#include <lxcall.h>
#include <lxdata.h>
#include <lxfile.h>
#include <lxlist.h>
#include <lxport-inter.h>
#include <lxport-types.h>

#define __CLOSE_PIPES__			0

#define lxport(lxport) \
    ((struct lxport_type *) (lxport))

/*
 * the opened files are dead.  closing them is a matter of keeping the number
 * of used descriptors small.  yet, there may still be data to read.
 */
int
_libx1f4i0_lxport_lose_port(void *lxport, struct waitlink_type *waitlink_data,
			    int *detail)
{
    int status = 0;
    struct waitlink_type *waitlink_last, *waitlink_next;
    struct xsport_type *xsport_data;
    void *call;

    if (detail) {
	int effect;

	effect = *detail;

	waitlink_data->status = effect;

	if (WIFEXITED(effect)) {
	    if (WEXITSTATUS(effect)) {
		status = _libx1f4i0_lxport_stat_wait
		    (lxport, waitlink_data, effect);
	    }
	} else {
	    if (WIFSIGNALED(effect)) {
		status = _libx1f4i0_lxport_stat_wait
		    (lxport, waitlink_data, effect);
	    }
	}
    }

    call = waitlink_data->path;
    if (call) {
	waitlink_data->path = NULL;
    }

    call = waitlink_data->argv;
    if (call) {
	waitlink_data->argv = NULL;

	free(call);
    }

    waitlink_last = waitlink_data->waitlink_last;
    waitlink_next = waitlink_data->waitlink_next;
    if (waitlink_last) {
	waitlink_last->waitlink_next = waitlink_next;
    } else {
	lxport(lxport)->link_u.wait = waitlink_next;
    }
    if (waitlink_next) {
	waitlink_next->waitlink_last = waitlink_last;
    }

    xsport_data = waitlink_data->xsport;
    if (xsport_data) {
#if __CLOSE_PIPES__
	int excess;
	struct x1f4_dxfile_type *dxfile_data;
#endif				/* __CLOSE_PIPES__ */

	waitlink_data->port = -1;

	waitlink_data->waitlink_last = NULL;
	waitlink_data->waitlink_next = NULL;

#if __CLOSE_PIPES__
	dxfile_data = xsport_data->line;
	if (dxfile_data->file) {
	    excess = x1f4_lose_dxfile
		(lxport(lxport)->link_l.file, dxfile_data);
	    if (excess) {
		status = _libx1f4i0_lxcall_land_slip(excess, status);
	    }
	}

	dxfile_data = xsport_data->text;
	if (dxfile_data->file) {
	    excess = x1f4_lose_dxfile
		(lxport(lxport)->link_l.file, dxfile_data);
	    if (excess) {
		status = _libx1f4i0_lxcall_land_slip(excess, status);
	    }
	}
#endif				/* __CLOSE_PIPES__ */
    } else {
	free(waitlink_data);
    }

    return status;
}
