/*
 * lxroll-a.0.c
 * Copyright (C) 2008-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <e4-m.0.h>
#include <e4.h>
#include <lxcall.h>
#include <lxroll-defs.h>
#include <lxroll-inter.h>
#include <lxroll-types.h>

#define MAKE_SINGLE(a, b)		a

#define FFLAGS_KT			X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK

#define screen(screen) \
    ((struct screen_type *) (screen))

static int init_text(struct lxroll_type *, unsigned,
		     const struct screen_type *);
static int line_text(struct lxroll_type *, unsigned,
		     const struct screen_type *, void *, void **);
static int link_text(struct lxroll_type *, unsigned,
		     const struct screen_type *);

static void bor_fftmm(u_case_args_____1);
static void born_fffm(u_case_args_____1);
static void borne_ffm(u_case_args_____1);
static void borne_fmm(u_case_args_____1);

static const struct line_type slip_line[] = {
/* *INDENT-OFF* */
#define l_accept_count			4
    {	{	"accept",
		_libx1f4i0_lxroll_a_logique,	X1f4_E4_VOID,
		NULL,				l_accept_count,
		FFLAGS_KT,			6		},
	born_fffm,							},
#define l_accept_reach			(0 + l_accept_count)
#define l_pipe_count			3
    {	{	"pipe",
		_libx1f4i0_lxroll_p_logique,	X1f4_E4_VOID,
		NULL,				l_pipe_count,
		FFLAGS_KT,			4		},
	borne_ffm,							},
#define l_pipe_reach			(l_accept_reach + l_pipe_count)
#define i_connect_count			5
    {	{	"tcpip_connect",
		_libx1f4i0_lxroll_c_forward,	X1f4_E4_VOID,
		NULL,				i_connect_count,
		FFLAGS_KT,			13		},
	bor_fftmm,							},
#define i_connect_reach			(l_pipe_reach + i_connect_count)
#define i_listen_count			3
    {	{	"tcpip_listen",
		_libx1f4i0_lxroll_l_forward,	X1f4_E4_VOID,
		NULL,				i_listen_count,
		FFLAGS_KT,			12		},
	borne_fmm,							},
#define i_listen_reach			(i_connect_reach + i_listen_count)
/* *INDENT-ON* */
#define slip_line_count                 i_listen_reach
};

static int
init_text(struct lxroll_type *lxroll_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    int status;

    if (bits & CODELINK_LINK) {
	lxroll_data->link_w = screen_data->lock_w;
    } else {
	lxroll_data->link_w.data = NULL;
	lxroll_data->link_w.free = _x1f4_e4_free_data;
	lxroll_data->link_w.link = _x1f4_e4_link_data;
	lxroll_data->link_w.mode = _x1f4_e4_mode_data;
    }

    if (bits & RESOURCE_LINK) {
	lxroll_data->link_m = screen_data->lock_m;
    } else {
	lxroll_data->link_m.data = NULL;
	lxroll_data->link_m.free = _x1f4_e4_free_data;
	lxroll_data->link_m.link = _x1f4_e4_link_data;
	lxroll_data->link_m.mode = _x1f4_e4_mode_data;
    }

    if (1) {
	lxroll_data->link_v = screen_data->lock_v;
    }

    if (bits & TEXTFLAT_LINK) {
	lxroll_data->link_e = screen_data->lock_e;
    } else {
	lxroll_data->link_e.data = NULL;
	lxroll_data->link_e.line = NULL;
	lxroll_data->link_e.post = NULL;
	lxroll_data->link_e.push = NULL;
    }

    status = link_text(lxroll_data, bits, screen_data);

    return status;
}


static int
line_text(struct lxroll_type *lxroll_data, unsigned bits,
	  const struct screen_type *screen_data, void *data, void **args)
{
    struct x1f4_linetext_type *linetext_data;

    linetext_data = data;

    if (1) {
	_libx1f4i0_lxcall_side_line
	    (lxroll_data, bits, screen_data, &linetext_data, args,
	     sizeof(slip_line) / sizeof(struct line_type), slip_line);
    }
    if (1) {
	_libx1f4i0_lxcall_zero_line(linetext_data);
    }

    return 0;
}


static int
link_text(struct lxroll_type *lxroll_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    int status;
    unsigned args_count, line_count;
    void *data;

    args_count = 0;
    line_count = 1;

    {
	args_count += slip_line_count;
	line_count += sizeof(slip_line) / sizeof(struct line_type);
    }

    status = lxroll_data->link_w.link
	(lxroll_data->link_w.data, &data,
	 args_count * sizeof(int)
	 + line_count * sizeof(struct x1f4_linetext_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	void *args;

	args = (struct x1f4_linetext_type *) data + line_count;

	lxroll_data->link_f.data = data;

	line_text(lxroll_data, bits, screen_data, data, &args);
    }

    return status;
}


static void
bor_fftmm(u_case_args_____1)
{
    int file, *line;

    line = *args;

    linetext_data->function.args = line;

    file = screen(screen)->link_i.file;

    *line++ = file;
    *line++ = file;
    *line++ = X1f4_E4_TEXT;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_MODE;

    *args = line;
}


static void
born_fffm(u_case_args_____1)
{
    int file, *line;

    line = *args;

    linetext_data->function.args = line;

    file = screen(screen)->link_i.file;

    *line++ = file;
    *line++ = file;
    *line++ = file;
    *line++ = X1f4_E4_MODE;

    *args = line;
}


static void
borne_ffm(u_case_args_____1)
{
    int file, *line;

    line = *args;

    linetext_data->function.args = line;

    file = screen(screen)->link_i.file;

    *line++ = file;
    *line++ = file;
    *line++ = X1f4_E4_MODE;

    *args = line;
}


static void
borne_fmm(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = screen(screen)->link_i.file;
    *line++ = X1f4_E4_MODE;
    *line++ = X1f4_E4_MODE;

    *args = line;
}


int
x1f4_init_lxroll(void **lxroll, unsigned bits,
	         const struct screen_type *screen_data)
{
    int (*link)(void *, void **, unsigned), status;
    void *data, *text;

    if (bits & CODELINK_LINK) {
	data = screen_data->lock_w.data;
	link = screen_data->lock_w.link;
    } else {
	data = (void *) 0;
	link = _x1f4_e4_link_data;
    }

    status = link(data, &text, sizeof(struct lxroll_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	status = init_text(text, bits, screen_data);
	if (status) {
	    if (bits & CODELINK_LINK) {
		screen_data->lock_w.free(data, text);
	    } else {
		_x1f4_e4_free_data(data, text);
	    }
	} else {
	    *lxroll = text;
	}
    }

    return status;
}
