/*
 * lxtrap-f.u.c
 * Copyright (C) 2011-2015, 2017, 2018, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <e4.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>

#define POST_BITS \
    X1f4_LXCALL_POST_ARGS | X1f4_LXCALL_POST_DATA | X1f4_LXCALL_POST_LINE

#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

#define lxtrap(trap) \
    ((struct lxtrap_type *) (trap))

extern const void *const x1f4_a1_walk_link;

typedef struct pick_type {
    struct fxdata_type finite[2];
    struct x1f4_dxcast_type dxcast, serial[2];
    struct x1f4_function_type function;
    struct x1f4_linetext_type linetext;
} pick_type;

typedef struct slip_type {
    int (*call)(void *, int, void *, struct x1f4_dispatch_type **), deck[2],
	miss;
    const int *over;
    int *enter, *point, *reach, *still;
    struct fxdata_type finite[2];
    struct lxtrap_type *lxtrap_data;
    struct x1f4_dispatch_type *dispatch_data;
    struct x1f4_dxcast_type dxcast, serial[2];
    struct x1f4_function_type function;
    struct x1f4_linetext_type linetext, *linetext_data;
    unsigned class;
    void *back, *clear[3], **order, **plane, **track;
} slip_type;

static int fold_call(struct lxtrap_type *, void *, void **,
		     struct x1f4_linetext_type *, int,
		     const struct x1f4_function_type *, int, int, int);
static int lift_call(struct lxtrap_type *, void *, void **,
		     struct x1f4_linetext_type *, int,
		     const struct x1f4_function_type *, int, int, int);
static int part_call(void *, int, void *, struct x1f4_dispatch_type **);

/*
 * function does setup for an unused output
 */
static int
fold_call(struct lxtrap_type *lxtrap_data, void *output, void **input,
	  struct x1f4_linetext_type *linetext_data, int lead,
	  const struct x1f4_function_type *function_text, int line, int jack,
	  int lock)
{
    int linked, status;
    unsigned class, count, grade, parts, scale = 0;
    void *beta;

    class = function_text->count;

    count = class - lead - line;

    grade = line + 2;

    parts = function_text->flags;

    if (parts & X1f4_E4_POST_TYPE) {
	if (linetext_data->function.flags
	    & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	    scale = 1;
	} else {
	    parts &= ~X1f4_E4_POST_TYPE;
	}
    }

    linked = sizeof(struct pick_type);

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &beta,
	 linked + (sizeof(void *) + (sizeof(int) << scale)) * grade
	 + sizeof(void *));
    if (status) {
	status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
    } else {
	int *args, deck[2];
	const int *over;
	int *enter, *point, *reach, *still, type;
	struct pick_type *call_data;
	struct x1f4_function_type *function_call;
	unsigned bits;
	void **frame, **order, **plane;

	call_data = beta;

	function_call = &call_data->function;

	function_call->count = grade;

	function_call->flags = parts & X1f4_E4_POST_TYPE;

	frame = (void *) (call_data + 1);
	frame++;

	function_call->args = args = (void *) (frame + grade);

	frame[-1] = input[-1];

	type = linetext_data->function.type;

	if (type != X1f4_E4_CASE) {
	    l_USER(&call_data->dxcast, output);

	    call_data->dxcast.call = 0;
	    call_data->dxcast.type = type;

	    output = &call_data->dxcast.data;
	} else {
	}

	bits = POST_BITS;

	over = function_text->args;

	_libx1f4i0_lxcall_keep_call
	    (frame, input - count, parts, args, over - count, line,
	     lead + count, jack, lock);

	if (jack < lock) {
	} else {
	    jack++;
	}

	deck[0] = jack;
	deck[1] = lock;

	reach = args + jack;
	still = args + lock;

	order = frame + jack;
	plane = frame + lock;

	input += lead + line;
	over += lead + line;

	line += 2;

	*reach = *over;
	*order = *input;
	if (scale) {
	    enter = reach + line;
	    *enter = over[class];
	} else {
	    enter = NULL;
	}

	input++;
	over++;

	*still = *over;
	*plane = *input;
	if (scale) {
	    point = still + line;
	    *point = over[class];
	} else {
	    point = NULL;
	}

	count >>= 1;

	{
	    /* the args, count, flags set should do */
	    call_data->linetext = *linetext_data;

	    call_data->linetext.function.type = X1f4_E4_CASE;

	    status = _libx1f4i0_lxcall_post_call
		(&call_data->linetext, function_call, output, 0, frame, bits,
		 &lxtrap_data->link_v, &lxtrap_data->link_p,
		 &lxtrap_data->link_q, &lxtrap_data->link_e,
		 lxtrap_data->trap_e.eelookup_data);
	    if (status) {
		count = 0;
	    } else {
		count--;
	    }
	}
	for (; count; count--) {
	    /* the args, count, flags set should do */
	    call_data->linetext = *linetext_data;

	    call_data->linetext.function.type = X1f4_E4_CASE;

	    input++;
	    over++;

	    *reach = *over;
	    *order = *input;
	    if (enter) {
		*enter = over[class];
	    }

	    input++;
	    over++;

	    *still = *over;
	    *plane = *input;
	    if (point) {
		*point = over[class];
	    }

	    status = _libx1f4i0_lxcall_rail_call
		(&call_data->linetext, function_call, output, 0, frame, bits,
		 &lxtrap_data->link_p, &lxtrap_data->link_q,
		 &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data, 2,
		 deck, call_data->finite, call_data->serial);
	    if (status) {
		break;
	    }
	}
    }

    return status;
}


/*
 * function does setup for an unused output
 */
static int
lift_call(struct lxtrap_type *lxtrap_data, void *output, void **input,
	  struct x1f4_linetext_type *linetext_data, int lead,
	  const struct x1f4_function_type *function_text, int line, int jack,
	  int lock)
{
    int linked, status;
    unsigned class, count, grade, parts, scale = 0;
    void *beta, **track;

    class = function_text->count;

    count = class - lead - line;

    grade = line + 2;

    parts = function_text->flags;

    if (parts & X1f4_E4_POST_TYPE) {
	if (linetext_data->function.flags
	    & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	    scale = 1;
	} else {
	    parts &= ~X1f4_E4_POST_TYPE;
	}
    }

    linked = sizeof(struct slip_type);

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &beta,
	 linked + (sizeof(void *) + (sizeof(int) << scale)) * grade
	 + sizeof(void *));
    if (status) {
	status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
    } else {
	int *args;
	const int *over;
	int type;
	struct slip_type *call_data;
	struct x1f4_function_type *function_call;
	void **clear, **frame, *series;

	call_data = beta;

	function_call = &call_data->function;

	function_call->count = grade;

	function_call->flags = parts & X1f4_E4_POST_TYPE;

	frame = (void *) ((char *) beta + linked);
	frame++;

	function_call->args = args = (void *) (frame + grade);

	track = input[1];

	frame[-1] = track[-1];

	type = linetext_data->function.type;

	series = input[0];

	if (type != X1f4_E4_CASE) {
	    l_USER(&call_data->dxcast, series);

	    call_data->dxcast.call = 0;
	    call_data->dxcast.type = type;

	    series = &call_data->dxcast.data;
	} else {
	}

	over = function_text->args;

	_libx1f4i0_lxcall_keep_call
	    (frame, track - count, parts, args, over - count, line,
	     lead + count, jack, lock);

	if (jack < lock) {
	} else {
	    jack++;
	}

	call_data->reach = args + jack;
	call_data->still = args + lock;

	call_data->order = frame + jack;
	call_data->plane = frame + lock;

	track += lead + line;
	over += lead + line;

	line += 2;

	*call_data->reach = *over;
	*call_data->order = *track;
	if (scale) {
	    call_data->enter = call_data->reach + line;
	    *call_data->enter = over[class];
	} else {
	    call_data->enter = NULL;
	}

	track++;
	over++;

	*call_data->still = *over;
	*call_data->plane = *track;
	if (scale) {
	    call_data->point = call_data->still + line;
	    *call_data->point = over[class];
	} else {
	    call_data->point = NULL;
	}

	clear = call_data->clear;

	clear[0] = series;
	clear[1] = frame;

	call_data->linetext = *linetext_data;

	call_data->linetext.function.type = X1f4_E4_CASE;

	status = _libx1f4i0_lxcall_post_call
	    (&call_data->linetext, function_call, output, 0, clear, POST_BITS,
	     &lxtrap_data->link_v, &lxtrap_data->link_p,
	     &lxtrap_data->link_q, &lxtrap_data->link_e,
	     lxtrap_data->trap_e.eelookup_data);
	if (status) {
	} else {
	    struct x1f4_dispatch_type *dispatch_data;

	    dispatch_data = input[2] = clear[2];

	    if (dispatch_data) {
		call_data->back = dispatch_data->back;
		call_data->call = dispatch_data->call;
		call_data->miss = count >> 1;
		call_data->over = over;

		call_data->class = class;
		call_data->track = track;

		call_data->deck[0] = jack;
		call_data->deck[1] = lock;

		call_data->lxtrap_data = lxtrap_data;

		call_data->linetext_data = linetext_data;

		dispatch_data->back = call_data;
		dispatch_data->call = part_call;

		call_data->linetext.function.type = type;

		call_data->dispatch_data = dispatch_data;
	    }
	}
    }

    return status;
}


static int
part_call(void *back, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    do {
	int slip;
	struct slip_type *call_data;

	if (status) {
	    break;
	}

	call_data = back;

	status = call_data->call(call_data->back, 0, output, dispatch);
	if (status) {
	    break;
	}

	slip = call_data->miss;
	slip--;
	if (slip) {
	    const int *over;
	    struct lxtrap_type *lxtrap_data;
	    struct x1f4_dispatch_type *dispatch_data;
	    void **frame, **track;

	    call_data->miss = slip;

	    lxtrap_data = call_data->lxtrap_data;

	    frame = (void *) (call_data + 1);
	    frame++;

	    over = call_data->over;
	    track = call_data->track;

	    over++;
	    track++;

	    *call_data->reach = *over;
	    *call_data->order = *track;
	    if (call_data->enter) {
		*call_data->enter = over[call_data->class];
	    }

	    over++;
	    track++;

	    *call_data->still = *over;
	    *call_data->plane = *track;
	    if (call_data->point) {
		*call_data->point = over[call_data->class];
	    }

	    call_data->over = over;
	    call_data->track = track;

	    /* the args, count, flags set should do */
	    call_data->linetext = *call_data->linetext_data;

	    call_data->linetext.function.type = X1f4_E4_CASE;

	    call_data->clear[2] = call_data->dispatch_data;

	    status = _libx1f4i0_lxcall_into_call
		(&call_data->linetext, &call_data->function, 0,
		 call_data->clear, POST_BITS | X1f4_LXCALL_POST_PLOT,
		 &lxtrap_data->link_p, &lxtrap_data->link_q,
		 &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data, 2,
		 call_data->deck, call_data->finite, call_data->serial);
	    if (status) {
		break;
	    }

	    dispatch_data = call_data->clear[2];

	    if (dispatch_data) {
		call_data->back = dispatch_data->back;
		call_data->call = dispatch_data->call;

		dispatch_data->back = call_data;
		dispatch_data->call = part_call;

		call_data->linetext.function.type = call_data->dxcast.type;

		*dispatch = dispatch_data;
	    }
	} else {
	    *dispatch = NULL;
	}
    } while (0);

    return status;
}


int
_libx1f4i0_lxtrap_u_forward(void *context, void *output, void **input)
{
    int status;

    do {
	X1f4_E4_C_MODE jack, line, lock;
	struct lxtrap_type *lxtrap_data;
	struct x1f4_dxcast_type *dxcast_data;
	const struct x1f4_function_type *function_text;
	struct x1f4_linetext_type *linetext_data;
	unsigned class, grade;
	void **track;

	lxtrap_data = context;

	if (output == x1f4_a1_walk_link) {
	    track = input[1];
	} else {
	    track = input;
	}

	dxcast_data = I_USER(track[0]);

	if (X1f4_E4_CALL < dxcast_data->type) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_roll(lxtrap_data, "__u2call", 8);
	    break;
	}

	linetext_data = I_USER(&dxcast_data->data);

	if (linetext_data->function.function) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_zero(lxtrap_data, "__u2call", 8);
	    break;
	}

	function_text = ((void **) track[-1])[X1f4_E4_RENDITION];

	class = function_text->count - 4;

	line = I_MODE(track[1]);
	if (line < 0 || class < line) {
	    status = _libx1f4i0_lxtrap_stat_line(lxtrap_data, line, class);
	    break;
	}

	if ((class - line) & 1) {
	    status = _libx1f4i0_lxtrap_stat_odds(lxtrap_data);
	    break;
	}

	grade = line + 1;

	jack = I_MODE(track[2]);

	if (jack < -(int) grade || !(jack < (int) grade)) {
	    status = _libx1f4i0_lxtrap_stat_land(context, grade, jack);

	    break;
	}

	if (jack < (X1f4_E4_C_MODE) 0) {
	    jack += grade;
	}

	grade++;

	lock = I_MODE(track[3]);

	if (lock < -(int) grade || !(lock < (int) grade)) {
	    status = _libx1f4i0_lxtrap_stat_land(context, grade, lock);

	    break;
	}

	if (lock < 0) {
	    lock += grade;
	}

	if (line < class) {
	    if (output == x1f4_a1_walk_link) {
		status = lift_call
		    (lxtrap_data, output, input, linetext_data, 4,
		     function_text, line, jack, lock);
	    } else {
		status = fold_call
		    (lxtrap_data, output, input, linetext_data, 4,
		     function_text, line, jack, lock);
	    }
	} else {
	    status = 0;

	    if (output == x1f4_a1_walk_link) {
		input[2] = NULL;
	    }
	}
    } while (0);

    return status;
}
