/*
 * sfar.f.c
 * Copyright (C) 2009-2012, 2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <sfas-defs.h>
#include <sfas-inter.h>
#include <sfas-names.h>
#include <sfas-types.h>

#define true(e)				1

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define sfnote(sfnote)			((struct sfnote_type *) (sfnote))

typedef struct fplink_type {
    struct fplink_type *fplink_data;
    struct fpnode_type fpnode;
} fplink_type;

static int join_ever(void *, void *);
static int join_half(void *, struct fpnode_type *, int, struct fpnode_type *);
static int join_last(void *, struct fpnode_type *, struct fpnode_type *,
		     unsigned, int, unsigned, byte *);
static int join_lead(void *, void *);
static int join_node(void *, void *, struct fpnode_type *,
		     struct fpnode_type *, int);
static int join_pair(void *, void *);
static int join_post(void *, struct fpnode_type *, int, struct fpnode_type *,
		     unsigned, int, unsigned, byte *, unsigned, byte *);
static int join_used(void *, void *);
static int rule_node(struct fpnode_type *, struct fpnode_type *, unsigned, int,
		     int *, unsigned *, byte **, unsigned *, byte **);

static int
join_ever(void *sfdata, void *sftext)
{
    int status = 0;
    struct fpnode_type *fpnode_data, *fpnode_slip, *fpnode_text;
    unsigned data, effect;

    fpnode_data = sfnote(sfdata)->link_a.fpdeck.fpnode_data;
    fpnode_text = sfnote(sftext)->link_a.fpdeck.fpnode_data;

    data = fpnode_data->bits;

    effect = Effect(data);

    fpnode_slip = fpnode_data->data;

    if (prefix(fpnode_slip + effect) ^ prefix(fpnode_text)) {
	do {
	    unsigned fast;
	    void *node;

	    fast = effect + 1;

	    node = fpnode_slip;

	    if (fast & 3) {
	    } else {
		status = sfnote(sfdata)->link_m.mode
		    (sfnote(sfdata)->link_m.data, &node,
		     sizeof(struct fpnode_type) * (fast + 4));
		if (status) {
		    status = LINK_ERROR;
		    if (1) {
			break;
		    }
		}
	    }

	    fpnode_slip = node;
	    fpnode_slip += fast;

	    *fpnode_slip = *fpnode_text;

	    fpnode_data->data = node;

	    fpnode_data->bits = eset(fast + 1);

	    sfnote(sftext)->link_a.fpdeck.fpnode_data = NULL;

            status = sfnote(sfdata)->link_m.free
                (sfnote(sfdata)->link_m.data, fpnode_text);
            if (status) {
                status = FREE_ERROR;
                if (1) {
                    break;
		}
	    }
	} while (0);
    } else {
	do {
	    struct fpnode_type *fpnode_miss;
	    void *node;

	    status = sfnote(sfdata)->link_m.link
		(sfnote(sfdata)->link_m.data, &node,
		 sizeof(struct fpnode_type) << 2);
	    if (status) {
		status = LINK_ERROR;
		if (1) {
		    break;
		}
	    }

	    fpnode_miss = node;

	    *fpnode_miss = *fpnode_text;

	    fpnode_text->data = node;
	    fpnode_text->bits = eset(1);

	    status = join_node(sfdata, sftext, fpnode_data, fpnode_text, 0);
	} while (0);
    }

    return status;
}


static int
join_half(void *sfdata, struct fpnode_type *fpnode_data, int join,
	  struct fpnode_type *fpnode_text)
{
    int status = 0;

    do {
	int excess;
	unsigned data, head, spread, text, tail;
	void *node;

	node = fpnode_data->data;

	head = fpnode_data->bits;

	tail = fpnode_text->bits;

	data = Spread(head);

	text = Spread(tail);

	spread = data - join + text;

	if (((data + 3) ^ (spread + 3)) & ~3) {
	    status = sfnote(sfdata)->link_m.mode
		(sfnote(sfdata)->link_m.data, &node,
		 sizeof(struct fpnode_type) * ((spread + 3) & ~3));
	    if (status) {
		status = LINK_ERROR;
		if (1) {
		    break;
		}
	    }

	    fpnode_data->data = node;
	}

	memcpy((struct fpnode_type *) node + data,
	       (struct fpnode_type *) fpnode_text->data + join,
	       sizeof(struct fpnode_type) * (text - join));

	fpnode_data->bits = kset(fpnode_data->bits, spread);

	if (sizeof(void *) < Length(tail)) {
	    status = sfnote(sfdata)->link_m.free
		(sfnote(sfdata)->link_m.data, fpnode_text->base);
	    if (status) {
		status = FREE_ERROR;
	    }
	}

	excess = sfnote(sfdata)->link_m.free
	    (sfnote(sfdata)->link_m.data, fpnode_text->data);
	if (excess) {
	    if (status) {
	    } else {
		status = FREE_ERROR;
	    }
	}
    } while (0);

    return status;
}


static int
join_last(void *sfdata, struct fpnode_type *fpnode_data,
	  struct fpnode_type *fpnode_text, unsigned post, int prefix,
	  unsigned ever, byte *text)
{
    int status = 0;

    do {
	int excess;
	struct fpnode_type *fpnode_slip;
	unsigned head, spread, tail;
	void *node;

	node = fpnode_data->data;

	head = fpnode_data->bits;

	spread = Spread(head);

	if (spread & 3) {
	} else {
	    status = sfnote(sfdata)->link_m.mode
		(sfnote(sfdata)->link_m.data, &node,
		 sizeof(struct fpnode_type) * (spread + 4));
	    if (status) {
		status = LINK_ERROR;
		if (1) {
		    break;
		}
	    }

	    fpnode_data->data = node;
	}

	fpnode_slip = node;

	fpnode_slip += spread;

	*fpnode_slip = *fpnode_text;

	tail = fpnode_text->bits;

	text += prefix;

	if (Extant(tail)) {
	    ever -= prefix;
	    ever--;

	    if (sizeof(void *) < ever) {
		void *base;

		base = fpnode_slip->base;

		fpnode_slip->bits = pset(*text, ever, tail);

		memmove(base, text + 1, ever);

		status = sfnote(sfdata)->link_m.mode
		    (sfnote(sfdata)->link_m.data, &base, ever);
		if (status) {
		    status = MODE_ERROR;
		}

		fpnode_slip->base = base;
	    } else {
		fpnode_slip->bits = pset(*text, ever, tail);

		memcpy(&fpnode_slip->base, text + 1, ever);

		if (sizeof(void *) < post + prefix + 1 + ever) {
		    excess = sfnote(sfdata)->link_m.free
			(sfnote(sfdata)->link_m.data, fpnode_text->base);
		    if (excess) {
			if (status) {
			} else {
			    status = FREE_ERROR;
			}
		    }
		} else {
		}
	    }
	} else {
	    fpnode_slip->bits = zset(tail, *text);
	}

	fpnode_data->data = node;
	fpnode_data->bits = nset(prefix, head, spread + 1);
    } while (0);

    return status;
}


static int
join_lead(void *sfdata, void *sftext)
{
    int status = 0;
    struct fpnode_type *fpnode_data, *fpnode_miss, *fpnode_text;
    unsigned text;

    fpnode_data = sfnote(sfdata)->link_a.fpdeck.fpnode_data;
    fpnode_text = sfnote(sftext)->link_a.fpdeck.fpnode_data;

    text = fpnode_text->bits;

    fpnode_miss = fpnode_text->data;

    if (prefix(fpnode_data) ^ prefix(fpnode_miss)) {
	do {
	    struct fpnode_type *fpnode_slip;
	    unsigned fast;
	    void *node;

	    fast = Spread(text);

	    node = fpnode_miss;

	    if (fast & 3) {
	    } else {
		status = sfnote(sfdata)->link_m.mode
		    (sfnote(sfdata)->link_m.data, &node,
		     sizeof(struct fpnode_type) * (fast + 4));
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		}
	    }

	    fpnode_slip = node;

	    memmove(fpnode_slip + 1, fpnode_slip,
		    sizeof(struct fpnode_type) * fast);

	    *fpnode_slip = *fpnode_data;

	    fpnode_data->data = node;

	    fpnode_data->bits = eset(1 + fast);

	    sfnote(sftext)->link_a.fpdeck.fpnode_data = NULL;

            status = sfnote(sfdata)->link_m.free
                (sfnote(sfdata)->link_m.data, fpnode_text);
            if (status) {
                status = FREE_ERROR;
                if (1) {
                    break;
		}
	    }
	} while (0);
    } else {
	do {
	    struct fpnode_type *fpnode_slip;
	    void *node;

	    status = sfnote(sfdata)->link_m.link
		(sfnote(sfdata)->link_m.data, &node,
		 sizeof(struct fpnode_type) << 2);
	    if (status) {
		status = LINK_ERROR;
		if (1) {
		    break;
		}
	    }

	    fpnode_slip = node;

	    *fpnode_slip = *fpnode_data;

	    fpnode_data->data = node;
	    fpnode_data->bits = eset(1);

	    status = join_node(sfdata, sftext, fpnode_data, fpnode_text, 0);
	} while (0);
    }

    return status;
}


static int
join_node(void *sfdata, void *sftext, struct fpnode_type *fpnode_data,
	  struct fpnode_type *fpnode_text, int offset)
{
    int join = 0, status;
    struct fplink_type *fplink_data = NULL;
    unsigned post = 0;

    while (1) {
	byte *data, *text;
	int course, prefix;
	unsigned ever, lead;

	course = rule_node
	    (fpnode_data, fpnode_text, post, offset, &prefix, &lead, &data,
	     &ever, &text);
	if (course) {
	    if (course < 0) {
		if (true(extant(fpnode_text))) {
		    struct fpnode_type *fpnode_miss;

		    fpnode_miss = fpnode_text->data;

		    if (prefix(fpnode_miss) ^ data[prefix]) {
			status = join_post
			    (sfdata, fpnode_data, 0, fpnode_text, post, prefix,
			     lead, data, ever, text);
			if (1) {
			    break;
			}
		    } else {
			struct fplink_type *fplink_trap;

			fplink_trap = malloc(sizeof(struct fplink_type));
			if (fplink_trap) {
			} else {
			    status = LINK_ERROR;
			    if (1) {
				break;
			    }
			}

			fplink_trap->fplink_data = fplink_data;
			fplink_data = fplink_trap;

			fplink_trap->fpnode = *fpnode_miss;

			status = join_post
			    (sfdata, fpnode_data, 1, fpnode_text, post, prefix,
			     lead, data, ever, text);
			if (status) {
			    break;
			}

			prefix++;

			fpnode_data = fpnode_miss;
			fpnode_text = &fplink_trap->fpnode;

			offset += prefix;

			post = 0;
		    }
		} else {
		}
	    } else {
		if (extant(fpnode_data)) {
		    struct fpnode_type *fpnode_slip;
		    unsigned head;

		    fpnode_slip = fpnode_data->data;

		    head = fpnode_data->bits;

		    if (prefix(fpnode_slip + Effect(head)) ^ text[prefix]) {
			status = join_last
			    (sfdata, fpnode_data, fpnode_text, post, prefix,
			     ever, text);
			if (1) {
			    break;
			}
		    } else {
			prefix++;
			offset += prefix;

			fpnode_data = fpnode_slip + Effect(head);

			post += prefix;
		    }
		} else {
		    join = 1;
		}
	    }
	} else {
	    if (prefix ^ lead) {
		join = 1;
	    } else {
		if (extant(fpnode_data)) {
		    if (true(extant(fpnode_text))) {
			struct fpnode_type *fpnode_miss, *fpnode_slip;
			unsigned head;

			fpnode_slip = fpnode_data->data;

			fpnode_miss = fpnode_text->data;

			head = fpnode_data->bits;

			if (prefix(fpnode_slip + Effect(head))
			    ^ prefix(fpnode_miss)) {
			    status = join_half
				(sfdata, fpnode_data, 0, fpnode_text);
			    if (1) {
				break;
			    }
			} else {
			    int effect;
			    struct fplink_type *fplink_trap;

			    fplink_trap = malloc(sizeof(struct fplink_type));
			    if (fplink_trap) {
			    } else {
				status = LINK_ERROR;
				if (1) {
				    break;
				}
			    }

			    fplink_trap->fplink_data = fplink_data;
			    fplink_data = fplink_trap;

			    effect = effect(fpnode_data);

			    fplink_trap->fpnode = *fpnode_miss;

			    status = join_half
				(sfdata, fpnode_data, 1, fpnode_text);
			    if (status) {
				break;
			    }

			    prefix++;
			    offset += prefix;

			    fpnode_data = fpnode_data->data;
			    fpnode_data += effect;

			    fpnode_text = &fplink_trap->fpnode;
			    post = 0;
			}
		    } else {
		    }
		} else {
		    if (true(extant(fpnode_text))) {
			status = join_post
			    (sfdata, fpnode_data, 0, fpnode_text, post, prefix,
			     lead, data, ever, text);
			if (1) {
			    break;
			}
		    } else {
		    }
		}
	    }
	}

	if (join) {
	    int excess;
	    struct fpnode_type *fpnode_slip;
	    unsigned head, tail;
	    void *node;

	    status = sfnote(sfdata)->link_m.link
		(sfnote(sfdata)->link_m.data, &node,
		 sizeof(struct fpnode_type) << 2);
	    if (status) {
		status = LINK_ERROR;
		if (1) {
		    break;
		}
	    }

	    fpnode_slip = node;

	    *fpnode_slip = *fpnode_data;

	    head = fpnode_data->bits;

	    data += prefix;

	    if (Extant(head)) {
		lead -= prefix;
		lead--;

		if (sizeof(void *) < lead) {
		    void *base;

		    status = sfnote(sfdata)->link_m.link
			(sfnote(sfdata)->link_m.data, &base, lead);
		    if (status) {
			sfnote(sfdata)->link_m.free
			    (sfnote(sfdata)->link_m.data, node);

			status = LINK_ERROR;

			break;
		    }

		    fpnode_slip->base = base;
		    fpnode_slip->bits = pset(*data, lead, head);

		    memcpy(base, data + 1, lead);
		} else {
		    fpnode_slip->bits = pset(*data, lead, head);

		    memcpy(&fpnode_slip->base, data + 1, lead);
		}
	    } else {
		fpnode_slip->bits = zset(head, *data);
	    }

	    fpnode_slip++;

	    *fpnode_slip = *fpnode_text;

	    tail = fpnode_text->bits;

	    text += prefix;

	    if (Extant(tail)) {
		ever -= prefix;
		ever--;

		if (sizeof(void *) < ever) {
		    void *base;

		    base = fpnode_slip->base;

		    fpnode_slip->bits = pset(*text, ever, tail);

		    memmove(base, text + 1, ever);

		    status = sfnote(sfdata)->link_m.mode
			(sfnote(sfdata)->link_m.data, &base, ever);
		    if (status) {
			status = MODE_ERROR;
		    }

		    fpnode_slip->base = base;
		} else {
		    fpnode_slip->bits = pset(*text, ever, tail);

		    memcpy(&fpnode_slip->base, text + 1, ever);

		    if (sizeof(void *) < post + prefix + 1 + ever) {
			excess = sfnote(sfdata)->link_m.free
			    (sfnote(sfdata)->link_m.data, fpnode_text->base);
			if (excess) {
			    if (status) {
			    } else {
				status = FREE_ERROR;
			    }
			}
		    } else {
		    }
		}
	    } else {
		fpnode_slip->bits = zset(tail, *text);
	    }

	    fpnode_data->data = node;
	    fpnode_data->bits = nset(prefix, fpnode_data->bits, 2);

	    if (Extant(head)) {
		if (sizeof(void *) < prefix) {
		    void *base;

		    base = fpnode_data->base;

		    excess = sfnote(sfdata)->link_m.mode
			(sfnote(sfdata)->link_m.data, &base, prefix);
		    if (excess) {
			if (status) {
			} else {
			    status = MODE_ERROR;
			}
		    } else {
			fpnode_data->base = base;
		    }
		} else {
		    if (sizeof(void *) < prefix + 1 + lead) {
			byte *base;

			base = fpnode_data->base;

			memcpy(&fpnode_data->base, base, prefix);

			excess = sfnote(sfdata)->link_m.free
			    (sfnote(sfdata)->link_m.data, base);
			if (excess) {
			    if (status) {
			    } else {
				status = FREE_ERROR;
			    }
			}
		    } else {
		    }
		}
	    } else {
		if (sizeof(void *) < prefix) {
		    void *base;

		    excess = sfnote(sfdata)->link_m.link
			(sfnote(sfdata)->link_m.data, &base, prefix);
		    if (excess) {
			fpnode_data->bits = lset(0, fpnode_data->bits);

			if (status) {
			} else {
			    status = LINK_ERROR;
			}
		    } else {
			memcpy(base, data - prefix, prefix);

			fpnode_data->base = base;
		    }
		} else {
		    memcpy(&fpnode_data->base, data - prefix, prefix);
		}
	    }

	    break;
	}
    }

    while (fplink_data) {
	struct fplink_type *fplink_free;

	fplink_free = fplink_data;
	fplink_data = fplink_data->fplink_data;
	free(fplink_free);
    }

    fpnode_text = sfnote(sftext)->link_a.fpdeck.fpnode_data;

    sfnote(sftext)->link_a.fpdeck.fpnode_data = NULL;

    if (true(option)) {
	int excess;

	excess = sfnote(sfdata)->link_m.free
	    (sfnote(sfdata)->link_m.data, fpnode_text);
	if (excess) {
	    if (status) {
	    } else {
		status = FREE_ERROR;
	    }
	}
    }

    return status;
}


static int
join_pair(void *sfdata, void *sftext)
{
    int status;
    void *node;

    status = sfnote(sfdata)->link_m.link
	(sfnote(sfdata)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip, *fpnode_text;

	fpnode_data = sfnote(sfdata)->link_a.fpdeck.fpnode_data;
	fpnode_text = sfnote(sftext)->link_a.fpdeck.fpnode_data;

	fpnode_slip = node;

	fpnode_slip[0] = *fpnode_data;
	fpnode_slip[1] = *fpnode_text;

	fpnode_data->bits = eset(2);
	fpnode_data->data = fpnode_slip;

	sfnote(sftext)->link_a.fpdeck.fpnode_data = NULL;

	status = sfnote(sfdata)->link_m.free
	    (sfnote(sfdata)->link_m.data, fpnode_text);
	if (status) {
	    status = FREE_ERROR;
	}
    }

    return status;
}


static int
join_post(void *sfdata, struct fpnode_type *fpnode_data, int join,
	  struct fpnode_type *fpnode_text, unsigned post, int prefix,
	  unsigned lead, byte *data, unsigned ever, byte *text)
{
    int status = 0;

    do {
	int excess;
	struct fpnode_type *fpnode_slip;
	unsigned head, spread, tail;
	void *node;

	head = fpnode_data->bits;

	node = fpnode_text->data;

	tail = fpnode_text->bits;

	spread = Spread(tail);

	if (join) {
	    fpnode_slip = node;
	} else {
	    if (spread & 3) {
	    } else {
		status = sfnote(sfdata)->link_m.mode
		    (sfnote(sfdata)->link_m.data, &node,
		     sizeof(struct fpnode_type) * (spread + 4));
		if (status) {
		    status = LINK_ERROR;
		    if (1) {
			break;
		    }
		}
	    }

	    fpnode_slip = node;

	    memmove(fpnode_slip + 1, fpnode_slip,
		    sizeof(struct fpnode_type) * spread);
	}

	*fpnode_slip = *fpnode_data;

	data += prefix;

	if (Extant(head)) {
	    lead -= prefix;
	    lead--;

	    if (sizeof(void *) < lead) {
		void *base;

		status = sfnote(sfdata)->link_m.link
		    (sfnote(sfdata)->link_m.data, &base, lead);
		if (status) {
		    fpnode_slip->bits = pset(*data, 0, head);

		    status = LINK_ERROR;

		    break;
		}

		fpnode_slip->base = base;
		fpnode_slip->bits = pset(*data, lead, head);

		memcpy(base, data + 1, lead);
	    } else {
		fpnode_slip->bits = pset(*data, lead, head);

		memcpy(&fpnode_slip->base, data + 1, lead);
	    }
	} else {
	    fpnode_slip->bits = zset(head, *data);
	}

	fpnode_data->data = node;
	fpnode_data->bits = nset(prefix, head, spread + 1 - join);

	if (Extant(head)) {
	    if (sizeof(void *) < prefix) {
		void *base;

		base = fpnode_data->base;

		excess = sfnote(sfdata)->link_m.mode
		    (sfnote(sfdata)->link_m.data, &base, prefix);
		if (excess) {
		    if (status) {
		    } else {
			status = MODE_ERROR;
		    }
		} else {
		    fpnode_data->base = base;
		}
	    } else {
		if (sizeof(void *) < prefix + 1 + lead) {
		    byte *base;

		    base = fpnode_data->base;

		    memcpy(&fpnode_data->base, base, prefix);

		    excess = sfnote(sfdata)->link_m.free
			(sfnote(sfdata)->link_m.data, base);
		    if (excess) {
			if (status) {
			} else {
			    status = FREE_ERROR;
			}
		    }
		} else {
		}
	    }
	} else {
	    if (sizeof(void *) < prefix) {
		void *base;

		excess = sfnote(sfdata)->link_m.link
		    (sfnote(sfdata)->link_m.data, &base, prefix);
		if (excess) {
		    fpnode_data->bits = lset(0, fpnode_data->bits);

		    if (status) {
		    } else {
			status = LINK_ERROR;
		    }
		} else {
		    memcpy(base, data - prefix, prefix);

		    fpnode_data->base = base;
		}
	    } else {
		memcpy(&fpnode_data->base, data - prefix, prefix);
	    }
	}

	if (sizeof(void *) < Length(tail)) {
	    excess = sfnote(sfdata)->link_m.free
		(sfnote(sfdata)->link_m.data, fpnode_text->base);
	    if (excess) {
		if (status) {
		} else {
		    status = FREE_ERROR;
		}
	    }
	}
    } while (0);

    return status;
}


static int
join_used(void *sfdata, void *sftext)
{
    int status = 0;
    struct fpnode_type *fpnode_data, *fpnode_miss, *fpnode_slip, *fpnode_text;
    unsigned data, text;

    fpnode_data = sfnote(sfdata)->link_a.fpdeck.fpnode_data;
    fpnode_text = sfnote(sftext)->link_a.fpdeck.fpnode_data;

    data = fpnode_data->bits;
    text = fpnode_text->bits;

    fpnode_slip = fpnode_data->data;
    fpnode_slip += Effect(data);

    fpnode_miss = fpnode_text->data;

    if (prefix(fpnode_slip) ^ prefix(fpnode_miss)) {
	do {
	    int excess;
	    unsigned fast, lead;

	    lead = Spread(data);
	    fast = lead + Spread(text);
	    if (((lead + 3) ^ (fast + 3)) & ~3) {
		void *node;

		node = fpnode_data->data;
		status = sfnote(sfdata)->link_m.mode
		    (sfnote(sfdata)->link_m.data, &node,
		     sizeof(struct fpnode_type) * ((fast + 3) & ~3));
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		}

		fpnode_data->data = node;

		fpnode_slip = node;
		fpnode_slip += Effect(data);
	    } else {
	    }

	    memcpy(fpnode_slip + 1, fpnode_miss,
		   sizeof(struct fpnode_type) * Spread(text));

	    fpnode_data->bits = iset(data, fast);

	    sfnote(sftext)->link_a.fpdeck.fpnode_data = NULL;

            status = sfnote(sfdata)->link_m.free
                (sfnote(sfdata)->link_m.data, fpnode_miss);
            if (status) {
                status = FREE_ERROR;
	    }

            excess = sfnote(sfdata)->link_m.free
                (sfnote(sfdata)->link_m.data, fpnode_text);
            if (excess) {
		if (status) {
		} else {
		    status = FREE_ERROR;
		}
	    }
	} while (0);
    } else {
	status = join_node(sfdata, sftext, fpnode_data, fpnode_text, 0);
    }

    return status;
}


static int
rule_node(struct fpnode_type *fpnode_data, struct fpnode_type *fpnode_text,
	  unsigned post, int offset, int *prefix, unsigned *mind, byte **hike,
	  unsigned *slip, byte **fold)
{
    byte *data, *text;
    int course = 0;
    unsigned bits, call, ever, lead, size;

    bits = fpnode_data->bits;

    lead = Length(bits);
    if (Extant(bits)) {
	if (sizeof(byte *) < lead) {
	    data = fpnode_data->base;
	    if (lead < Record) {
	    } else {
		lead += strlen(back(data) + Record);
	    }
	} else {
	    data = (void *) &fpnode_data->base;
	}
    } else {
	data = fpnode_data->base;
	if (lead < Record) {
	} else {
	    lead += strlen(back(data) + Record);
	}

	data += offset;
	lead -= offset;
    }

    *hike = data;
    *mind = lead;

    bits = fpnode_text->bits;

    ever = Length(bits);
    if (Extant(bits)) {
	if (sizeof(byte *) < ever) {
	    text = fpnode_text->base;
	    if (ever < Record) {
	    } else {
		ever += strlen(back(text) + Record);
	    }
	} else {
	    text = (void *) &fpnode_text->base;
	}

	ever -= post;
	text += post;
    } else {
	text = fpnode_text->base;
	if (ever < Record) {
	} else {
	    ever += strlen(back(text) + Record);
	}

	ever -= offset;
	text += offset;
    }

    *fold = text;
    *slip = ever;

    call = ever < lead ? ever : lead;

    size = call;
    while (size) {
	if (*data ^ *text) {
	    break;
	} else {
	    data++;
	    text++;
	    size--;
	}
    }

    *prefix = call - size;

    if (size) {
    } else {
	if (ever < lead) {
	    course = -1;
	} else {
	    if (ever ^ lead) {
		course = 1;
	    } else {
	    }
	}
    }

    return course;
}


int
x1f4_join_sfnote(void *sfdata, void *sftext)
{
    int status;
    struct fpnode_type *fpnode_text;

    fpnode_text = sfnote(sftext)->link_a.fpdeck.fpnode_data;
    if (fpnode_text) {
	struct fpnode_type *fpnode_data;

	fpnode_data = sfnote(sfdata)->link_a.fpdeck.fpnode_data;
	if (fpnode_data) {
	    unsigned data, text;

	    data = fpnode_data->bits;
	    text = fpnode_text->bits;

	    if (Prefix(data)) {
		if (Prefix(text)) {
		    if (Prefix(data) ^ Prefix(text)) {
			status = join_pair(sfdata, sftext);
		    } else {
			status = join_node
			    (sfdata, sftext, fpnode_data, fpnode_text, 1);
		    }
		} else {
		    status = join_lead(sfdata, sftext);
		}
	    } else {
		if (Extant(data)) {
		    if (Prefix(text)) {
			status = join_ever(sfdata, sftext);
		    } else {
			if (true(Extant(text))) {
			    status = join_used(sfdata, sftext);
			} else {
			}
		    }
		} else {
		    if (Prefix(text)) {
			status = join_pair(sfdata, sftext);
		    } else {
			status = join_lead(sfdata, sftext);
		    }
		}
	    }
	} else {
	    status = 0;

	    sfnote(sfdata)->link_a.fpdeck.fpnode_data = fpnode_text;
	    sfnote(sftext)->link_a.fpdeck.fpnode_data = NULL;
	}
    } else {
	status = 0;
    }

    return status;
}
